#include "precomp.h"
#include "AncestorEnum.h" 

#include "ClassFactoryChunk.h"
#include "ClassFactoryMethod.h"
#include "../classhierarchy/Project.h"
#include <bfc/parse/paramparser.h>
#include "SecondaryAncestor.h"
#include "../classhierarchy/Workspace.h"

AncestorEnum::AncestorEnum() {
  m_chunk = NULL;
  m_workspace = NULL;
}

AncestorEnum::AncestorEnum(ClassFactoryChunk *chunk, Workspace *workspace) {
  m_chunk = chunk;
  m_workspace = workspace;
  enumerate();
}

AncestorEnum::~AncestorEnum() {
  m_ancestors.deleteAll();
}

void AncestorEnum::setChunk(ClassFactoryChunk *chunk) {
  m_chunk = chunk;
  enumerate();
}

void AncestorEnum::setWorkspace(Workspace *workspace) {
  m_workspace = workspace;
}

int AncestorEnum::findAncestor(const char *ancestorname) {
  foreach(m_ancestors)
    AncestorEntry *ae = m_ancestors.getfor();
    if (STRCASEEQL(ae->m_chunk->getClassName(), ancestorname))
      return foreach_index;
  endfor;
  return -1;
}

int AncestorEnum::getNumAncestors() {
  return m_ancestors.getNumItems();
}

AncestorEntry *AncestorEnum::enumAncestor(int n) {
  return m_ancestors.enumItem(n);
}

void AncestorEnum::enumerate() {
  ASSERT(m_chunk != NULL);
  m_ancestors.deleteAll();
  if (m_workspace == NULL) return;
  int i, n = m_chunk->getNumSecondaryAncestors();
  for (i=0;i<n;i++) {
    SecondaryAncestor *anc = m_chunk->enumSecondaryAncestor(i);
    const char *ancestor = anc->getClassName();
    ClassFactoryChunk *chunk = m_workspace->findChunk(ancestor);
    m_ancestors.addItem(new AncestorEntry(chunk, anc->getScope()));
  }
}
