#include "precomp.h"
#include "ClassFactoryBlockI.h"
#include "ClassFactoryMethodI.h"
#include "ClassFactoryDataI.h"
#include "ClassFactorySubChunkI.h"
#include "bfc/std.h"

// leaking

ClassFactoryBlockI::ClassFactoryBlockI() : method(NULL), codeblock(NULL) {
  m_data = NULL;
  m_chunk = NULL;
  m_subchunk = NULL;
}

ClassFactoryBlockI::ClassFactoryBlockI(ClassFactoryMethod *m) : method(static_cast<ClassFactoryMethodI *>(m)), codeblock(NULL), m_data(NULL), m_subchunk(NULL) { 
  m_chunk = NULL;
  m->setParentBlock(this);
}

ClassFactoryBlockI::ClassFactoryBlockI(ClassFactoryCodeBlock *cb) : method(NULL), m_data(NULL), codeblock(static_cast<ClassFactoryCodeBlockI *>(cb)), m_subchunk(NULL) { 
  m_chunk = NULL;
}

ClassFactoryBlockI::ClassFactoryBlockI(ClassFactoryData *data) : m_data(static_cast<ClassFactoryDataI *>(data)), method(NULL), codeblock(NULL), m_subchunk(NULL) { 
  m_chunk = NULL;
  data->setParentBlock(this);
}

ClassFactoryBlockI::ClassFactoryBlockI(ClassFactorySubChunk *subchunk) : m_data(NULL), m_subchunk(static_cast<ClassFactorySubChunkI *>(subchunk)), method(NULL), codeblock(NULL) { 
  m_chunk = NULL;
  m_subchunk->setParentBlock(this);
}

ClassFactoryBlockI::~ClassFactoryBlockI() { 
  delete method; 
  delete codeblock;
  delete m_data; 
  delete m_subchunk;
}

void ClassFactoryBlockI::copy(ClassFactoryBlock *block) {
  delete method; method = NULL; 
  delete codeblock; codeblock = NULL;
  delete m_data; m_data = NULL;
  switch (block->getBlockType()) {
    case CF_BLOCKTYPE_METHOD: {
      ClassFactoryMethodI *m = new ClassFactoryMethodI();
      m->copy(block->getMethod());
      setMethod(m);
      break;
    }
    case CF_BLOCKTYPE_CODEBLOCK: {
      ClassFactoryCodeBlockI *cb = new ClassFactoryCodeBlockI();
      cb->copy(block->getCodeBlock());
      setCodeBlock(cb);
      break;
    }
    case CF_BLOCKTYPE_DATA: {
      ClassFactoryDataI *data = new ClassFactoryDataI();
      data->copy(block->getData());
      setData(data);
      break;
    }
    case CF_BLOCKTYPE_SUBCHUNK: {
      ClassFactorySubChunkI *subchunk = new ClassFactorySubChunkI();
      subchunk->copy(block->getSubChunk());
      setSubChunk(subchunk);
      break;
    }
  }
}

int ClassFactoryBlockI::getBlockType() { 
  if (method != NULL) 
    return CF_BLOCKTYPE_METHOD; 
  if (codeblock != NULL) 
    return CF_BLOCKTYPE_CODEBLOCK; 
  if (m_data != NULL) 
    return CF_BLOCKTYPE_DATA; 
  if (m_subchunk != NULL) 
    return CF_BLOCKTYPE_SUBCHUNK; 
  return CF_BLOCKTYPE_UNDEFINED; 
}

ClassFactoryMethod *ClassFactoryBlockI::getMethod() { 
  return method; 
}

ClassFactoryData *ClassFactoryBlockI::getData() { 
  return m_data; 
}

ClassFactoryCodeBlock *ClassFactoryBlockI::getCodeBlock() { 
  return codeblock; 
}

ClassFactorySubChunk *ClassFactoryBlockI::getSubChunk() {
  return m_subchunk;
}

void ClassFactoryBlockI::setSubChunk(ClassFactorySubChunk *chunk) {
  if (m_subchunk != chunk) {
    delete m_subchunk;
  }
  m_subchunk = static_cast<ClassFactorySubChunkI*>(chunk);
  method = NULL;
  m_data = NULL;
  codeblock = NULL;
}

void ClassFactoryBlockI::setCodeBlock(ClassFactoryCodeBlock *block) { 
  if (codeblock != block) {
    delete codeblock;
  }
  codeblock = static_cast<ClassFactoryCodeBlockI *>(block); 
  method = NULL;  
  m_data = NULL;
  m_subchunk = NULL;
}

void ClassFactoryBlockI::setMethod(ClassFactoryMethod *m) { 
  if (method != m) {
    delete method;
  }
  method = static_cast<ClassFactoryMethodI*>(m); 
  method->setParentBlock(this);
  codeblock = NULL; 
  m_data = NULL;
  m_subchunk = NULL;
}

void ClassFactoryBlockI::setData(ClassFactoryData *data) { 
  if (data != m_data) {
    delete m_data;
  }
  m_data = static_cast<ClassFactoryDataI *>(data); 
  method = NULL;  
  codeblock = NULL;  
  m_subchunk = NULL;
}

void ClassFactoryBlockI::setFlag(int flag) { 
  flag_setFlag(flag); 
}

int ClassFactoryBlockI::isFlagSet(int flag) { 
  return flag_isFlagSet(flag); 
}

void ClassFactoryBlockI::clearFlag(int flag) { 
  flag_clearFlag(flag); 
}

int ClassFactoryBlockI::getStartLine() {
  return m_startline;
}

int ClassFactoryBlockI::getNumLines() {
  return m_numlines;
}

void ClassFactoryBlockI::setStartLine(int start) {
  m_startline = start;
}

void ClassFactoryBlockI::setNumLines(int num) {
  m_numlines = num;
}

void ClassFactoryBlockI::setParentChunk(ClassFactoryChunk *chunk) {
  m_chunk = chunk;
}


ClassFactoryChunk *ClassFactoryBlockI::getParentChunk() {
  return m_chunk;
}
