#ifndef __CLASSFACTORYCHUNKI_H
#define __CLASSFACTORYCHUNKI_H

#include "ClassFactoryChunk.h"

// todo : add support for raw params on other contexts than class

/*[interface.header.h]
class XmlReaderParams;

enum {
  CF_FLAG_METHOD_FOUND_IN_CPP=0,
  CF_FLAG_METHOD_FOUND_IN_H=100,
  CF_FLAG_DATA_FOUND=110,
  CF_FLAG_CLASS_NEW_IMPLEMETATION=200,
  CF_FLAG_METHOD_IMPLEMENTED=300,
  CF_FLAG_HAS_CPP_CODEBLOCKS=400,
  CF_FLAG_HAS_H_CODEBLOCKS=500,
};
*/

/*<?<autoheader/>*/
#include "ClassFactoryChunk.h"
#include "ClassFactoryChunkX.h"

class ClassFactoryBlock;
class ClassFactoryChunk;
class ClassFactoryCodeBlock;
class ClassFactoryMethod;
class HelpSection;
class SecondaryAncestor;
/*?>*/

#include "ClassFactoryBlockI.h"
#include "ClassFactoryCodeBlockI.h"
#include "HelpSectionI.h"
#include "FlagHolder.h"
#include "SecondaryAncestorI.h"

#include <bfc/string/string.h>
#include <bfc/ptrlist.h>
#include <bfc/tlist.h>
#include <api/xml/xmlparams.h>

using namespace wasabi;

/*<? <classdecl name="ClassFactoryChunk" implname="ClassFactoryChunkI" factory="Interface" inherit="FlagHolder"/>*/
class ClassFactoryChunkI : public ClassFactoryChunkX,
                           public FlagHolder {
/*?>*/
public:
  NODISPATCH ClassFactoryChunkI(const char *_classname = NULL);
  NODISPATCH virtual ~ClassFactoryChunkI();
  DISPATCH(30) virtual const char *getClassName();
  DISPATCH(40) virtual int getNumBlocks();
  DISPATCH(50) virtual ClassFactoryBlock *enumBlock(int n);
  DISPATCH(60) virtual void addBlock(ClassFactoryBlock *b, int pos = -1);
  DISPATCH(70) virtual void removeBlock(ClassFactoryBlock *b);
  DISPATCH(80) virtual void setClassName(const char *_classname);
  DISPATCH(90) virtual XmlReaderParams *getParams();
  DISPATCH(91) virtual XmlReaderParams *getRawParams();
  DISPATCH(100) virtual void setParams(XmlReaderParams *_params);
  DISPATCH(101) virtual void setRawParams(XmlReaderParams *_params);
  DISPATCH(110) virtual void dropCodeBlocks(const char *btype, int methodstoo);
  DISPATCH(120) virtual int getBlockNum(ClassFactoryBlock *block);
  DISPATCH(130) virtual void setImplementationName(const char *_implname);
  DISPATCH(140) virtual const char *getImplementationName();
  DISPATCH(150) virtual ClassFactoryCodeBlock *getHeader(const char *type);
  DISPATCH(160) virtual ClassFactoryCodeBlock *getFooter(const char *type);
  DISPATCH(170) virtual void setHeader(ClassFactoryCodeBlock *block);
  DISPATCH(180) virtual void setFooter(ClassFactoryCodeBlock *block);
  DISPATCH(190) virtual ClassFactoryMethod *getConstructor(int n);
  DISPATCH(200) virtual ClassFactoryMethod *getDestructor(int n);
  DISPATCH(210) virtual void setFlag(int flag);
  DISPATCH(220) virtual int isFlagSet(int flag);
  DISPATCH(230) virtual void clearFlag(int flag);
  DISPATCH(240) virtual void addErrorMessage(const char *msg, int col = -1);
  DISPATCH(250) virtual int getNumErrors();
  DISPATCH(260) virtual const char *enumError(int n, int *col);
  DISPATCH(270) virtual void clearErrors();
  DISPATCH(280) virtual void setAncestor(const char *_ancestor);
  DISPATCH(290) virtual const char *getAncestor();
  DISPATCH(300) virtual void copy(ClassFactoryChunk *c);
  DISPATCH(310) virtual void clearAllFlags(int methodstoo = 1);
  DISPATCH(320) virtual int haveCodeBlocks(const char *btype);
  DISPATCH(330) virtual void deleteBlock(ClassFactoryBlock *block);
  DISPATCH(340) virtual void setClassKW(const char *_kw);
  DISPATCH(350) virtual const char *getClassKW();
  DISPATCH(360) virtual int isConstructor(ClassFactoryMethod *m);
  DISPATCH(370) virtual int isDestructor(ClassFactoryMethod *m);
  DISPATCH(380) virtual void setClassType(int ct);
  DISPATCH(390) virtual int getClassType();
  DISPATCH(400) virtual int getNumSecondaryAncestors();
  DISPATCH(410) virtual SecondaryAncestor *enumSecondaryAncestor(int n);
  DISPATCH(420) virtual void removeSecondaryAncestor(const char *anc);
  DISPATCH(430) virtual void removeSecondaryAncestorNum(int n);
  DISPATCH(440) virtual void addSecondaryAncestor(const char *anc, int scope);
  DISPATCH(450) virtual int moveBlock(ClassFactoryBlock *block, int newpos);
  DISPATCH(460) virtual void addHelpSection(HelpSection *section);
  DISPATCH(470) virtual int getNumHelpSections();
  DISPATCH(480) virtual HelpSection *enumHelpSection(int n);
  DISPATCH(490) virtual void deleteHelpSection(HelpSection *s);
  DISPATCH(500) virtual ClassFactoryChunk *getParentChunk();
  DISPATCH(510) virtual void setParentChunk(ClassFactoryChunk *chunk);

private:
  String classname;
  String implname;
  PtrList < ClassFactoryBlockI > blocks;
  PtrList < ClassFactoryCodeBlockI > headers;
  PtrList < ClassFactoryCodeBlockI > footers;
  PtrList < SecondaryAncestorI > secancestors;
  XmlReaderParamsI params;
  XmlReaderParamsI raw_params;
  PtrList < String > errors;
  TList < int > errcols;
  String autoimplname;
  String ancestor;
  PtrList < HelpSectionI > m_helpsections;
  ClassFactoryChunk * m_parentChunk;
};

class ClassFactoryChunkSortInterface {
  public:
    static int compareItem(ClassFactoryChunk *p1, ClassFactoryChunk *p2) {
      return STRICMP(p1->getClassName(), p2->getClassName());
    }
    static int compareAttrib(const char *attrib, ClassFactoryChunk *item) {
      return STRICMP(attrib, item->getClassName());
    }
};

class ClassFactoryChunkSortImplementation {
  public:
    static int compareItem(ClassFactoryChunk *p1, ClassFactoryChunk *p2) {
      return STRICMP(p1->getImplementationName(), p2->getImplementationName());
    }
    static int compareAttrib(const char *attrib, ClassFactoryChunk *item) {
      return STRICMP(attrib, item->getImplementationName());
    }
};

#endif // __CLASSFACTORYCHUNKI_H
