#include "precomp.h"
#include "ClassFactoryDataI.h"
#include "ClassFactoryMethodI.h"


ClassFactoryDataI::ClassFactoryDataI() {
  m_parent = NULL;
  m_sourceline = -1;
}

ClassFactoryDataI::ClassFactoryDataI(const char *name, const char *type) {
  m_parent = NULL;
  m_type = type;
  m_name = name;
  m_sourceline = -1;
}

ClassFactoryDataI::~ClassFactoryDataI() {
}

void ClassFactoryDataI::copy(ClassFactoryData *data) {
  flag_copy(static_cast<ClassFactoryDataI*>(data));
  setDataType(data->getDataType());
  setCtorCode(data->getCtorCode());
  setDtorCode(data->getDtorCode());
  setName(data->getName());
  setParentBlock(data->getParentBlock());
  while (xparams.getNbItems()) xparams.removeItem(xparams.getItemName(0));
  XmlReaderParams *par = data->getParams();
  int p, np = par->getNbItems();
  for (p=0;p<np;p++) {
    xparams.addItem(par->getItemName(p), par->getItemValue(p));
  }
}

const char *ClassFactoryDataI::getName() {
  return m_name;
}

void ClassFactoryDataI::setName(const char *name) {
  m_name = name;
}

void ClassFactoryDataI::setDataType(const char *type) {
  m_type = type;
}

const char *ClassFactoryDataI::getDataType() {
  return m_type;
}

void ClassFactoryDataI::setInitValue(const char *val) {
  xparams.removeItem("init");
  xparams.addItem("init", val);
}

const char *ClassFactoryDataI::getInitValue() {
  return xparams.getItemValue2("init");
}

void ClassFactoryDataI::setCtorCode(const char *code) {
  m_ctorcode = code;
}

void ClassFactoryDataI::setDtorCode(const char *code) {
  m_dtorcode = code;
}

const char *ClassFactoryDataI::getCtorCode() {
  return m_ctorcode;
}

const char *ClassFactoryDataI::getDtorCode() {
  return m_dtorcode;
}

void ClassFactoryDataI::setParentBlock(ClassFactoryBlock *block) {
  m_parent = block;
}

ClassFactoryBlock *ClassFactoryDataI::getParentBlock() {
  return m_parent;
}

void ClassFactoryDataI::setScope(int s) {
  xparams.removeItem("scope");
  switch (s) {
    case SCOPE_PRIVATE: xparams.addItem("scope", "private"); break;
    case SCOPE_PROTECTED: xparams.addItem("scope", "protected"); break;
  }
}

int ClassFactoryDataI::getScope() {
  String s = xparams.getItemValue("scope");
  if (s.iscaseequal("private")) return SCOPE_PRIVATE;
  else if (s.iscaseequal("protected")) return SCOPE_PROTECTED;
  else if (s.iscaseequal("public")) return SCOPE_PUBLIC;
  else return SCOPE_PUBLIC;
}

XmlReaderParams *ClassFactoryDataI::getParams() { 
  return &xparams; 
}

void ClassFactoryDataI::setParams(XmlReaderParams *_params) {
  for (int i=0;i<_params->getNbItems();i++) {
    const char *par = _params->getItemName(i);
    if (STRCASEEQL(par, "name")) continue;
    if (STRCASEEQL(par, "type")) continue;
    xparams.addItem(_params->getItemName(i), _params->getItemValue(i));
  }
}

void ClassFactoryDataI::setClassName(const char *name) {
  m_classname = name;
}

const char *ClassFactoryDataI::getClassName() {
  if (m_classname.isempty()) return NULL; // use parent chunk.getClassName();
  return m_classname;
}

void ClassFactoryDataI::setFlag(int flag) { flag_setFlag(flag); }

int ClassFactoryDataI::isFlagSet(int flag) { return flag_isFlagSet(flag); }

void ClassFactoryDataI::clearFlag(int flag) { flag_clearFlag(flag); }

int ClassFactoryDataI::isStatic() {
  return xparams.getItemValueInt("static");
}

void ClassFactoryDataI::setStatic(int is) {
  xparams.removeItem("static");
  if (is) xparams.addItem("static", "1");
}

void ClassFactoryDataI::setSourceLine(int sl) {
  xparams.replaceItem("line", StringPrintf("%d", sl));
}

int ClassFactoryDataI::getSourceLine() {
  return ATOI(xparams.getItemValue("line"));
}

