#include "precomp.h"
#include "ClassFactoryMethodI.h"
#include "ClassFactoryBlock.h"
#include "ClassFactoryChunk.h"

ClassFactoryMethodI::ClassFactoryMethodI(const char *_name, const char *_rettype, const char *_defret) : name(_name), rettype(_rettype), defret(_defret), m_sourceline(-1) { 
  parentblock = NULL;
  ln_hstart = ln_cppstart = ln_inlinestart = ln_hnum = ln_cppnum = ln_inlinenum = -1;
}

ClassFactoryMethodI::~ClassFactoryMethodI() { 
  params.deleteAll(); 
  codeblocks.deleteAll(); 
  headers.deleteAll(); 
  footers.deleteAll(); 
}

void ClassFactoryMethodI::copy(ClassFactoryMethod *m) {
  ClassFactoryMethodI *_m = static_cast<ClassFactoryMethodI *>(m);
  while (xparams.getNbItems()) xparams.removeItem(xparams.getItemName(0));
  XmlReaderParams *par = m->getParams();
  int p, np = par->getNbItems();
  for (p=0;p<np;p++) {
    xparams.addItem(par->getItemName(p), par->getItemValue(p));
  }
  flag_copy(static_cast<ClassFactoryMethodI*>(m));
  setName(m->getName());
  setReturnType(m->getReturnType());
  setDefaultReturnValue(m->getDefaultReturnValue());
  setClassName(m->getClassName());
  codeblocks.deleteAll();
  headers.deleteAll();
  footers.deleteAll();
  params.deleteAll();
  int cb, ncb = _m->codeblocks.getNumItems();
  for (cb=0;cb<ncb;cb++) {
    ClassFactoryCodeBlockI *block = new ClassFactoryCodeBlockI();
    block->copy(_m->codeblocks.enumItem(cb));
    codeblocks.addItem(block);
  }
  p, np = _m->params.getNumItems();
  for (p=0;p<np;p++) {
    ClassFactoryParamI *param = new ClassFactoryParamI();
    param->copy(_m->params.enumItem(p));
    params.addItem(param);
  }
  int cf, ncf = _m->footers.getNumItems();
  for (cf=0;cf<ncf;cf++) {
    ClassFactoryCodeBlockI *block = new ClassFactoryCodeBlockI();
    block->copy(_m->footers.enumItem(cf));
    footers.addItem(block);
  }
  int ch, nch = _m->headers.getNumItems();
  for (ch=0;ch<nch;ch++) {
    ClassFactoryCodeBlockI *block = new ClassFactoryCodeBlockI();
    block->copy(_m->headers.enumItem(ch));
    headers.addItem(block);
  }
}

const char *ClassFactoryMethodI::getName() { 
  return name;
} 

const char *ClassFactoryMethodI::getReturnType() { 
  return rettype; 
}

const char *ClassFactoryMethodI::getDefaultReturnValue() { 
  return defret; 
}

int ClassFactoryMethodI::getNumParams() { 
  return params.getNumItems(); 
}

ClassFactoryParam *ClassFactoryMethodI::enumParam(int n) { 
  return params[n]; 
}

void ClassFactoryMethodI::addParam(ClassFactoryParam *p) { 
  params.addItem(static_cast<ClassFactoryParamI *>(p)); 
}

void ClassFactoryMethodI::setName(const char *_name) { 
  name = _name;
} 

void ClassFactoryMethodI::setReturnType(const char *_rettype) { 
  rettype = _rettype; 
}

void ClassFactoryMethodI::setDefaultReturnValue(const char *_defret) { 
  defret = _defret; 
}

XmlReaderParams *ClassFactoryMethodI::getParams() { 
  return &xparams; 
}

void ClassFactoryMethodI::setParams(XmlReaderParams *_params) {
  for (int i=0;i<_params->getNbItems();i++) {
    const char *par = _params->getItemName(i);
    if (STRCASEEQL(par, "name")) continue;
    if (STRCASEEQL(par, "type")) continue;
    if (STRCASEEQL(par, "default")) continue;
    if (STRCASEEQL(par, "hstart")) { ln_hstart = ATOI(_params->getItemValue(i)); continue; }
    if (STRCASEEQL(par, "cppstart")) { ln_cppstart = ATOI(_params->getItemValue(i)); continue; }
    if (STRCASEEQL(par, "inlinestart")) { ln_inlinestart = ATOI(_params->getItemValue(i)); continue; }
    if (STRCASEEQL(par, "hnum")) { ln_hnum = ATOI(_params->getItemValue(i)); continue; }
    if (STRCASEEQL(par, "cppnum")) { ln_cppnum = ATOI(_params->getItemValue(i)); continue; }
    if (STRCASEEQL(par, "inlinenum")) { ln_inlinenum = ATOI(_params->getItemValue(i)); continue; }
    xparams.addItem(_params->getItemName(i), _params->getItemValue(i));
  }
}

void ClassFactoryMethodI::dropCodeBlocks(const char *type) {
  if (type == NULL) { 
    codeblocks.deleteAll(); 
    headers.deleteAll();
    footers.deleteAll();
    return; 
  }
  int i;
  for (i=0;i<codeblocks.getNumItems();i++) {
    ClassFactoryCodeBlockI *cb = codeblocks.enumItem(i);
    const char *t = cb->getParams()->getItemValue("type");
    if (STRCASEEQLSAFE(type, t)) {
      codeblocks.removeItem(cb);
      delete cb;
      i--;
    }
  }
  for (i=0;i<headers.getNumItems();i++) {
    ClassFactoryCodeBlockI *cb = headers.enumItem(i);
    const char *t = cb->getParams()->getItemValue("type");
    if (STRCASEEQLSAFE(type, t)) {
      headers.removeItem(cb);
      delete cb;
      i--;
    }
  }
  for (i=0;i<footers.getNumItems();i++) {
    ClassFactoryCodeBlockI *cb = footers.enumItem(i);
    const char *t = cb->getParams()->getItemValue("type");
    if (STRCASEEQLSAFE(type, t)) {
      footers.removeItem(cb);
      delete cb;
      i--;
    }
  }
}

void ClassFactoryMethodI::setCodeBlock(ClassFactoryCodeBlock *cb) { 
  const char *type = cb->getParams()->getItemValue("type");
  for (int i=0;i<codeblocks.getNumItems();i++) {
    ClassFactoryCodeBlockI *cb2 = codeblocks.enumItem(i);
    const char *t = cb2->getParams()->getItemValue("type");
    if (STRCASEEQLSAFE(t, type)) {
      codeblocks.removeItem(cb2);
      delete cb2;
      break;
    }
  }
  codeblocks.addItem(static_cast<ClassFactoryCodeBlockI *>(cb)); 
}

ClassFactoryCodeBlock *ClassFactoryMethodI::getCodeBlock(const char *type) { 
  for(int i=0;i<codeblocks.getNumItems();i++) {
    ClassFactoryCodeBlockI *cb = codeblocks.enumItem(i);
    const char *t = cb->getParams()->getItemValue("type");
    if (STRCASEEQLSAFE(t, type)) return cb;
  }
  return NULL; 
}

void ClassFactoryMethodI::setFlag(int flag) { flag_setFlag(flag); }

int ClassFactoryMethodI::isFlagSet(int flag) { return flag_isFlagSet(flag); }

void ClassFactoryMethodI::clearFlag(int flag) { flag_clearFlag(flag); }

void ClassFactoryMethodI::setClassName(const char *name) {
  classname = name;
}

const char *ClassFactoryMethodI::getClassName() {
  if (classname.isempty()) return NULL; // use parent chunk.getClassName();
  return classname;
}

void ClassFactoryMethodI::setHeader(const char *header, const char *type) {
  foreach(headers)
    ClassFactoryCodeBlockI *cb = headers.getfor();
    if (STRCASEEQLSAFE(cb->getType(), type)) {
      headers.removeItem(cb);
      delete cb;
      break;
    }
  endfor;
  if (STRCASEEQLSAFE(type, "h")) {
    if (STRCASEEQLSAFE(header, xparams.getItemValue2("postdecl")))
      return;
  }
  ClassFactoryCodeBlockI *cb = new ClassFactoryCodeBlockI(header, type);
  headers.addItem(cb);
}

void ClassFactoryMethodI::setFooter(const char *footer, const char *type) {
  foreach(footers)
    ClassFactoryCodeBlockI *cb = footers.getfor();
    if (STRCASEEQLSAFE(cb->getType(), type)) {
      footers.removeItem(cb);
      delete cb;
      break;
    }
  endfor;
  ClassFactoryCodeBlockI *cb = new ClassFactoryCodeBlockI(footer, type);
  footers.addItem(cb);
}

void ClassFactoryMethodI::setPostDeclModifier(const char *postdeclmodifier) {
  xparams.removeItem("postdecl");
  if (STRCASEEQLSAFE(postdeclmodifier, "=0") || STRCASEEQLSAFE(postdeclmodifier, "= 0")) {
    setVirtual(1, 1);
    return;
  }
  if (STRCASEEQLSAFE(postdeclmodifier, getHeader("h"))) {
    return;
  }
  xparams.addItem("postdecl", postdeclmodifier);
}

const char *ClassFactoryMethodI::getPostDeclModifier() {
  return xparams.getItemValue2("postdecl");
}

const char *ClassFactoryMethodI::getHeader(const char *type) {
  foreach(headers)
    ClassFactoryCodeBlockI *cb = headers.getfor();
    if (STRCASEEQLSAFE(cb->getType(), type)) {
      return cb->getCodeBlock();
    }
  endfor;
  return NULL;
}

const char *ClassFactoryMethodI::getFooter(const char *type) {
  foreach(footers)
    ClassFactoryCodeBlockI *cb = footers.getfor();
    if (STRCASEEQLSAFE(cb->getType(), type)) {
      return cb->getCodeBlock();
    }
  endfor;
  return NULL;
}

void ClassFactoryMethodI::setStatic(int s) {
  xparams.removeItem("static");
  if (s) xparams.addItem("static", StringPrintf("%d", s));
}

int ClassFactoryMethodI::isStatic() {
  return xparams.getItemValueInt("static");
}

void ClassFactoryMethodI::setScope(int s) {
  xparams.removeItem("scope");
  switch (s) {
    case SCOPE_PRIVATE: xparams.addItem("scope", "private"); break;
    case SCOPE_PROTECTED: xparams.addItem("scope", "protected"); break;
  }
}

int ClassFactoryMethodI::getScope() {
  String s = xparams.getItemValue("scope");
  if (s.iscaseequal("private")) return SCOPE_PRIVATE;
  else if (s.iscaseequal("protected")) return SCOPE_PROTECTED;
  else if (s.iscaseequal("public")) return SCOPE_PUBLIC;
  else return SCOPE_PUBLIC;
}

void ClassFactoryMethodI::setInline(int s, int where) {
  xparams.removeItem("inline");
  xparams.removeItem("inlinebody");
  if (s) xparams.addItem("inline", StringPrintf("%d", s));
  if (where) xparams.addItem("inlinebody", StringPrintf("%d", where));
}

int ClassFactoryMethodI::isInline(int *where) {
  if (where) *where = xparams.getItemValueInt("inlinebody");
  return xparams.getItemValueInt("inline");
}

void ClassFactoryMethodI::setVirtual(int v, int pure) {
  xparams.removeItem("virtual");
  xparams.removeItem("pure");
  if (v) xparams.addItem("virtual", StringPrintf("%d", v));
  if (pure) xparams.addItem("pure", StringPrintf("%d", pure));
}

int ClassFactoryMethodI::isVirtual(int *pure) {
  if (pure != NULL)
    *pure = xparams.getItemValueInt("pure");
  return xparams.getItemValueInt("virtual");
}

void ClassFactoryMethodI::addHelpSection(HelpSection *section) {
  m_helpsections.addItem(static_cast<HelpSectionI*>(section));
}

int ClassFactoryMethodI::getNumHelpSections() {
  return m_helpsections.getNumItems();
}

HelpSection *ClassFactoryMethodI::enumHelpSection(int n) {
  return m_helpsections.enumItem(n);
}

void ClassFactoryMethodI::deleteHelpSection(HelpSection *s) {
  m_helpsections.removeItem(static_cast<HelpSectionI*>(s));
  delete static_cast<HelpSectionI*>(s);
}

void ClassFactoryMethodI::setParentBlock(ClassFactoryBlock *block) {
  parentblock = block;
}

ClassFactoryBlock *ClassFactoryMethodI::getParentBlock() {
  return parentblock;
}

int ClassFactoryMethodI::getStartLine(const char *type) {
  if (STRCASEEQL(type, "h")) return ln_hstart;
  else if (STRCASEEQL(type, "cpp")) return ln_cppstart;
  else if (STRCASEEQL(type, "inline")) return ln_inlinestart;
  return -1;
}

int ClassFactoryMethodI::getNumLines(const char *type) {
  if (STRCASEEQL(type, "h")) return ln_hnum;
  else if (STRCASEEQL(type, "cpp")) return ln_cppnum;
  else if (STRCASEEQL(type, "inline")) return ln_inlinenum;
  return -1;
}

void ClassFactoryMethodI::setStartLine(int n, const char *type) {
  if (STRCASEEQL(type, "h")) ln_hstart = n;
  else if (STRCASEEQL(type, "cpp")) ln_cppstart = n;
  else if (STRCASEEQL(type, "inline")) ln_inlinestart = n;
}


void ClassFactoryMethodI::setNumLines(int n, const char *type) {
  if (STRCASEEQL(type, "h")) ln_hnum = n;
  else if (STRCASEEQL(type, "cpp")) ln_cppnum = n;
  else if (STRCASEEQL(type, "inline")) ln_inlinenum = n;
}

int ClassFactoryMethodI::isConstructor() {
  ClassFactoryChunk *chunk = parentblock->getParentChunk();
  return chunk->isConstructor(this);
}

int ClassFactoryMethodI::isDestructor() {
  ClassFactoryChunk *chunk = parentblock->getParentChunk();
  return chunk->isDestructor(this);
}

void ClassFactoryMethodI::setSourceLine(int sl) {
  xparams.replaceItem("line", StringPrintf("%d", sl));
}

int ClassFactoryMethodI::getSourceLine() {
  return ATOI(xparams.getItemValue("line"));
}

