#include "precomp.h"
#include "ClassFactoryParser.h"

#include "../outputfactories/CPPToXHDMerge.h"
#include "../outputfactories/Vfs.h"

#include "ClassFactoryMethodI.h"
#include "ClassFactoryParamI.h"
#include "ClassFactoryDataI.h"
#include "ClassFactoryBlockI.h"
#include "ClassFactoryCodeBlockI.h"
#include "ClassFactoryChunkI.h"

#include <bfc/parse/paramparser.h>

#define BUFLEN 8*1024

CPPParser::CPPParser(const char *s, char sep/* =' ' */) {
  char *p = (char *)s;
  while (*p == ' ') p++;
  while (*p) {
    char *t = CPPToXHDMerge::STRCHR_cpp(p, sep);
    if (!t) {
      m_strings.addItem(new String(p));
      break;
    }
    String s;
    s.catn(p, t-p);
    m_strings.addItem(new String(s));
    p = t+1;
    while (*p == ' ') p++;
  }
}

CPPParser::~CPPParser() {
  m_strings.deleteAll();
}

int CPPParser::getNumItems() {
  return m_strings.getNumItems();
}

const char *CPPParser::enumItem(int n) {
  return m_strings.enumItem(n)->getValue();
}

void ClassFactoryParser::grabWord(String &str, const char *word, int *var, int multiple) {
  if (var) *var = 0;
  String rec;
  const char *p = str.getValue();
  int c = 0;
  while (p && *p) {
    const char *d = (multiple || !c) ? STRSTR(p, word) : NULL;
    if (d) {
      const char *e = d + STRLEN(word);
      if (!(d == str.getValue() || *(d+1) == '\n' || *(d+1) == '\t' || *(d+1) == ' ')) { rec.catn(p, 1); p++; continue; }
      if (*e != '\n' && *e != '\t' && *e != ' ' && *e != 0) { rec.catn(p, 1); p++; continue; }
      rec.catn(p, d-p);
      p+=strlen(word);
      while (*p && (*p == '\n' || *p == '\t' || *p == ' ')) p++;
      if (var) *var = 1;
      c++;
    } else {
      rec.cat(p);
      break;
    }
  }
  str = rec;
}

void ClassFactoryParser::grabInOut(String &str, int *in, int *out, int *inout) {
  grabWord(str, "IN", in);
  grabWord(str, "OUT", out);
  grabWord(str, "INOUT", inout);
}

ClassFactoryChunk *ClassFactoryParser::parse(const char *text, ClassFactoryChunk *chunk) {

  if (text == NULL) return NULL;

  char str[BUFLEN]="";
  char ptype[BUFLEN]="";
  errorstr = "";

  //int p = 0;

  VFILE *vf = Vfopen_mem(text, "rb", STRLEN(text));

  while (!Vfeof(vf)) {
    String paramtype;
    String paramname;
    String _classname;
    String type;
    String method;
    int nodisp = 0;
    int dodisp = 0;
    int dispid = -1;
    int hide = 0;
    String methodkws;
    int scripted = 0;
    int isevent = 0;

    CPPToXHDMerge::Vfgets_cpp(str, BUFLEN, vf);

    str[BUFLEN-1] = 0;
    CRToSpace(str);

    redo:

    formatCPP(str, BUFLEN);
    str[BUFLEN-1] = 0;
    reduceSpaces(str);

    if (isEmpty(str)) continue;

    char *brace = CPPToXHDMerge::STRCHR_cpp(str, '{');
    if (brace) *brace = 0;

    char *rawcol = CPPToXHDMerge::STRCHR_cpp(str, ':');
    char *doublecol = CPPToXHDMerge::STRSTR_cpp(str, "::");
    char *col = CPPToXHDMerge::STRSTR_cpp(str, "):");
    if (col) *(col+1)= 0;
    col = CPPToXHDMerge::STRSTR_cpp(str, ") :");
    if (col) *(col+1)= 0;

    char *sc = CPPToXHDMerge::STRCHR_cpp(str, ';');
    if (sc) *sc = 0;

    // line : type methodname ( param_list ) ;
    // param_list is multiple

    const char *par = CPPToXHDMerge::STRSTR_cpp(str, "(");

    if (par) {
      // look for a second set of parenthesis, if that's the case, look at that as a flag on the function :
      // DISPATCH(0) function(int param);

      char *closepar = CPPToXHDMerge::STRSTR_cpp(par, ")");
      if (closepar) {
        char *secpar = CPPToXHDMerge::STRSTR_cpp(closepar+1, "(");
        if (secpar) {
          // remeber the first part for later
          if (!methodkws.isempty()) methodkws += " ";
          methodkws.catn(str, closepar+1-str);
          String s(closepar+1);
          STRCPY(str, s);
          goto redo;
        }
      }
    }

    if ((par && (rawcol && !col && !doublecol) && (rawcol < par)) || (!par && rawcol && !col && !doublecol)) {
      break;
    }

    if (!par && STRCASEEQL(str, "ctor")) STRCPY(str, StringPrintf("%s();", chunk->getClassName()));
    else if (!par && STRCASEEQL(str, "dtor")) STRCPY(str, StringPrintf("virtual ~%s();", chunk->getClassName()));

    par = CPPToXHDMerge::STRSTR_cpp(str, "(");

    const char *closeopenpar = STRSTR(str, ")(");
    if (!closeopenpar)
      closeopenpar = STRSTR(str, ") (");

    if (par == NULL || closeopenpar) {

      int doabort = 0;

      if (!closeopenpar && sc) { // try for data declaration
        String lastdatatype;
        String lastclass;
        // could be multiple data, ie: int a, b, c, d;
        CPPParser ppv(str, ',');
        const char *_param;
        const char *_type;
        const char *_class;
        int misstatic = 0;
        int misscripted = 0;

        for (int i=0;i<ppv.getNumItems();i++) {
          String initval;
          String decl = ppv.enumItem(i);
          // look for default val
          char *eqchar = CPPToXHDMerge::STRCHR_cpp(str, '=');
          if (eqchar != NULL) {
            char *val = eqchar+1;
            while (val && *val == ' ') val++;
            initval = val;
            decl = "";
            decl.catn(str, eqchar-str);
            while (decl.lastChar() == ' ') { decl.trunc(decl.len()-1); }
            while (decl.getChar(0) == ' ') { decl = String(decl.getValue()+1); }
          }
          // first block contains type
          if (i == 0) {
            ParamParser ppp(decl, " ");
            String reconst;
            for (int ppi=0;ppi<ppp.getNumItems();ppi++) {
              const char *p = ppp.enumItem(ppi);
              if (STRCASEEQL(p, "static")) { misstatic = 1; continue; }
              if (STRCASEEQL(p, "SCRIPT")) { misscripted = 1; continue; }
              if (STREQL(p, "friend")
                  || STREQL(p, "class")
                  || STREQL(p, "using")
                  || STREQL(p, "namespace")) {
                doabort = 1;
                break;
              }
              if (!reconst.isempty()) reconst+=" ";
              reconst+=ppp.enumItem(ppi);
            }
            if (doabort) break;
            decl = reconst;
            if (!parseTypeDecl(decl, &_type, &_param, &_class, "data", str, lastdatatype)) break;
            //if no type on first block, not a data declaration
            if (_type == NULL || !*_type) {
              doabort = 1;
              break;
            }
            paramname = _param;
            paramtype = _type;
            _classname = _class;
          } else {
            paramname = decl;
            paramtype = paramtype;
            _classname = lastclass;
          }
          lastdatatype = paramtype;
          lastclass = _classname;
          ClassFactoryDataI *d = new ClassFactoryDataI(paramname, paramtype);
          d->setStatic(misstatic);
          if (misscripted) d->getParams()->replaceItem("scriptable", StringPrintf("%d", misscripted));
          else d->getParams()->removeItem("scriptable");
          if (!initval.isempty()) {
            d->setInitValue(initval);
          }
          if (!_classname.isempty())
          d->setClassName(_classname);
          chunk->addBlock(new ClassFactoryBlockI(d));
        }

        continue;
      }

      errorstr += "// ???: "; errorstr += str; errorstr += "\n";
      continue;
    }

    char *z = (char *)par+1;
    int gotstuff = 0;
    while (*z) {
      if (*z != ' ' && *z != '\t' && *z != ')') {
        gotstuff = 1;
        break;
      }
      if (*z == ')') break;
      z++;
    }

    ParamParser pp(str, "()");
    if (pp.getNumItems() < 2) {
      if (!CPPToXHDMerge::STRSTR_cpp(str, "()") && !CPPToXHDMerge::STRSTR_cpp(str, "( )")) { // look for parameterless method
        errorstr += "// ???: "; errorstr += str; errorstr += "\n";
        continue; // syntax error on the line, cancel it
      }
    }

    type = "";

    String suffix;
    char *eop = CPPToXHDMerge::STRCHR_cpp(str, ')');
    if (eop != NULL) {
      suffix = eop+1;
      *(eop+1)=0;
    }

    const char *typeandmethod = pp.enumItem(0);
    const char *paramlist = NULL;
    if (gotstuff)
      paramlist = pp.enumItem(1);

    if (!typeandmethod) continue;
    char str2[BUFLEN];
    STRCPY(str2, typeandmethod);
    reduceSpaces(str2);
    ParamParser vp(str2, " ");

    int misvirtual = 0;
    int misstatic = 0;
    int misinline = 0;
    int ispure = 0;

    if (!suffix.isempty()) {
      if (STREQL(suffix, "=0") || STREQL(suffix, " =0") || STREQL(suffix, "= 0") || STREQL(suffix, " = 0"))
        ispure = 1;
    }

    String s;
    int t=0;
    for (int k=0;k<vp.getNumItems();k++) {
      if (STRCASEEQL(vp.enumItem(k), "virtual")) { misvirtual = 1; continue; }
      if (STRCASEEQL(vp.enumItem(k), "static")) { misstatic = 1; continue; }
      if (STRCASEEQL(vp.enumItem(k), "inline")) { misinline = 1; continue; }

      if (t > 0) s+= " ";
      s+=vp.enumItem(k);
      t++;
    }
    STRCPY(str2, s);
    reduceSpaces(str2);

    const char *_type, *_method, *_class;
    if (!parseTypeDecl(str2, &_type, &_method, &_class, "method", str)) continue;

    type = _type;
    method = _method;
    _classname = _class;

    String rc="";
    for (int fx=0;fx<2;fx++) {
      ParamParser pt(fx == 0 ? methodkws : type, " ");
      for (int it=0;it<pt.getNumItems();it++) {
        String its = pt.enumItem(it);
        if (!STRNCMP(its, "DISPATCH", 8)) {
          dodisp = 1;
          char *p = (char *)(its.getValue()+8);
          if (*p == '(') {
            char *d = STRCHR(p, ')');
            if (d != NULL) {
              String sdispid;
              sdispid.catn(p+1, d-p-1);
              dispid = ATOI(sdispid);
            }
          }
          continue;
        }
        if (!STRCMP(its, "NODISPATCH")) {
          nodisp = 1;
          continue;
        }
        if (!STRCMP(its, "HIDE")) {
          hide = 1;
          continue;
        }
        if (!STRCMP(its, "SCRIPT")) {
          scripted = 1;
          continue;
        }
        if (!STRCMP(its, "EVENT")) {
          isevent = 1;
          continue;
        }
        if (!rc.isempty()) rc += " ";
        rc += pt.enumItem(it);
      }
    }

    type = rc;

    if (STRCASEEQL(method, "ctor")) method = StringPrintf("%s();", chunk->getClassName());
    else if (STRCASEEQL(method, "dtor")) method = StringPrintf("virtual ~%s();", chunk->getClassName());

    ClassFactoryMethodI *m = new ClassFactoryMethodI(method, type);
    m->setStatic(misstatic);
    m->setVirtual(misvirtual, ispure);
    m->setInline(misinline, 0);

    if (nodisp && !dodisp) m->getParams()->addItem("nodispatch", "1");
    if (dodisp) m->getParams()->addItem("dispatch", "1");
    if (dispid != -1) m->getParams()->addItem("dispatch_id", StringPrintf("%d", dispid));

    m->getParams()->replaceItem("scriptable", StringPrintf("%d", scripted));
    m->getParams()->replaceItem("isevent", StringPrintf("%d", isevent));
    m->getParams()->replaceItem("hide", StringPrintf("%d", hide));

    if (!_classname.isempty())
      m->setClassName(_classname);

    if (paramlist != NULL) {

      STRCPY(str2, paramlist);
      reduceSpaces(str2);
      ParamParser pp3(str2, ",");
      String autoname;
      int quit = 0;
      for (int i=0;i<pp3.getNumItems() && !quit;i++) {
        const char *param = pp3.enumItem(i);

        // param : type name = defval
        // type is multiple (possibly a ref/ptr, ie: 'const char *a')

        ParamParser pp5(param, " ");
        if (pp5.getNumItems() == 0) {
          delete m;
          break;
        } else if (pp5.getNumItems() == 1) {
          // ie : function(void); or function(int,...);
          const char *kw = pp5.enumItem(0);
          while (kw && *kw == ' ') { kw++; }
          if (STREQL(kw, "void")) {
            continue; // next param, but no param, and so we handle function(void, void) heh
          }
          //if (!STREQL(kw, "...")) {
            // we have a type, so lets give it a name
            //autoname = StringPrintf("%s p%d", kw, i);
            //param = autoname;
          //}
        }

        char str3[BUFLEN];
        STRCPY(str3, param);
        reduceSpaces(str3);
        ParamParser pp4(str3, "=");

        const char *typeandname = pp4.enumItem(0);
        String defval = pp4.enumItem(1); // null if not there
        while (isOneOf(defval.getChar(0), " \t")) defval = String(defval.getValue()+1);
        while (isOneOf(defval.lastChar(), " \t")) defval.trunc(defval.len()-1);

        STRCPY(str3, typeandname);
        reduceSpaces(str3);

        const char *_param;
        const char *_type;
        int inout=0;
        if (!parseTypeDecl(str3, &_type, &_param, NULL, "parameter", str, NULL, &inout)) break;
        paramname = _param;
        paramtype = _type;

        ClassFactoryParamI *p = new ClassFactoryParamI(paramname, paramtype, defval);
        p->setInOut(inout);
        m->addParam(p);

      }
    }
    if (m != NULL) {
      chunk->addBlock(new ClassFactoryBlockI(m));
    }
    if (!errorstr.isempty())
      chunk->addBlock(new ClassFactoryBlockI(new ClassFactoryCodeBlockI(errorstr, "h")), 0);
  }

  Vfclose(vf);

  return chunk;
}

void ClassFactoryParser::release(ClassFactoryChunk *chunk) {
  delete static_cast<ClassFactoryChunkI *>(chunk);
}


int ClassFactoryParser::getLine(const char *text, char *str, int maxlen, int *pos) {
  ASSERT(pos != NULL);
  const char *p = text + *pos;
  char *d = str;
  int n=0;

  if (*p == 0) return 0; // eof

  while (*p && *p != '\n' && *p != '\r' && n < maxlen) {
    *d++ = *p++; (*pos)++; n++;
  }
  *d = 0;

  if (*p && *p != '\n' && *p != '\r' && n == maxlen) { // line too long, cancel it entirely
    *str = 0;
    return 1;
  }

  // we have a line, now eat the remaining cr
  eatCR(p, pos);

  // do some formating to help the parser
  formatCPP(str, maxlen);
  reduceSpaces(str);

  return 1;
}

int ClassFactoryParser::isEmpty(const char *str) {
  const char *p = str;
  while (*p) {
    if (*p != ' ' && *p != '\t') return 0;
    p++;
  }
  return 1;
}

const char *ClassFactoryParser::eatCR(const char *p, int *pos) {
  if (*p == '\n') {
    p++;
    if (pos != NULL) (*pos)++;
    if (*p == '\r') {
      p++;
      if (pos != NULL) (*pos)++;
    }
  } else if (*p == '\r') {
    p++;
    if (pos != NULL) (*pos)++;
    if (*p == '\n') {
      p++; if (pos != NULL) (*pos)++;
    }
  }
  return p;
}

void ClassFactoryParser::trimSpaces(String &_p) {
  const char *p = _p.getValue();
  while (p && isOneOf(*p, " \t\n\r")) STRCPY((char *)p, p+1);
  while (p && STRLEN(p) > 0 && isOneOf(*(p+STRLEN(p)-1), " \t\n\r")) *(char *)(p+STRLEN(p)-1) = 0;
}

void ClassFactoryParser::reduceSpaces(char *_p) {
  char str[BUFLEN]="";
  const char *p = _p;
  char *d = str;
  int count = 1;
  int ddot = 0;
  while (*p) {
    if (*p == ':') ddot++;
    if (*p == ' ' && count != 0) { p++; continue; }
    if (*p != ' ') count = 0;
    else count++;
    *d++ = *p++;
    if (ddot == 2) {
      while (*p && *p == ' ') p++;
      ddot = 0;
    }
  }
  *d = 0; d--;
  while (d >= str) {
    if (*d == ' ')
      *d = 0;
    else
      break;
    d--;
  }
  STRCPY(_p, str);
}

int ClassFactoryParser::parseTypeDecl(const char *txt, const char **t, const char **m, const char **c, const char *what, const char *line, const char *deftype, int *_inout) {
  typedecl_type = "";
  typedecl_name = "";
  typedecl_class = "";

  // first let see if it is a pointer or a ref
  char *s = CPPToXHDMerge::STRCHR_cpp(txt, '*');

  if (s == NULL)
    s = CPPToXHDMerge::STRCHR_cpp(txt, '&');

  if (s != NULL) {
    // if it is, we're going to use the first '*' or '&' as the separation position between type and method
    // skip additional *&
    while (*s && (*s == '*' || *s == '&' || *s == ' ' || *s == '\t')) s++;

    int l = MIN(s-txt, BUFLEN-1);

    char before[BUFLEN];
    STRNCPY(before, txt, l); before[l] = 0;

    char after[BUFLEN];
    STRNCPY(after, s, BUFLEN-1); after[BUFLEN-1] = 0;
    reduceSpaces(after);

    typedecl_name = after;
    typedecl_type = before;
  } else {
    int shift = 0;
    ParamParser pp(txt, " ");
    int n = pp.getNumItems();
    typedecl_name = pp.enumItem(n-1);
    if (n > 1) {
      String beforename = pp.enumItem(n-2);
      if (STRSTR(beforename, "operator")) {
        typedecl_name = beforename + " " + typedecl_name;
        shift = 1;
      }
    }
    if (n == 1) {
      typedecl_type = deftype;
    }
    for (int i=0;i<n-(1+shift);i++) {
      if (!typedecl_type.isempty()) typedecl_type.cat(" ");
      typedecl_type.cat(pp.enumItem(i));
    }
  }

  // ok, it's been separated in two, check that the name is valid
  if (CPPToXHDMerge::STRCHR_cpp(typedecl_name.getValue(), ' ') && !CPPToXHDMerge::STRSTR_cpp(typedecl_name.getValue(), "operator")) { // more than one word ?
    errorstr += "// ???: "; errorstr += line; errorstr += StringPrintf(" [check %s \"%s\"]\n", what, txt);
    return 0; // syntax error on the line, cancel it
  }

  // could be a class::method, parse that
  char *p = NULL;
  char *_p = typedecl_name.getNonConstVal();
  p = CPPToXHDMerge::STRSTR_cpp(typedecl_name, "::");
  if (p != NULL) { // Ok, we have a class::method block
    typedecl_class.trunc(0);
    typedecl_class.catn(_p, p-_p);
    p += 2;
    typedecl_name = String(p);
  }

  fixupType(typedecl_type);

  // grab IN, OUT and INOUT
  int in, out, inout;
  grabInOut(typedecl_type, &in, &out, &inout);
  trimSpaces(typedecl_type);

  if (_inout) {
    *_inout = 0;
    if (in) *_inout |= IO_IN;
    if (out) *_inout |= IO_OUT;
    if (inout) *_inout |= IO_INOUT;
  }

  // set pointers
  if (t != NULL) *t = typedecl_type;
  if (m != NULL) *m = typedecl_name;
  if (c != NULL) *c = typedecl_class;

  return 1;
}

void ClassFactoryParser::formatCPP(const char *s, int maxlen) {
  char *str = (char *)MALLOC(maxlen*3);
  STRCPY(str, s);
  char *p = (char *)str;

  eatComments(str);

  while (*p) {
    if (*p == '\t') *p = ' ';
    p++;
  }

  char *dstr = str;
  while (*dstr == ' ' || *dstr == '\r' || *dstr == '\n') dstr++;
  p = dstr;

  while (*p) {
    if (*p == ' '&& *(p+1) == ' ') {
      STRCPY(p, p+1);
      continue;
    }
    if (p > str && *p == ':' && *(p+1) == ':') {
      if (*(p-1) == ' ') {
        STRCPY(p-1, p);
        continue;
      }
    }
    if (*p == ':' && *(p+1) == ':') {
      if (*(p+2) == ' ') {
        STRCPY(p+2, p+3);
        continue;
      }
    }
    if (p > str && *(p-1) != ' ' && isOneOf(*p, "*&") && !(isOneOf(*(p+1), "*&"))) {
      int l = STRLEN(p);
      memmove(p+1, p, l);
      p[l] = 0;
      *p = ' ';
      continue;
    }
    if (p > str && *p == ':' && *(p-1) != ':' && *(p+1) != ':' && *(p-1) != ' ') {
      int l = STRLEN(p);
      memmove(p+1, p, l);
      p[l+1] = 0;
      *p = ' ';
      continue;
    }
    if (p > str && isOneOf(*p, "<>") && *(p-1) != ' ') {
      int l = STRLEN(p);
      memmove(p+1, p, l);
      p[l+1] = 0;
      *p = ' ';
      continue;
    }
    if (isOneOf(*p, "<>") && *(p+1) != ' ') {
      int l = STRLEN(p);
      memmove(p+2, p+1, l);
      p[l+1] = 0;
      *(p+1) = ' ';
      continue;
    }
    if (*p == ':' && *(p-1) != ':' && *(p+1) != ':' && *(p+1) != ' ') {
      int l = STRLEN(p);
      memmove(p+2, p+1, l);
      p[l+1] = 0;
      *(p+1) = ' ';
      continue;
    }
    if (p > str && isOneOf(*p, ",;") && *(p-1) == ' ') {
      STRCPY(p-1, p);
      continue;
    }
    if (p > str && *(p-1) == '}' && (*p+1) == ';') {
      STRCPY(p, p+1);
      continue;
    }
/*    if (!STRNCMP(p, "//", 2)) {
      while (*p && *p != '\r' && *p != '\n')
        *p++ = ' ';
      if (*p && isOneOf(*p, "\r\n")) *p++ = ' ';
      if (*p && isOneOf(*p, "\r\n")) *p++ = ' ';
      p--;
    }*/
    p++;
  }

  STRNCPY((char *)s, dstr, maxlen);
  ((char *)s)[maxlen-1]=0;
  FREE(str);
}

int ClassFactoryParser::isOneOf(char c, const char *list) {
  const char *p = list;
  while (*p) {
    if (c == *p) return 1;
    p++;
  }
  return 0;
}

int ClassFactoryParser::isIdentifierChar(char c) {
  if (c >= 'a' && c <= 'z') return 1;
  if (c >= 'A' && c <= 'Z') return 1;
  if (c >= '0' && c <= '9') return 1;
  if (c == '_') return 1;
  return 0;
}

void ClassFactoryParser::eatComments(const char *str) {
  char *p = (char *)str;
  while (*p) {
    if (!STRNCMP(p, "//", 2)) {
      while (*p && *p != '\r' && *p != '\n')
        *p++ = ' ';
      continue;
    }
    if (!STRNCMP(p, "/*", 2)) {
      while (*p && STRNCMP(p, "*/", 2))
        *p++ = ' ';
      int i;
      for (i=0;*p && i<2;i++)
        *p++ = ' ';
      continue;
    }
  p++;
  }
}

void ClassFactoryParser::CRToSpace(const char *str) {
  char *p = (char *)str;
  while (*p) {
    if (*p == '\r' || *p == '\n') *p = ' ';
    p++;
  }
}

void ClassFactoryParser::fixupType(String &str) {
  char *p = str.getNonConstVal();
  String s="";
  while (p && *p) {
    s.catn(p, 1);
    if (*p == '*' || *p == '&') {
      p++;
      while (*p && *p == ' ')
        p++;
    } else p++;
  }
  str = s;
}

String ClassFactoryParser::typedecl_name;
String ClassFactoryParser::typedecl_type;
String ClassFactoryParser::typedecl_class;
String ClassFactoryParser::errorstr;
