#include "precomp.h"
#include "ClassScanner.h"
#include <bfc/parse/paramparser.h>
#include "classhierarchy/ClassFactoryParser.h"

#ifndef MAXBUF
#define MAXBUF 8192
#endif

ClassScanner::ClassScanner(const char *basefilename, const char *path) {
  m_basefilename = basefilename;
  m_path = path;
  scan();
};

ClassScanner::ClassScanner() {
}

ClassScanner::~ClassScanner() {
  m_classes.deleteAll();
};

void ClassScanner::setFiles(const char *basefilename, const char *path) {
  m_path = path;
  m_basefilename = basefilename;
}

int ClassScanner::getNumClasses() {
  return m_classes.getNumItems();
}

const char *ClassScanner::enumClass(int n) {
  return m_classes.enumItem(n)->getValue();
}

int ClassScanner::scan() {
  int r = 0;
  m_classes.deleteAll();
  if (!Std::isDirChar(m_path.lastChar())) m_path.cat(DIRCHARSTR);
  for (int z=0;z<2;z++) {
    String flname = StringPrintf("%s%s.%s", m_path.getValue(), m_basefilename.getValue(), z == 0 ? "h" : "cpp");
    FILE *hf = FOPEN(flname, "rt");
    if (hf != NULL) {
      r++;
      char line[MAXBUF];
      while (!feof(hf)) {
        FGETS(line, MAXBUF, hf);
        line[MAXBUF-1] = 0;
        ClassFactoryParser::reduceSpaces(line);
        ClassFactoryParser::formatCPP(line, MAXBUF);
        char *p = line+STRLEN(line)-1;
        while (p > line) {
          if (*p != '\r' && *p != '\n') break;
          p--;
        }
        p++; *p = 0;
        ParamParser pp(line, " ");
        int nw = pp.getNumItems();
        if (nw == 0) continue;
        if (STRCASEEQL(pp.enumItem(0), "class") && nw > 1) {
          String *classname = new String(pp.enumItem(1));
          int ininheritance = 0;
          String lastkeyword;
          if (nw > 1) {
            int i = 1;
            while (i < nw) {
              if (STRCASEEQL(pp.enumItem(i), ":")) {
                *classname = lastkeyword;
                ininheritance = 1;
              } else if (STRCASEEQL(pp.enumItem(i), "{")) {
                *classname = lastkeyword;
                break;
              } else if (String(pp.enumItem(i)).lastChar() == '{') {
                *classname = lastkeyword;
                classname->trunc(classname->len()-1);
                break;
              } else if (STRCASEEQL(pp.enumItem(i), ";") || String(pp.enumItem(i)).lastChar() == ';') {
                delete classname;
                classname = NULL;
                break;
              } else if (STRCASEEQL(pp.enumItem(i), ",")) {
              } else if (STRCASEEQL(pp.enumItem(i), "public")) {
              } else if (STRCASEEQL(pp.enumItem(i), "private")) {
              } else if (STRCASEEQL(pp.enumItem(i), ".")) {
              } else {
                if (ininheritance) {
                  // read ancestor
                } else {
                  // got a classname or a keyword
                  lastkeyword = pp.enumItem(i);
                }
              }
              i++;
            }
            if (classname) {
              if (!classname->isempty()) {
                *classname = lastkeyword;
              }
              if (classname->lastChar() == '\n')
                classname->trunc(classname->len()-1);
              if (classname->lastChar() == '\r')
                classname->trunc(classname->len()-1);
              if (!classname->isempty())
                m_classes.addItem(classname);
            }
          }
        }
      }
      FCLOSE(hf);
    }
  }
  return r;
}
