#include "precomp.h"
#include "HelperEnum.h"
#include "ClassFactoryChunk.h"
#include <bfc/parse/paramparser.h>
#include <api/xml/xmlparams.h>
#include "../outputfactories/svc_srcClassFactory.h"

HelperEnum::HelperEnum() {
  m_chunk = NULL;
  makeList();
}

HelperEnum::HelperEnum(ClassFactoryChunk *chunk) {
  setChunk(chunk);
}

HelperEnum::~HelperEnum() {
  m_helpers.deleteAll();
}

void HelperEnum::setChunk(ClassFactoryChunk *chunk) {
  m_chunk = chunk;
  makeList();
}

int HelperEnum::getNumHelpers() {
  return m_helpers.getNumItems();
}

const char *HelperEnum::enumHelper(int n) {
  return m_helpers.enumItem(n)->getValue();
}

int HelperEnum::findHelper(const char *helper) {
  foreach(m_helpers)
    if (m_helpers.getfor()->iscaseequal(helper)) return foreach_index;
  endfor;
  return -1;
}

void HelperEnum::makeList() {
  m_helpers.deleteAll();
  if (m_chunk != NULL) {
    String factories = m_chunk->getParams()->getItemValue("xhd_factories");
    ParamParser pp(factories, ";", 1);
    if (pp.hasString("Interface")) {
      m_helpers.addItem(new String("Interface"));
    }
    for (int i=0;i<pp.getNumItems();i++) {
      String hname = pp.enumItem(i);
      if (STRCASEEQL(hname, "Interface")) continue;
      if (STRCASEEQL(hname, "Implementation")) continue;
      m_helpers.addItem(new String(hname));
    }
  } else {
    m_helpers.addItem(new String("Interface"));
    SrcClassFactoryEnum enf;
    svc_srcClassFactory *svc = enf.getFirst();
    while (svc != NULL) {
      for (int i=0;i<svc->getNumAvailableFormats();i++) {
        String name = svc->enumAvailableFormat(i);
        if (!STRCASEEQL(name, "Interface")
         && !STRCASEEQL(name, "Implementation")) {
          m_helpers.addItem(new String(name));
        }
      }
      WASABI_API_SVC->service_release(svc);
      svc = enf.getNext();
    }
  }
}

HelperDescription::HelperDescription(const char *format) {
  SrcClassFactoryEnum en(format);
  svc_srcClassFactory *svc = en.getFirst();
  if (svc != NULL) {
    // make the helper process its template
    svc->setChunk(NULL);
    m_desc = svc->getDescription();
    WASABI_API_SVC->service_release(svc);
  }
}

