#include "precomp.h"
#include "HelperFileEnum.h"
#include "ClassFactoryChunk.h"
#include <bfc/parse/paramparser.h>
#include "../outputfactories/svc_srcClassFactory.h"
#include <api/xml/xmlparams.h>

HelperFileEnum::HelperFileEnum() {
  m_chunk = NULL;
  m_num = -1;
}

HelperFileEnum::HelperFileEnum(ClassFactoryChunk *chunk, int helpernum) {
  setHelper(chunk, helpernum);
}

HelperFileEnum::HelperFileEnum(ClassFactoryChunk *chunk, const char *helpername) {
  String factories = chunk->getParams()->getItemValue("xhd_factories");
  ParamParser pp(factories, ";", 1);
  int n = 0;
  m_num = -1;
  if (pp.hasString("Interface")) {
    n++;
    if (STRCASEEQL(helpername, "Interface"))
      m_num = 0;
  }
  if (m_num == -1) {
    for (int i=0;i<pp.getNumItems();i++) {
      if (STRCASEEQL(pp.enumItem(i), "Interface")) continue;
      if (STRCASEEQL(helpername, pp.enumItem(i))) {
        m_num = n;
        break;
      }
      n++;
    }
  }
  if (m_num == -1) return;
  setHelper(chunk, m_num);
}

HelperFileEnum::~HelperFileEnum() {
  m_files.deleteAll();
}

void HelperFileEnum::setHelper(ClassFactoryChunk *chunk, int helpernum) {
  m_chunk = chunk;
  m_num = helpernum;
  makeList();
}

int HelperFileEnum::getNumFiles() {
  return m_files.getNumItems();
}

const char *HelperFileEnum::enumFile(int n) {
  m_retstring = m_files.enumItem(n);
  return m_retstring;
}

void HelperFileEnum::makeList() {
  ASSERT(m_chunk != NULL);
  ASSERT(m_num != -1);
  m_files.deleteAll();

  String prevanc = m_chunk->getAncestor();

  String factories = m_chunk->getParams()->getItemValue("xhd_factories");
  ParamParser pp(factories, ";", 1);
  m_chunk->setAncestor(NULL);
  int n = 0;
  if (pp.hasString("Interface")) {
    if (m_num == n) {
      m_name = "Interface";
      SrcClassFactoryEnum scfe(m_name);
      svc_srcClassFactory *fact = scfe.getFirst();
      if (fact != NULL) {
        fact->setChunk(m_chunk);
        for (int j=0;j<fact->getNumModules();j++) {
          m_files.addItem(new String(fact->enumModuleName(j)));
        }
        WASABI_API_SVC->service_release(fact);
      }
    }
    n++;
  }
  for (int i=0;i<pp.getNumItems();i++) {
    if (STRCASEEQL(pp.enumItem(i), "Interface")) continue;
    if (STRCASEEQL(pp.enumItem(i), "Implementation")) continue;
    if (m_num == n) {
      m_name = pp.enumItem(i);
      SrcClassFactoryEnum scfe(m_name);
      svc_srcClassFactory *fact = scfe.getFirst();
      if (fact != NULL) {
        String prevanc = m_chunk->getAncestor();
        m_chunk->setAncestor(NULL);
        fact->setChunk(m_chunk);
        for (int j=0;j<fact->getNumModules();j++) {
          m_files.addItem(new String(fact->enumModuleName(j)));
        }
        WASABI_API_SVC->service_release(fact);
        m_chunk->setAncestor(prevanc);
      }
    }
    n++;
  }
  m_chunk->setAncestor(prevanc);
}
