#include "precomp.h"
#include "XHDProcessor.h"
#include <api/xml/xmlparams.h>
#include <bfc/parse/paramparser.h>
#include "ClassFactoryChunk.h"
#include "../outputfactories/svc_srcClassFactory.h"
#include "../outputfactories/ImplementationFactory.h"
#include "../outputfactories/OutputFactoryBase.h"
#include "../classhierarchy/Project.h"

extern "C" unsigned long int _crc32(const char *buffer);

XHDProcessor::XHDProcessor(int full, ClassFactoryChunk *chunk, const char *stopfactory, int forcestartcpp, int forcestarth, ProgressCallback *percent) {
  m_full = full;
  m_workspace = g_workspace;
  m_ppercent = percent;
  m_fact = NULL;
  m_chunk = chunk;

  m_prevancestor = chunk->getAncestor();

  chunk->clearAllFlags();
  chunk->setAncestor(NULL);

  if (chunk->haveCodeBlocks("cpp") && !forcestartcpp) chunk->setFlag(CF_FLAG_HAS_CPP_CODEBLOCKS);
  if (chunk->haveCodeBlocks("h") && !forcestarth) chunk->setFlag(CF_FLAG_HAS_H_CODEBLOCKS);

  int stopped = 0;

  // enumerate all factories in chunk
  const char *xfactories = chunk->getParams()->getItemValue("xhd_factories");
  ParamParser pp(xfactories, ";", 1);
  int f, nf = pp.getNumItems();
  for (f=0;f<nf;f++) {
    if (full && m_ppercent) m_ppercent->setPercent((int)(((float)f / (float)nf) * 100));
    String factory = pp.enumItem(f);
    if (factory.iscaseequal("Implementation")) continue; // Implementation must be last

    // find the factory
    SrcClassFactoryEnum e(factory);
    svc_srcClassFactory *cf = e.getFirst();
    // set the chunk and process so that everybody does what he needs on the chunk
    if (cf != NULL) {
      cf->setWorkspace(m_workspace);
      cf->setChunk(chunk);
      if (full)
        cf->process();
      ProcessorEntry *pe = new ProcessorEntry();
      pe->m_svc = cf;
      pe->m_needfriend = cf->needFriend();
      pe->m_classname = chunk->getAncestor();
      pe->m_format = factory;
      XmlReaderParams *par = chunk->getParams();
#ifdef ODE_IDE
      par->removeItem(StringPrintf("%s_n", cf->getName()));
      par->addItem(StringPrintf("%s_n", cf->getName()), StringPrintf("%d", cf->getNumModules()));
      for (int ms=0;ms<cf->getNumModules();ms++) {
        par->removeItem(StringPrintf("%s_%d", cf->getName(), ms));
        par->addItem(StringPrintf("%s_%d", cf->getName(), ms), cf->enumModuleName(ms));
      }
#endif
      for (int ff = 0; ff < cf->getNumModules(); ff++)
        pe->m_files.addItem(new String(cf->enumModuleName(ff)));
      m_layers.addItem(pe);
      if (stopfactory && factory.iscaseequal(stopfactory)) {
        stopped = 1;
        break;
      }
    }
  }

  if (!stopped) {
    m_fact = new ImplementationFactory();
    m_fact->setChunk(chunk);
    if (full) {
      m_fact->setProgressCallback(m_ppercent);
      m_fact->setWorkspace(m_workspace);
      m_fact->process();
    }

    ProcessorEntry *pe = new ProcessorEntry();
    pe->m_svc = m_fact;
    pe->m_releaseservice = 0;
    pe->m_format = "Implementation";
    pe->m_classname = chunk->getAncestor();
    for (int ff = 0; ff < m_fact->getNumModules(); ff++)
      pe->m_files.addItem(new String(m_fact->enumModuleName(ff)));
    m_layers.addItem(pe);

    /*
    // write temp files

    char tp[1024];
    char tf[1024];

    GetTempPath(1024, tp);
    GetTempFileName(tp, "cf", 0, tf);
    FILE *fh = FOPEN(tf, "wb");
    const char *str = fact->enumModuleData(0);
    FWRITE(str, STRLEN(str), 1, fh);
    FCLOSE(fh);
    unsigned long hcrc = crc32file(tf);
    UNLINK(tf);

    GetTempPath(1024, tp);
    GetTempFileName(tp, "cf", 0, tf);
    FILE *fcpp = FOPEN(tf, "wb");
    str = fact->enumModuleData(1);
    FWRITE(str, STRLEN(str), 1, fcpp);
    FCLOSE(fcpp);
    unsigned long cppcrc = crc32file(tf);
    UNLINK(tf);
    */

    if (full) {
      unsigned long hcrc = _crc32(m_fact->enumModuleData(0));
      unsigned long cppcrc = _crc32(m_fact->enumModuleData(1));

      chunk->getParams()->removeItem("hcrc");
      chunk->getParams()->removeItem("cppcrc");
      chunk->getParams()->addItem("hcrc", StringPrintf("%d", hcrc));
      chunk->getParams()->addItem("cppcrc", StringPrintf("%d", cppcrc));
    }
  }

}

XHDProcessor::~XHDProcessor() {
  m_layers.deleteAll();
  delete m_fact;
  if (!m_full)
    m_chunk->setAncestor(m_prevancestor);
}

svc_srcClassFactory *XHDProcessor::getFactory(const char *name) {
  foreach(m_layers)
    if (STRCASEEQLSAFE(name, m_layers.getfor()->m_format)) return m_layers.getfor()->m_svc;
  endfor;
  return NULL;
}

ProcessorEntry *XHDProcessor::getFactoryEntry(const char *name) {
  foreach(m_layers)
    if (STRCASEEQLSAFE(name, m_layers.getfor()->m_format)) return m_layers.getfor();
  endfor;
  return NULL;
}

int XHDProcessor::getNumFactories() {
  return m_layers.getNumItems();
}

svc_srcClassFactory *XHDProcessor::enumFactory(int n) {
  return m_layers.enumItem(n)->m_svc;
}

ProcessorEntry *XHDProcessor::enumProcessorEntry(int n) {
  return m_layers.enumItem(n);
}
