#include "precomp.h"
#include "XMLToChunk.h"
#include <bfc/string/stringdict.h>
#include <bfc/pair.h>
#include <bfc/string/url.h>

#include "../classhierarchy/ClassFactorySubChunkI.h"
#include "../classhierarchy/ClassFactoryDataI.h"
#include "../classhierarchy/Workspace.h"
#include "../classhierarchy/Project.h"

int g_noVar = 0;

enum XHD_TAGS {
  XHD_CLASS,
  XHD_XHD,
  XHD_METHOD,
  XHD_DATA,
  XHD_PARAM,
  XHD_DEFINE,
  XHD_UNDEF,
  XHD_IFDEF,
  XHD_ELSE,
  XHD_CPP,
  XHD_HELPSECTION,
  XHD_HELPELEMENT,
};

BEGIN_STRINGDICTIONARY(_XHDTAGS);
SDI("class", XHD_CLASS);
SDI("xhd", XHD_XHD);
SDI("method", XHD_METHOD);
SDI("data", XHD_DATA);
SDI("param", XHD_PARAM);
SDI("define", XHD_DEFINE);
SDI("undef", XHD_UNDEF);
SDI("ifdef", XHD_IFDEF);
SDI("else", XHD_ELSE);
SDI("cpp", XHD_CPP);
SDI("helpsection", XHD_HELPSECTION);
SDI("helpelement", XHD_HELPELEMENT);
END_STRINGDICTIONARY(_XHDTAGS, xhdTags);

XMLToChunk::XMLToChunk() {
  curChunk = NULL;
  curMethod = NULL;
  curCodeBlock = NULL;
  ifdefedout = 0;
  cbcount = 0;
  header = 1;
  footer = 0;
  lastChunk = NULL;
  nextChunk = NULL;
  curData = NULL;
}

XMLToChunk::XMLToChunk(const char *file) {
  curChunk = NULL;
  curMethod = NULL;
  curCodeBlock = NULL;
  ifdefedout = 0;
  cbcount = 0;
  header = 1;
  footer = 0;
  lastChunk = NULL;
  nextChunk = NULL;
  curData = NULL;
  XmlReader::registerCallback("*", static_cast<XmlReaderCallback*>(this));
  XmlReader::loadFile(file, NULL, static_cast<XmlReaderCallback*>(this));
  XmlReader::unregisterCallback(static_cast<XmlReaderCallback*>(this));
}


XMLToChunk::~XMLToChunk() {
  chunks.deleteAll();
  //vars.deleteAll();
}

ClassFactoryChunk *XMLToChunk::enumChunk(int n) {
  return chunks.enumItem(n);
}

int XMLToChunk::getNumChunks() {
  return chunks.getNumItems();
}

void XMLToChunk::xmlReaderOnStartElementCallback(const char *xmlpath, const char *xmltag, XmlReaderParams *params) {
  int tag = xhdTags.getId(xmltag);
  if (ifdefedout && tag != XHD_IFDEF && tag != XHD_ELSE) return;
  XmlReaderParamsI processed_params;
  for (int i=0; i<params->getNbItems(); i++) {
    String name = params->getItemName(i);
    String value = processVariables(params->getItemValue(i));
    processed_params.addItem(name, value);
  }
  //params = &processed_params;
  switch (tag) {
    case XHD_CLASS: {
      header = 0;
      String cn = params->getItemValue("name");
      Url::decode(cn);
      if (curChunk == NULL) { // new chunk
        if (nextChunk != NULL) {
          curChunk = nextChunk;
          nextChunk = NULL;
          curChunk->setClassName(cn);
        } else curChunk = new ClassFactoryChunkI(cn);
        //OutputDebugString(StringPrintf("xhd>class %s\n", cn));
        String impl = params->getItemValue("implementation");
        String ance = params->getItemValue("ancestor");
        Url::decode(impl);
        Url::decode(ance);
        curChunk->setImplementationName(impl);
        curChunk->setParams(&processed_params);
        curChunk->setRawParams(params);
        lastChunk = curChunk;
      }
      break;
    }
#ifdef ODE_IDE
    case XHD_XHD: {
      if (!g_workspace) break;
      if (curChunk == NULL) break;
      XmlReaderParams *par = g_workspace->getParams();
      String outpath = par ? par->getItemValue("target_dir") : "";
      if (!Std::isDirChar(outpath[0]) && (outpath[0] != 0 && outpath[1] != ':')) {
        String s = g_workspace->getDirectory();
        if (!Std::isDirChar(s.lastChar())) s.cat(Std::dirCharStr());
        outpath = StringPrintf("%s%s", s, outpath);
      }
      if (Std::isDirChar(outpath.lastChar())) outpath.trunc(outpath.len()-1);

      String tfile = params->getItemValue("file");
      if (Std::isDirChar(tfile.getChar(0))) tfile = String(tfile.getValue()+1);
      IdeXHD *add = NULL;
      String destfilename = StringPrintf("%s%s%s", outpath, DIRCHARSTR, tfile);
      add = new IdeXHD(destfilename, g_workspace->getWorkspaceName());
      if (add->enumChunk(0) == NULL) delete add;
      else {
        g_workspace->parserAddSubXHD(add);
      }
      ClassFactoryChunk *lchunk = add->enumChunk(0);
      lchunk->setParentChunk(curChunk);
      curChunk->addBlock(new ClassFactoryBlockI(new ClassFactorySubChunkI(lchunk)));
      break;
    }
#endif
    case XHD_METHOD:
      if (curMethod == NULL) { // new method
        String mn = params->getItemValue("name");
        Url::decode(mn);
        //OutputDebugString(StringPrintf("xhd> method %s\n", mn));
        curMethod = new ClassFactoryMethodI(mn);
        String mt = params->getItemValue("type");
        Url::decode(mt);
        curMethod->setReturnType(mt);
        String def = params->getItemValue("default");
        Url::decode(def);
        curMethod->setDefaultReturnValue(def);
        curMethod->setParams(params);
      }
      break;
    case XHD_DATA:
      if (curData == NULL) { // new method
        String dn = params->getItemValue("name");
        Url::decode(dn);
        //OutputDebugString(StringPrintf("xhd> data %s\n", mn));
        String dt = params->getItemValue("type");
        Url::decode(dt);
        curData = new ClassFactoryDataI(dn, dt);
        curData->setDataType(dt);
        curData->setParams(params);
      }
      break;
    case XHD_PARAM:
      if (curMethod != NULL) {
        String pn = params->getItemValue("name");
        String pt = params->getItemValue("type");
        String pd = params->getItemValue("default");
        Url::decode(pn);
        Url::decode(pt);
        Url::decode(pd);
        ClassFactoryParamI *par = new ClassFactoryParamI(pn, pt, pd);
        par->setParams(params);
        //OutputDebugString(StringPrintf("xhd>  param %s\n", pn));
        curParam = par;
      }
      break;
    case XHD_DEFINE: {
      String dn = params->getItemValue("name");
      String dv = params->getItemValue("value");
      Url::decode(dn);
      Url::decode(dv);
      g_globalVars->setString(dn, dv);
      //vars.addItem(new XMLVariable(dn, dv));
      break;
    }
    case XHD_UNDEF: {
      int p=-1;
      String un = params->getItemValue("name");
      Url::decode(un);
      //XMLVariable *v = vars.findItem(un, &p);
      //if (v != NULL) {
      //  vars.removeByPos(p);
      //  delete v;
      //}
      g_globalVars->removeString(un);
    }
    break;
    case XHD_IFDEF: {
      ifdefstack.push(ifdefedout);
      if (!ifdefedout) {
        String in = params->getItemValue("name");
        Url::decode(in);
        //XMLVariable *v = vars.findItem(in);
        //if (v == NULL) {
        if (g_globalVars->haveString(in)) {
          ifdefedout=1;
        }
      }
    }
    break;
    case XHD_ELSE:
      ifdefedout = !ifdefedout;
      break;
    case XHD_CPP:
      if (curCodeBlock != NULL) {
        cbcount++;
        printf("Embedded code blocks, ignoring tag pair.\n");
      } else {
        curCodeBlock = new ClassFactoryCodeBlockI();
        curCodeBlock->setParams(params);
      }
      break;
    case XHD_HELPSECTION:
      curHelpSection = new HelpSectionI();
      if (curParam) {
        // parameter help section
      } else if (curMethod) {
        // method help section
      } else if (curChunk) {
        // class help section
      } else {
        // oops!
      }
      break;
  }
}

void XMLToChunk::xmlReaderOnEndElementCallback(const char *xmlpath, const char *xmltag) {
  int tag = xhdTags.getId(xmltag);
  if (ifdefedout && tag != XHD_IFDEF && tag != XHD_ELSE) return;
  switch (tag) {
    case XHD_CLASS:
      footer = 1;
      chunks.addItem(curChunk);
      curChunk = NULL;
      break;
    case XHD_METHOD:
      curChunk->addBlock(new ClassFactoryBlockI(curMethod));
      curMethod = NULL;
      break;
    case XHD_DATA:
      curChunk->addBlock(new ClassFactoryBlockI(curData));
      curData = NULL;
      break;
    case XHD_PARAM:
      curMethod->addParam(curParam);
      curParam = NULL;
      break;
    case XHD_ELSE:
      ifdefedout = !ifdefedout;
      break;
    case XHD_IFDEF:
      ifdefstack.pop(&ifdefedout);
      break;
    case XHD_HELPSECTION:
      if (curParam) {
        // parameter help section
        curParam->addHelpSection(curHelpSection);
      } else if (curMethod) {
        // method help section
        curMethod->addHelpSection(curHelpSection);
      } else if (curChunk) {
        // class help section
        curChunk->addHelpSection(curHelpSection);
      } else {
        // oops!
        delete curHelpSection;
      }
      curHelpSection = NULL;
      break;
    case XHD_HELPELEMENT:
      if (curHelpSection) {
        curHelpSection->addElement(curHelpElement);
      } else {
        delete curHelpElement;
      }
      curHelpElement = NULL;
      break;
    case XHD_CPP:
      if (cbcount == 0) {
        if (curMethod != NULL)
          if (STRCASEEQLSAFE(curCodeBlock->getType(), "hcpp")) {
            if (curMethod != NULL) curMethod->setHeader(curCodeBlock->getCodeBlock(), "cpp");
            delete curCodeBlock;
          } else if (STRCASEEQLSAFE(curCodeBlock->getType(), "hh")) {
            if (curMethod != NULL) curMethod->setHeader(curCodeBlock->getCodeBlock(), "h");
            delete curCodeBlock;
          } else if (STRCASEEQLSAFE(curCodeBlock->getType(), "hinline")) {
            if (curMethod != NULL) curMethod->setHeader(curCodeBlock->getCodeBlock(), "inline");
            delete curCodeBlock;
          } else if (STRCASEEQLSAFE(curCodeBlock->getType(), "finline")) {
            if (curMethod != NULL) curMethod->setFooter(curCodeBlock->getCodeBlock(), "inline");
            delete curCodeBlock;
          } else if (STRCASEEQLSAFE(curCodeBlock->getType(), "fcpp")) {
            if (curMethod != NULL) curMethod->setFooter(curCodeBlock->getCodeBlock(), "cpp");
            delete curCodeBlock;
          } else if (STRCASEEQLSAFE(curCodeBlock->getType(), "fh")) {
            if (curMethod != NULL) curMethod->setFooter(curCodeBlock->getCodeBlock(), "h");
            delete curCodeBlock;
          } else {
            curMethod->setCodeBlock(curCodeBlock);
          }
        else {
          ClassFactoryChunkI *chunk = curChunk;
          const char *type = curCodeBlock->getType();
          if (STRCASEEQLSAFE(type, "class")
              || STRCASEEQLSAFE(type, "cpp")
              || STRCASEEQLSAFE(type, "h")) {
            if (chunk == NULL && !footer) {
              if (nextChunk == NULL)
                nextChunk = new ClassFactoryChunkI();
              chunk = nextChunk;
            }
            if (header)
              chunk->setHeader(curCodeBlock);
            else if (footer) {
              ASSERT(lastChunk != NULL);
              lastChunk->setFooter(curCodeBlock);
            } else {
              if (STRCASEEQLSAFE(curCodeBlock->getType(), "class")) {
                chunk->setHeader(curCodeBlock);
              } else {
                chunk->addBlock(new ClassFactoryBlockI(curCodeBlock));
              }
            }
          } else {
            chunk->addBlock(new ClassFactoryBlockI(curCodeBlock));
          }
        }
        curCodeBlock = NULL;
      } else
        cbcount--;
      break;
  }
}

void XMLToChunk::xmlReaderOnCharacterDataCallback(const char *xmlpath, const char *xmltag, const char *str) {
  int tag = xhdTags.getId(xmltag);
  if (ifdefedout) return;
  switch (tag) {
    case XHD_HELPELEMENT:
      if (curHelpElement) {
        curHelpElement->setHtml(str);
      }
      break;
    case XHD_CPP:
      if (curCodeBlock == NULL) break;
      curCodeBlock->setCodeBlock(str, curCodeBlock->getType());
      break;
  }
}

const char *XMLToChunk::processVariables(const char *str) {
  varbuf[0] = 0;
  if (g_noVar) return str;

  char *p = (char *)str;
  char *d = varbuf;
  while (*p && d < varbuf+BUFMAX-256) {
    if (*p == '$') {
      // walk til next separator
      char *pz = p+1;
      while (*pz) {
        if (!((*pz >= 'A' && *pz <= 'Z') || (*pz >= 'a' && *pz<= 'z') || (*pz >= '0' && *pz <= '9') || *pz == '_'))
          break;
        pz++;
      }
      // otherwise, we have a word between p ($) and pz (next char), now get it
      char kw[256];
      STRNCPY(kw, p+1, pz-p-1); kw[pz-p-1] = 0;
      // and replace it with the variable value
      //XMLVariable *v = vars.findItem(kw);
      GlobalString *v = g_globalVars->getStringObject(kw);
      if (v != NULL) {
        STRCPY(d, v->getValue());
        d += STRLEN(d);
      } else {
        STRNCPY(d, p, pz-p);
        d += pz-p;
        *d = 0;
      }
      p = pz;
    } else {
      // continue walkin
      *d++ = *p++;
    }
  }
  *d = 0;
  varbuf[BUFMAX-256] = 0;

  return varbuf;
}

void XMLToChunk::deleteChunk(ClassFactoryChunk *chunk) {
  ClassFactoryChunkI *c = static_cast<ClassFactoryChunkI *>(chunk);
  if (chunks.haveItem(c)) {
    chunks.removeItem(c);
    delete c;
  }
}
