#include "precomp.h"
#include "CPPToXHDMerge.h"
#include "../classhierarchy/Project.h"
#include "../classhierarchy/Workspace.h"
#include "svc_srcClassFactory.h"
#include <bfc/parse/paramparser.h>

#include "../classhierarchy/ClassFactoryMethodI.h"
#include "../classhierarchy/ClassFactoryCodeBlockI.h"
#include "../classhierarchy/ClassFactoryDataI.h"
#include "../classhierarchy/ClassFactoryBlockI.h"
#include "../classhierarchy/ClassFactoryChunkI.h"
#include "../classhierarchy/ClassFactoryParser.h"
#include "../classhierarchy/XHDProcessor.h"

#define MAXBUF 8 * 1024

CPPToXHDMerge::CPPToXHDMerge(ClassFactoryChunk *_chunk, const char *path, int import, const char *_importorigclassname, const char *_importimplclassname, ProgressCallback *percent, Workspace *workspace) {
  m_workspace = workspace;
  m_ppercent = percent;
  isimport = import;
  chunk = _chunk;
  importorigclassname = _importorigclassname;
  importimplclassname = _importimplclassname;
  mergecpp = StringPrintf("%s%s", path, _chunk->getParams()->getItemValue("class_module"));
  mergeh = StringPrintf("%s%s", path, _chunk->getParams()->getItemValue("class_header"));
  mergefiles = 1;
  merge();
}

CPPToXHDMerge::CPPToXHDMerge(ClassFactoryChunk *_chunk, const char *_cppdata, const char *_hdata, int import, const char *_importorigclassname, const char *_importimplclassname, ProgressCallback *percent, Workspace *workspace) {
  //ASSERT(_cppdata && _hdata);
  m_workspace = workspace;
  m_ppercent = percent;
  isimport = import;
  chunk = _chunk;
  mergecpp = _cppdata;
  mergeh = _hdata;
  mergefiles = 0;
  importimplclassname = _importimplclassname;
  importorigclassname = _importorigclassname;
  merge();
}

void CPPToXHDMerge::setIsImport(int import) {
  isimport = import;
}


void CPPToXHDMerge::Vfgets_cpp(char *dest, int n, VFILE *fl, const char *search, int *scopemodifier, int *_bracecount, int *_bracketcount, int *linecount) {
  char *d = dest;
  int lastscope = SCOPE_NEUTRAL;
  int l=0;
  int inquotes = 0;
  int incquotes = 0;
  int escape = 0;
  int incomment = 0;
  int inlinecomment = 0;
  int inblockcomment = 0;
  int ispreproc = 0;
  int gotachar = 0;
  String lastword;
  String prevword;
  String curword;
  char t[2]="X";
  int gotscope = 0;
  int gotfriend = 0;
  int gotclass = 0;
  int endatnextcr = 0;
  int parcount = 0;
  int bracecount = 0;
  int bracketcount = 0;
  if (scopemodifier != NULL) lastscope = *scopemodifier;
  if (_bracecount) bracecount = *_bracecount;
  if (_bracketcount) bracketcount = *_bracketcount;
  int gotminus = 0;

  while (!Vfeof(fl) && (dest == NULL || l<n-2)) {
    char c = Vfgetc(fl);
    if (dest != NULL) *d++ = c; l++; t[0] = c;

    int pos = Vftell(fl);
    char nextc = Vfgetc(fl);
    Vfseek(fl, pos, SEEK_SET);

    if (search != NULL &&
        (parcount == 0 || (parcount == 1 && c == ')')) &&
        (bracketcount == 0 || (bracketcount == 1 && (c == '>' && !lastword.iscaseequal("operator") && !gotminus))) &&
        !incomment &&
        !inquotes &&
        !incquotes &&
        !escape &&
        c == *search &&
        (*(search+1)==0 || *(search+1) == nextc)
        ) {
      // ok we have 2 chars right, enough to check for more
      int pos = Vftell(fl); // record pos
      Vfseek(fl, -1, SEEK_CUR);
      char *g = (char *)search;
      int cont=0;
      while (*g) {
        char t = Vfgetc(fl);
        if (t != *g++) {
          cont = 1;
          break;
        }
      }
      Vfseek(fl, pos, SEEK_SET);
      if (cont) continue;
      Vfseek(fl, -1, SEEK_CUR);
      break;
    }

    int isidentchar = ClassFactoryParser::isIdentifierChar(c);
    if (!inquotes && !incquotes && !incomment && !isidentchar) {
      if (!lastword.isempty())
        prevword = lastword;
      lastword = curword;
      curword = "";

      if (!lastword.isempty() && STREQL(lastword, "class"))
        gotclass = 1;

      if (!gotclass && !lastword.isempty()) {
        if (STREQL(lastword, "public")) {
          gotscope = 1;
          lastscope = SCOPE_PUBLIC;
        } else if (STREQL(lastword, "private")) {
          gotscope = 1;
          lastscope = SCOPE_PRIVATE;
        } else if (STREQL(lastword, "protected")) {
          gotscope = 1;
          lastscope = SCOPE_PROTECTED;
        } else if (STREQL(lastword, "friend")) {
          gotfriend = 1;
          endatnextcr = 1;
        }
      }
    } else {
      curword.cat(t);
    }

    if (gotminus && c != ' ') {
      gotminus = 0;
      if (c == '>') {
        continue;
      }
    }
    if (!gotachar && c == '#') {
      ispreproc = 1;
      continue;
    }
    if (!inquotes && !incquotes && !incomment && c == '/' && nextc == '/') {
      if (dest != NULL) { *d++ = Vfgetc(fl); l++; }
      inlinecomment = 1;
      inblockcomment = 0;
      incomment = 1;
      continue;
    }
    if (!inquotes && !incquotes && !incomment && c == '/' && nextc == '*' ) {
      if (dest != NULL) { *d++ = Vfgetc(fl); l++; }
      incomment = 1;
      inblockcomment = 1;
      inlinecomment = 0;
      continue;
    }
    if (!gotachar && !incomment && c != ' ' && c != '\t' && c != '\r' && c != '\n') {
      gotachar = 1;
      gotscope = 0; // reset scope kw
    }
    if (!inquotes && !incquotes && incomment && !inlinecomment && c == '*' && nextc == '/') {
      if (dest != NULL) { *d++ = Vfgetc(fl); l++; }
      inblockcomment = 0;
      incomment = 0;
      continue;
    }
    if (!inquotes && !incquotes && inlinecomment && (c == '\r' || c == '\n')) {
      if (linecount) (*linecount)++;
      if ((nextc == '\r' || nextc == '\n') && c != nextc) {
        if (dest != NULL) { *d++ = Vfgetc(fl); l++; }
      }
      inlinecomment = 0;
      incomment = 0;
      break;
    }
    if (!incomment && escape == 2) {
      escape = 0;
      continue;
    }
    if (!incomment && escape == 1) {
      escape = 0;
      continue;
    }
    if (!incomment && c == '\\') {
      if (!escape) escape = 1; else escape = 0;
      continue;
    }
    if (!incomment && !escape && c == '\"') {
      inquotes = !inquotes;
      continue;
    }
    if (!inquotes && !incomment && !escape && c == '\'') {
      incquotes = !incquotes;
      continue;
    }

    if (!inquotes && !incomment && !escape && c == '{') {
      bracecount++;
    }

    if (!inquotes && !incomment && !escape && c == '}') {
      bracecount--;
    }

    if (!inquotes && !incomment && !escape && c == '(') {
      parcount++;
      continue;
    }

    if (!inquotes && !incomment && !escape && c == ')') {
      parcount--;
      continue;
    }

    if (!inquotes && !incomment && !escape && c == '-') {
      gotminus = 1;
      continue;
    }

    if (!inquotes && !incomment && !escape && c == '<' && !prevword.iscaseequal("operator")) {
      bracketcount++;
      continue;
    }

    if (!inquotes && !incomment && !escape && c == '>' && !prevword.iscaseequal("operator")) {
      bracketcount--;
      continue;
    }

    if (!incomment && !inquotes && !incquotes && c == '}' && nextc == ';') {
      endatnextcr = 1;
      continue;
    }
    if (!incomment && !inquotes && !incquotes && parcount == 0 && (c == ';' || c == '{' || c == '}' || c == '<' || c == '>' || (!gotclass && gotscope && c == ':'))) {
      // end of statement;
      endatnextcr = 1;
      continue;
    }
    if (!inblockcomment && (ispreproc || !gotachar || endatnextcr) && (c == '\r' || c == '\n')) {
      if (linecount) (*linecount)++;
      if ((nextc == '\r' || nextc == '\n') && c != nextc) {
        if (dest != NULL) { *d++ = Vfgetc(fl); l++; }
      }
      break;
    }
    if (c == '\r' || c == '\n') {
      if (linecount) (*linecount)++;
      if ((nextc == '\r' || nextc == '\n') && c != nextc) {
        if (dest != NULL) { *d++ = Vfgetc(fl); l++; }
      }
    }
  }
  if (dest != NULL) *d = 0;
  if (scopemodifier) *scopemodifier = lastscope;
  if (_bracecount) *_bracecount = bracecount;
  if (_bracketcount) *_bracketcount = bracketcount;

//  OutputDebugString(dest);
//  OutputDebugString(">>\n");
}

char *CPPToXHDMerge::STRSTR_cpp(const char *str, const char *srch) {
  VFILE *vf = Vfopen_mem(str, "rb", STRLEN(str));
  char *r = NULL;
  while (1) {
    Vfgets_cpp(NULL, 0, vf, srch);
    if (Vfeof(vf)) {
      Vfclose(vf);
      return NULL;
    }
    r = (char *)(str + Vftell(vf));
    if (STRNICMP(r, srch, STRLEN(srch))) r = NULL;
    else break;
  }
  Vfclose(vf);
  return r;
}

char *CPPToXHDMerge::STRCHR_cpp(const char *str, char c) {
  char t[2]="X";
  *t = c;
  return STRSTR_cpp(str, t);
}

void CPPToXHDMerge::parseScope(const char *str, int *modifier) {
  VFILE *vf = Vfopen_mem(str, "rb", STRLEN(str));
  while (!Vfeof(vf))
    Vfgets_cpp(NULL, 0, vf, NULL, modifier);
  Vfclose(vf);
}

void CPPToXHDMerge::updateScope(const char *str, int *curscope) {
  int modifier = SCOPE_NEUTRAL;
  VFILE *vf = Vfopen_mem(str, "rb", STRLEN(str));
  while (!Vfeof(vf))
    Vfgets_cpp(NULL, 0, vf, NULL, &modifier);
  Vfclose(vf);
  if (modifier != SCOPE_NEUTRAL)
    *curscope = modifier;
}

int CPPToXHDMerge::countBraces(const char *str) {
  if (str == NULL) return 0;
  int bc = 0;
  VFILE *vf = Vfopen_mem(str, "rb", STRLEN(str));
  Vfgets_cpp(NULL, 0, vf, NULL, NULL, &bc);
  Vfclose(vf);
  return bc;
}

void CPPToXHDMerge::merge() {

  ClassFactoryChunkI *cp = NULL;
  int didstuff = 0;

  PtrList<ClassFactoryMethodI> wrongparamsmathods;
  PtrList<String> friends;
  PtrList<Ancestor> ancestors;

  XHDProcessor preproc(0, chunk);

  int curpass = 0;
  for (int z=1;z>=0;z--) {
    int curscope = SCOPE_PRIVATE;

    const char *merge = NULL;

    merge = z == 0 ? mergecpp : mergeh;

    int totalpass = 0;

    if (!mergecpp.isempty() && !mergeh.isempty()) totalpass = 2;
    else if (!mergecpp.isempty() || !mergeh.isempty()) totalpass = 1;

    String thisclass = chunk->getImplementationName();

    if (merge == NULL || *merge == 0)
      continue;

    VFILE *inf = NULL;
    if (mergefiles)
      inf = Vfopen(merge, "rb", TRUE);
    else
      inf = Vfopen_mem(merge, "rb", STRLEN(merge));

    if (inf == NULL) {
      // do not output an error -- conflicts with idexhd when double clicking on a file item that doesn't exist yet, we still want the xhd to generate the files so we shouldn't fail the merge
      //chunk->addErrorMessage(StringPrintf("Could not open %s", merge));
      continue;
    }

    Vfseek(inf, 0, SEEK_END);
    int size = Vftell(inf);
    Vfseek(inf, 0, SEEK_SET);

    didstuff = 1;
    int isinclass = 0;
    int classcount = 0;
    int gbracecount = 0;

    ASSERT(chunk != NULL);
    const char *blocktypes = z==0?"cpp":"h";
    const char *blocktypes2 = z==0?"cpp":"class";
    chunk->dropCodeBlocks(blocktypes, 1);
    if (z == 1) chunk->dropCodeBlocks(blocktypes2, 1);

    if (z == 1) chunk->dropCodeBlocks("interface", 0);
    if (z == 1) chunk->dropCodeBlocks("interface.header.h", 0);
    if (z == 0) chunk->dropCodeBlocks("interface.header.cpp", 0);
    if (z == 1) chunk->dropCodeBlocks("interface.footer.h", 0);
    if (z == 0) chunk->dropCodeBlocks("interface.footer.cpp", 0);

    char line[MAXBUF];
    char origline[MAXBUF];
    int linecount = 0;
    ClassFactoryBlock *lastblock = NULL;

    isinclass = 0;
    while (!Vfeof(inf)) {
      int eatline = 0;
      if (m_ppercent) m_ppercent->setPercent((int)(((curpass*(1.0f/(float)totalpass)) + ((float)Vftell(inf) / (float)size) / totalpass) * 100));
      Sleep(0);
      *line = 0;
      nextline:
      lastimport = "";
      int scopemodifier = SCOPE_NEUTRAL;
      Vfgets_cpp(line, MAXBUF-1, inf, NULL, &scopemodifier, &gbracecount, NULL, &linecount);
      if (scopemodifier != SCOPE_NEUTRAL)
        curscope = scopemodifier;
      line[MAXBUF-1] = 0;
//      mainline:
      STRCPY(origline, line); // save unmodified string
      ClassFactoryParser::reduceSpaces(line);
      if (startsWith(line, "#") || startsWith(line, "//")) {
        dumbcode += origline;
        continue;
      }
      if (startsWith(line, "/*[") && STRCHR(line, ']')) {
        dumbcode += origline;
        String blocktype;
        char *p = STRCHR(line, '[');
        ASSERT(p != NULL);
        char *d = STRCHR(line, ']');
        ASSERT(d != NULL);
        p++;
        blocktype.catn(p, d-p);
        ClassFactoryCodeBlockI *block = new ClassFactoryCodeBlockI();

        d++;
        String blockdata;
        char *z = STRSTR(line, "*/");
        if (z != NULL) {
          blockdata.catn(d, z-d);
        } else {
          blockdata.cat(d);
          while (!Vfeof(inf)) {
            Vfgets_cpp(line, MAXBUF-1, inf, NULL, NULL, NULL, NULL, &linecount);
            dumbcode += line;
            d = STRSTR(line, "*/");
            if (d != NULL) {
              blockdata.catn(line, d-line);
              break;
            }
            blockdata.cat(line);
          }
        }
        block->setCodeBlock(blockdata, blocktype);
        chunk->addBlock(new ClassFactoryBlockI(block));
        continue;
      }
      ClassFactoryParser::formatCPP(line, MAXBUF);
      if (*line) {
        ParamParser pp(line, " ,()");
        if (isinclass) {
          if (!STRNCMP(line, "};", 2) || (pp.getNumItems() >= 2 && STREQL(pp.enumItem(0), "}") && STREQL(pp.enumItem(1), ";"))) {
            if (isinclass && classcount == 0) {
              char *sc = STRCHR_cpp(origline, ';');
              String remaining = sc+1;
              if (!remaining.isempty()) {
                chunk->setFooter(new ClassFactoryCodeBlockI(remaining, blocktypes2));
              }
              if (!dumbcode.isempty()) {
                char *p = dumbcode.getNonConstVal();
/*                while (p && *p) {
                  if (*p != '\r' && *p != '\n') break;
                  p++;
                }*/
                if (p && *p) {
                  if (!chunk->getHeader(blocktypes2))
                    chunk->setHeader(new ClassFactoryCodeBlockI(dumbcode, blocktypes2)); // that doesn't seem right
                  else
                    chunk->addBlock(new ClassFactoryBlockI(new ClassFactoryCodeBlockI(dumbcode, blocktypes)));
                }
                dumbcode.reset();
              }
              isinclass = 0;
              continue;
            } else if (classcount > 0) {
              classcount--;
            } else {
              chunk->addErrorMessage("unbalanced brackets", linecount);
            }
          }
        }
        if (classcount > 0) {
          dumbcode += origline;
          continue;
        }
        if (z == 0) { // if cpp, we only need to look for "::"
          if (!STRSTR_cpp(line, "::")) {
            dumbcode += origline;
            continue;
          }
        } else if (z == 1) { // but if h, we need to look a bit deeper
          int gotclass=0;
          int gotname=0;
          int gotcolumn=0;
          int hadcolumn=0;
          int lastscope=SCOPE_PRIVATE;
          int forwdecl=0;
          String templatestr;
          int tpltcount = 0;
          int intemplate = 0;
          String localdumbcode;
          String classkw;
          String ourancestor = StringPrintf("%sX", chunk->getClassName());
          char *here = origline-1;
          String thisw;
          char *fb = NULL;
          char *parsepos = origline;
          for (int g=0;g<pp.getNumItems()&&fb==NULL;g++) {
            thisw = pp.enumItem(g);
            while (thisw.getChar(0) == '\r' || thisw.getChar(0) == '\n') thisw = String(thisw.getValue()+1);
            while (thisw.len() > 0 && thisw.lastChar() == '\r' || thisw.lastChar() == '\n') thisw.trunc(thisw.len()-1);
            if (thisw.isempty()) continue;
            char *parsepos = g == 0 ? origline : here;
            here++;
            char *next = STRSTR(here, thisw);
            int _l = STRLEN(thisw)-1;
            while (next == NULL && _l > 0) {
              String srch;
              srch.catn(thisw, _l);
              next = STRSTR(here, srch);
              _l--;
            }
            if (next == NULL) {
              // er what to do ? just go on, it'll duplicate a word but will catch up at the next one
            } else
              here = next;
            if (thisw.getChar(0) == '{') break;
            //char *fb = STRCHR_cpp(thisw, '{');
            //if (fb != NULL) *fb = 0;
            if (!gotclass && !intemplate && STREQL(thisw, "class")) {
              // grab what was before "class" and add it to the dumb block
              ASSERT(here != NULL);
              localdumbcode.catn(parsepos, here-parsepos);
              gotclass = 1;
              continue;
            } else if (!gotclass) {
              ASSERT(here != NULL);
              localdumbcode.catn(parsepos, here-parsepos);
            }
            if (!gotclass) {
              if (STREQL(thisw, "template")) {
                if (STRCHR(line, '<'))
                  intemplate = 1;
                templatestr += thisw;
                templatestr += " ";
                continue;
              }
            }
            if (STREQL(thisw, "<")) {
              tpltcount++;
              if (!gotclass) {
                templatestr += thisw;
                templatestr += " ";
                continue;
              }
            }
            if (!gotclass && STREQL(thisw, ">")) {
              tpltcount--;
              if (!gotclass) {
                templatestr += thisw;
                templatestr += " ";
                if (tpltcount == 0) intemplate = 0;
                continue;
              }
              continue;
            }
            if (intemplate) {
              if (!gotclass) {
                templatestr += thisw;
                templatestr += " ";
              }
              continue;
            }
            if (gotclass) {
              if (STREQL(thisw, ":")) {
                gotcolumn = 1;
                continue;
              }
              if (!gotname) {
                char *fc = STRCHR_cpp(thisw, ':');
                if (fc != NULL) {
                  *fc = 0;
                  hadcolumn = 1;
                }
              }
            }
            char *fsc = STRCHR_cpp(thisw, ';');
            if (fsc != NULL) {
              forwdecl = 1;
              *fsc = 0;
              break;
            }
            char *fcr = STRCHR_cpp(thisw, '\r');
            if (!fcr) fcr = STRCHR_cpp(thisw, '\n');
            if (fcr != NULL) {
              // got a classname[CRLF]
              *fcr = 0;
            }
            if (gotclass && !gotname && !gotcolumn) {
             if (!isimport && STREQL(thisw, thisclass) ||
                (isimport && STREQL(thisw, importimplclassname))) {
                  gotname = 1;
                  if (hadcolumn)
                    gotcolumn = 1;
                  hadcolumn = 0;
                  continue;
              }
            }
            if (tpltcount == 0 && gotclass && thisw[STRLEN(thisw)-2]==';') { forwdecl = 1; break; }
            if (tpltcount == 0 && gotname && STREQL(thisw, "public")) { lastscope = SCOPE_PUBLIC; continue; }
            if (tpltcount == 0 && gotname && STREQL(thisw, "private")) { lastscope = SCOPE_PRIVATE; continue; }
            if (tpltcount == 0 && gotname && STREQL(thisw, "protected")) { lastscope = SCOPE_PROTECTED; continue; }
            if (gotname && !STREQL(thisw, ourancestor)) {
              if (!STREQLSAFE(thisw, chunk->getAncestor()) && !STREQLSAFE(thisw, chunk->getClassName())) {
                if (tpltcount == 0)
                  ancestors.addItem(new Ancestor(thisw, lastscope));
                else {
                  String *s = &ancestors.getLast()->classname;
                  if (STRCHR(s->getValue(), '<') && !ClassFactoryParser::isOneOf(s->lastChar(), "<>") && !STREQL(thisw, ">")) {
                    s->cat(", ");
                  } else {
                    if (ClassFactoryParser::isOneOf(s->lastChar(), "<>") && (STREQL(thisw, ">") || STREQL(thisw, "<")))
                      s->cat(" ");
                  }
                  s->cat(thisw);
                }
              }
            }
            if (!gotcolumn && tpltcount == 0) {
              if (!classkw.isempty()) classkw += " ";
              classkw += thisw;
            }
            lastscope = SCOPE_PRIVATE;
          }
          if (isinclass && gotclass && !gotname && !forwdecl) { // we got a class in a class, but it's not the one we manage, and it wasn't a forward declaration either
            classcount++;
          }
          if (gotname) { // class OurClass : public OurOtherclass, public Blah, Blop {
            if (!classkw.isempty())
              chunk->setClassKW(classkw);
            isinclass = 1;
            String remaining;
            if ((thisw != NULL && *thisw == '{') || fb != NULL) {
              char *p = STRCHR_cpp(origline, '{'); p++;
              remaining = p;
            }
            dumbcode += localdumbcode;
            if (!dumbcode.isempty()) {
              char *p = dumbcode.getNonConstVal();
              if (p && *p) {
                chunk->setHeader(new ClassFactoryCodeBlockI(dumbcode, "h"));
              }
            }
            dumbcode = remaining;
            continue;
          }
        }
        if (isinclass) {
          char *p = STRSTR_cpp(line, "friend");
          if (p != NULL) {
            ParamParser fpp(p, " ");
            String thefriend = fpp.getLastItem();
            while (ClassFactoryParser::isOneOf(thefriend.lastChar(), " \t\r\n;")) thefriend.trunc(thefriend.len()-1);
            int cln = preproc.getNumFactories();
            int icl = 0;
            int continueloop = 0;
            for (icl = 0; icl<cln && !continueloop; icl++) {
              ProcessorEntry *pe = preproc.enumProcessorEntry(icl);
              if (STRCASEEQLSAFE(pe->m_classname, thefriend)) {
                char *s = STRSTR_cpp(origline, "friend");
                String linestart;
                linestart.catn(origline, s-origline);
                while (ClassFactoryParser::isOneOf(linestart.lastChar(), " \t\r\n")) linestart.trunc(linestart.len()-1);
                dumbcode += linestart;
                char *d = STRCHR_cpp(s, ';');
                if (d) d++;
                String lineend = d;
                while (ClassFactoryParser::isOneOf(lineend.getChar(0), " \t\r\n")) lineend = String(lineend.getValue()+1);
                dumbcode += lineend;
                continueloop=1;
                break;
              }
            }
            if (continueloop) continue;
            friends.addItem(new String(thefriend));
            // let it got thru
          }
        }
        reparsemember:
        int matchedbyname = 0;
        int isexternalinline = 0;
        delete cp;
        cp = new ClassFactoryChunkI(chunk->getClassName());
        ClassFactoryParser::parse(line, cp);
        if (cp != NULL) {
          if (cp->getNumBlocks() > 0) {
            for (int xb=0;xb<cp->getNumBlocks();xb++) {
              ClassFactoryBlock *cb = cp->enumBlock(xb);
              if (cb->getBlockType() == CF_BLOCKTYPE_METHOD) {
                ClassFactoryMethod *cm = cb->getMethod();
                String methodname = cm->getName(); // Method or Class::Method
                String classname = cm->getClassName();
                {
                  ParamParser pp(cm->getReturnType(), " ");
                  if (STRCASEEQLSAFE(pp.enumItem(0), "friend")) continue;
                }
                int recognized = 0;
                if (!isinclass) {
                  if (!isimport && classname.isequal(thisclass)) {
                    // body belongs to implementation, nothing to do
                    recognized = 1;
                  } else if (isimport) {
                    if (!importimplclassname.isempty() && importimplclassname.isequal(classname)) {
                      recognized = 1;
                    }
                  }
                }
                if (isinclass || recognized) { // Ok, we have a class::method block
                  // now, see if it is a match with an XHD entry
                  int skipsearch = 0; // set to 1 to abort looking for the method
                  for (int i_chunk=0;i_chunk<chunk->getNumBlocks() && !skipsearch;i_chunk++) {
                    int doesntmatch = 0;
                    ClassFactoryBlock *b = chunk->enumBlock(i_chunk);
                    if (b->getBlockType() == CF_BLOCKTYPE_METHOD) {
                      ClassFactoryMethod *m = b->getMethod();
                      String xhdmethodname = m->getName();
                      int constructor = chunk->isConstructor(m);
                      int destructor = chunk->isDestructor(m);
                      if (((!constructor && !destructor && STRCASEEQL(m->getName(), methodname)) ||
                          (constructor && STRCASEEQL(isimport ? importimplclassname : thisclass, methodname)) ||
                          (destructor && STRCASEEQL(isimport ? importimplclassname : thisclass, methodname.getValue()+1)))
                      ) {
                        matchedbyname++;
                        // method name matches, now check parameters
                        int cmnp = cm->getNumParams();
                        int mnp = m->getNumParams();
                        if (cmnp != mnp) doesntmatch = 1;
                        for (int j=0;j<cmnp && doesntmatch==0;j++) {
                          ClassFactoryParam *cp = cm->enumParam(j);
                          ClassFactoryParam *p = m->enumParam(j);
                          if (p == NULL) { doesntmatch=1; continue; } // not matching number of params
                          if (!STRCASEEQLSAFE(p->getType(), cp->getType())) { doesntmatch = 1; continue; } // param type does not match
                        }
                        if (!doesntmatch) {
                          if (z == 1 && lastblock != b && isinclass) {
                            int newpos = 0;
                            if (lastblock != NULL) {
                              newpos = chunk->getBlockNum(lastblock)+1;
                              for (int lm=newpos;lm<chunk->getNumBlocks();lm++) {
                                if (chunk->enumBlock(lm)->getBlockType() == CF_BLOCKTYPE_METHOD) break;
                                if (chunk->enumBlock(lm)->getBlockType() == CF_BLOCKTYPE_DATA) break;
                                newpos++;
                              }
                            }
                            int oldpos = chunk->getBlockNum(b);
                            if (newpos != oldpos) {
                              chunk->moveBlock(b, newpos);
                              oldpos = chunk->getBlockNum(b);
                              if (oldpos <= i_chunk) {
                                i_chunk--;
                              }
                            }
                            lastblock = b;
                          }
                          m->setFlag(z == 0 ? CF_FLAG_METHOD_FOUND_IN_CPP : CF_FLAG_METHOD_FOUND_IN_H);
                          m->setSourceLine(linecount);
                          if (z == 1) m->setScope(curscope);
                          // ok, seems like we do have a match, so, grab the inside of the method and insert it in the corresponding XHD method
                          BigString codeblock;
                          //todo: find end of declaration and grab method header ( : ident(blah) )
                          char *eop = findEndOfPar(origline);
                          if (eop) {
                            eop++;
                            // see if there is a method modifier with params

                            char *col = STRCHR_cpp(eop, ':');
                            char *seceop = findEndOfPar(eop);
                            char *p = STRCHR_cpp(eop, '{');

                            if (seceop != NULL && (col == NULL || col > seceop) && (p == NULL || p > seceop)) eop = seceop+1;

                            p = STRCHR_cpp(eop, '{');
                            char *s = STRCHR_cpp(eop, ';');

                            int blockcount = gbracecount;

                            if (p != NULL) {
                              int n = countBraces(p);
                              blockcount -= n;
                            }

                            if (blockcount > 0 && !(blockcount == 1 && isinclass)) {
                              dumbcode += origline;
                              continue;
                            }

                            // update method flags
                            if (z == 1 && isinclass) {
                              m->setStatic(cm->isStatic());
                              m->setInline(cm->isInline(), 0);
                              int pure=0;
                              int virt = cm->isVirtual(&pure);
                              m->setVirtual(virt, pure);
                              m->setReturnType(cm->getReturnType());
                              m->setDefaultReturnValue(cm->getDefaultReturnValue());

                              // update inouts

                              for (int j=0;j<cm->getNumParams();j++) {
                                ClassFactoryParam *cp = cm->enumParam(j);
                                ClassFactoryParam *p = m->enumParam(j);
                                p->setInOut(cp->getInOut());
                              }

                              // _if_ we got a new dispatch_id, record it, otherwise keep the previous one wether or not we're dispatched

                              if (cm->getParams()->getItemValueInt("dispatch_id",-9999999)!=-9999999) {
                                if (cm->getParams()->getItemValueInt("dispatch_id") == -1)
                                  m->getParams()->removeItem("dispatch_id");
                                else
                                  m->getParams()->replaceItem("dispatch_id", cm->getParams()->getItemValue("dispatch_id"));
                              }

                              // set default flags -- for now all off
                              m->getParams()->removeItem("nodispatch");
                              m->getParams()->removeItem("dispatch");
                              m->getParams()->removeItem("scriptable");
                              m->getParams()->removeItem("isevent");
                              m->getParams()->removeItem("hide");

                              // update with parsed flags

                              if (cm->getParams()->getItemValueInt("nodispatch",-1)!=-1) {
                                if (cm->getParams()->getItemValueInt("nodispatch",0))
                                  m->getParams()->addItem("nodispatch", cm->getParams()->getItemValue("nodispatch"));
                              }

                              if (cm->getParams()->getItemValueInt("dispatch",-1)!=-1) {
                                if (cm->getParams()->getItemValueInt("dispatch",0))
                                  m->getParams()->addItem("dispatch", cm->getParams()->getItemValue("dispatch"));
                              }

                              if (cm->getParams()->getItemValueInt("scriptable",-1)!=-1) {
                                if (cm->getParams()->getItemValueInt("scriptable",0))
                                  m->getParams()->addItem("scriptable", cm->getParams()->getItemValue("scriptable"));
                              }

                              if (cm->getParams()->getItemValueInt("isevent",-1)!=-1) {
                                if (cm->getParams()->getItemValueInt("isevent",0))
                                  m->getParams()->addItem("isevent", cm->getParams()->getItemValue("isevent"));
                              }

                              if (cm->getParams()->getItemValueInt("hide",-1)!=-1) {
                                if (cm->getParams()->getItemValueInt("hide",0))
                                  m->getParams()->addItem("hide", cm->getParams()->getItemValue("hide"));
                              }

                            } else if (z == 1) {
                              isexternalinline = 1;
                              m->setInline(cm->isInline(), 1);
                            }

                            String mheader;
                            if (p == NULL && s != NULL) { // no body !
                              String postdecl;
                              postdecl.catn(eop, s-eop);
                              m->setPostDeclModifier(postdecl);
                              if (!dumbcode.isempty()) {
                                char *p = dumbcode.getNonConstVal();
    /*                            while (p && *p) {
                                  if (*p != '\r' && *p != '\n') break;
                                  p++;
                                }*/
                                if (p && *p) {
                                  if (!isexternalinline && !chunk->getHeader(blocktypes2))
                                    chunk->setHeader(new ClassFactoryCodeBlockI(dumbcode, blocktypes2));
                                  else {
                                    int position = chunk->getBlockNum(b);
                                    chunk->addBlock(new ClassFactoryBlockI(new ClassFactoryCodeBlockI(dumbcode, isexternalinline ? "inline" : blocktypes)), position);
                                  }
                                }
                                dumbcode.reset();
                                // gotos sometimes are so convenient ...
                              }
                              if (*s && *(s+1)) {
                                m->setFooter(s+1, isexternalinline ? "inline" : blocktypes);
                              }
                              eatline = 1;
                              skipsearch = 1;
                              continue;
                            }
    //                        int braceonline = 0;
                            if (p == NULL) { // { wasn't present on the method line, so read until we find it
                              mheader = eop;
                              while (1) {
                                char c = Vfgetc(inf);
                                if (c == '{' || c == ';') break;
                                if (Vfeof(inf)) break;
                              }
                              *line = 0;
                              Vfgets(line, MAXBUF-1, inf); linecount++;
                              line[MAXBUF-1] = 0;
                              STRCPY(origline, line);
                              ClassFactoryParser::reduceSpaces(line);
                            } else { // otherwise we need to record whatever is behind it until next line
                              gbracecount--;
                              mheader.catn(eop, p-eop);
                              codeblock = ""; // so that an empty datablock gets inserted
                              p++;
                              String segline = p;
    /*                          while (*p && (*p != '\n' && *p != '\r')) {
                                segline.catn(p, 1);
                                p++;
                              }
                              while (*p && (*p == '\n' || *p == '\r')) {
                                p++;
                              }*/
                              if (segline.isempty()) *line = 0;
                              else { STRNCPY(line, segline, MAXBUF-1); line[MAXBUF-1] = 0; }
                              //ClassFactoryParser::reduceSpaces(line);
    /*                          if (!segline.isempty())
                                braceonline = 1;*/
                            }
                            int bracecount = 1;
                            int inquotes = 0;
                            int incquotes = 0;
                            int escape = 0;
                            int incomment = 0;
                            int inlinecomment = 0;
                            while (!skipsearch) {
                              // we need to count the number of { and }, we are past the fist
                              char *d = line;
                              while (*d && !skipsearch) {
                                if (!inquotes && !incquotes && !incomment && !STRNCMP(d, "//", 2)) {
                                  inlinecomment = 1;
                                  incomment = 1;
                                }
                                if (!inquotes && !incquotes && !incomment && !STRNCMP(d, "/*", 2)) {
                                  incomment = 1;
                                  inlinecomment = 0;
                                }
                                if (!inquotes && !incquotes && incomment && !inlinecomment && !STRNCMP(d, "*/", 2)) {
                                  incomment = 0;
                                }
                                if (!inquotes && !incquotes && inlinecomment && (*d == '\r' || *d == '\n')) {
                                  inlinecomment = 0;
                                  incomment = 0;
                                }
                                if (!incomment && escape == 2)
                                  escape = 0;
                                else if (!incomment && escape == 1)
                                  escape++;
                                if (!incomment && *d == '\\') {
                                  if (escape < 2) escape++;
                                } else if (!incomment && !escape && *d == '\"') {
                                    inquotes = !inquotes;
                                } else if (!inquotes && !incomment && !escape && *d == '\'') {
                                    incquotes = !incquotes;
                                }
                                if (!incomment && !inquotes && !incquotes) {
                                  if (*d == '{') {
                                    bracecount++;
                                  } else if (*d == '}') {
                                    bracecount--;
                                    if (bracecount == 0) { // end of method
                                      char *p = codeblock.getNonConstVal();
    /*                                  while (p && *p) {
                                        if (*p != '\r' && *p != '\n') break;
                                        p++;
                                      }*/
                                      if (p) {
                                        m->setCodeBlock(new ClassFactoryCodeBlockI(codeblock, isexternalinline ? "inline" : blocktypes));
    //                                    m->getParams()->removeItem("onlinebrace");
    //                                    m->getParams()->addItem("onlinebrace", StringPrintf("%d", braceonline));
                                      }
                                      if (!mheader.isempty()) {
                                        m->setHeader(mheader, isexternalinline ? "inline" : blocktypes);
                                      }
                                      if (!dumbcode.isempty()) {
                                        char *p = dumbcode.getNonConstVal();
    /*                                    while (p && *p) {
                                          if (*p != '\r' && *p != '\n') break;
                                          p++;
                                        }*/
                                        if (p && *p) {
                                          if (!isexternalinline && !chunk->getHeader(blocktypes2)) {
                                            chunk->setHeader(new ClassFactoryCodeBlockI(dumbcode, blocktypes2));
                                          }
                                          else {
                                            int position = chunk->getBlockNum(b);
                                            chunk->addBlock(new ClassFactoryBlockI(new ClassFactoryCodeBlockI(dumbcode, isexternalinline ? "inline" : blocktypes)), position);
                                          }
                                        }
                                        dumbcode.reset();
                                      }
                                      d++;
                                      int n = STRLEN(d);
    //                                  if (n >= 1 && (*d == '\r' || *d == '\n')) d++; n--;
    //                                  if (n >= 1 && (*d == '\r' || *d == '\n')) d++; n--;
                                      //STRCPY(line, d);
                                      // and now that we've successfully grabbed the function, we need to modify the parameter names to whatever names are actually used in the implementation, or it's not gonna compile
                                      int cmnp = cm->getNumParams();
                                      for (int j=0;j<cmnp;j++) {
                                        ClassFactoryParam *cp = cm->enumParam(j);
                                        ClassFactoryParam *p = m->enumParam(j);
                                        p->setName(cp->getName());
                                      }

                                      // now we need to assign what was behind the } as a footer for the member
                                      if (d && *d) {
                                        m->setFooter(d, isexternalinline ? "inline" : blocktypes);
                                      }

                                      eatline = 1;
                                      skipsearch = 1;
                                      continue;
                                    }
                                  }
                                } // if !inquotes
                                codeblock.catn(d, 1);
                                d++;
                              }
                              if (skipsearch) continue;
                              if (*d == 0) d = NULL;

                              incodenextline:

                              // get next line
                              if (Vfeof(inf)) {
                                chunk->addErrorMessage("End of file before end of method body, cannot merge method.");
                                break;
                              }
                              *line = 0;
                              Vfgets(line, MAXBUF-1, inf); linecount++;
                              line[MAXBUF-1] = 0;
                              if (startsWith(line, "#") || startsWith(line, "//")) {
                                codeblock += line;
                                goto incodenextline;
                              }
                            }
                          }
                        }
                      }
                    }
                  }
                  if (!skipsearch && /*(isimport || (matchedbyname && z == 1)) && */!STRCASEEQLSAFE(origline, lastimport)) {
                    if (z == 1 && isinclass) {
  //                    cm->getParams()->addItem("implementation", "1");
                      ClassFactoryMethodI *dm = new ClassFactoryMethodI();
                      if (matchedbyname && z == 1) wrongparamsmathods.addItem(dm);
                      dm->copy(cm);
                      dm->setSourceLine(linecount);
                      int isctor = 0;
                      int isdtor = 0;
                      if (STRCASEEQLSAFE(dm->getName(), importimplclassname)) {
                        dm->setName(chunk->getClassName());
                        isctor = 1;
                      }
                      if (STRCASEEQLSAFE(dm->getName(), StringPrintf("~%s", importimplclassname.getValue()))) {
                        dm->setName(StringPrintf("~%s", chunk->getClassName()));
                        isdtor = 1;
                      }
                      int newpos = 0;
                      if (lastblock != NULL) {
                        newpos = chunk->getBlockNum(lastblock)+1;
                        for (int lm=newpos;lm<chunk->getNumBlocks();lm++) {
                          if (chunk->enumBlock(lm)->getBlockType() == CF_BLOCKTYPE_METHOD) break;
                          if (chunk->enumBlock(lm)->getBlockType() == CF_BLOCKTYPE_DATA) break;
                          newpos++;
                        }
                      }
                      ClassFactoryBlockI *bl = new ClassFactoryBlockI(dm);
                      chunk->addBlock(bl, newpos);
                      lastblock = bl;
                      delete cp; cp = NULL;
                      lastimport = origline;
                      goto reparsemember;
                    }
                  }
                }
              } else if (cb->getBlockType() == CF_BLOCKTYPE_DATA) {
                ClassFactoryData *cd = cb->getData();
                String dataname = cd->getName(); // Method or Class::Method
                String classname = cd->getClassName();
                int recognized = 0;
                if (!isinclass) {
                  if (!isimport && classname.isequal(thisclass)) {
                    recognized = 1;
                  } else if (isimport) {
                    if (!importimplclassname.isempty() && importimplclassname.isequal(classname)) {
                      recognized = 1;
                    }
                  }
                }
                if (isinclass || recognized) { // Ok, we have a data block
                  // now, see if it is a match with an XHD entry
                  int found = 0;
                  for (int i_chunk=0;i_chunk<chunk->getNumBlocks() && !found;i_chunk++) {
                    ClassFactoryBlock *b = chunk->enumBlock(i_chunk);
                    if (b->getBlockType() == CF_BLOCKTYPE_DATA) {
                      ClassFactoryData *data = b->getData();
                      if (STRCASEEQLSAFE(data->getName(), cd->getName())) {
                        found = 1;
                        data->setFlag(CF_FLAG_DATA_FOUND);
                        data->setSourceLine(linecount);
                        data->setStatic(cd->isStatic() || !isinclass);
                        String init = cd->getInitValue();
                        if (!init.isempty()) data->setInitValue(init);

                        if (!dumbcode.isempty()) {
                          char *p = dumbcode.getNonConstVal();
                          if (p && *p) {
                            int position = chunk->getBlockNum(b);
                            chunk->addBlock(new ClassFactoryBlockI(new ClassFactoryCodeBlockI(dumbcode, blocktypes)), position);
                          }
                          dumbcode.reset();
                        }

                        if (z == 1 && isinclass) {
                          data->setDataType(cd->getDataType());
                          data->setScope(curscope);

                          int newpos = 0;
                          if (lastblock != NULL) {
                            newpos = chunk->getBlockNum(lastblock)+1;
                            for (int lm=newpos;lm<chunk->getNumBlocks();lm++) {
                              if (chunk->enumBlock(lm)->getBlockType() == CF_BLOCKTYPE_METHOD) break;
                              if (chunk->enumBlock(lm)->getBlockType() == CF_BLOCKTYPE_DATA) break;
                              newpos++;
                            }
                          }
                          int oldpos = chunk->getBlockNum(b);
                          if (newpos != oldpos) {
                            chunk->moveBlock(b, newpos);
                            oldpos = chunk->getBlockNum(b);
                            if (oldpos <= i_chunk) {
                              i_chunk--;
                            }
                          }
                        }

                        lastblock = b;
                        eatline = 1;
                        continue;
                      }
                    }
                  }
                  if (!found && z == 1 && isinclass) { // data not found
                    ClassFactoryDataI *data = new ClassFactoryDataI();
                    data->copy(cd);
                    data->setSourceLine(linecount);
                    data->setFlag(CF_FLAG_DATA_FOUND);
                    int newpos = 0;
                    if (lastblock != NULL) {
                      newpos = chunk->getBlockNum(lastblock)+1;
                      for (int lm=newpos;lm<chunk->getNumBlocks();lm++) {
                        if (chunk->enumBlock(lm)->getBlockType() == CF_BLOCKTYPE_METHOD) break;
                        if (chunk->enumBlock(lm)->getBlockType() == CF_BLOCKTYPE_DATA) break;
                        newpos++;
                      }
                    }
                    chunk->addBlock(new ClassFactoryBlockI(data), newpos);
                    goto reparsemember;
                  }
                }
              }
            }
            if (eatline) goto nextline;
          }
          delete cp; cp = NULL;
        }
      }
      dumbcode += origline;
    }
    if (!dumbcode.isempty()) {
      char *p = dumbcode.getNonConstVal();
      /*while (p && *p) {
        if (*p != '\r' && *p != '\n') break;
        p++;
      }*/
      if (p && *p) {
          if (chunk->haveCodeBlocks(blocktypes))
            chunk->setFooter(new ClassFactoryCodeBlockI(dumbcode, blocktypes));
          else
            chunk->setHeader(new ClassFactoryCodeBlockI(dumbcode, blocktypes));
      }
    }
    dumbcode.reset();
    if (classcount > 0 || isinclass)
      chunk->addErrorMessage("end of file before end of class (missing ; after class {} ?)", linecount);

    delete cp; cp = NULL;

    Vfclose(inf);

    if (z == 1 && !mergeh.isempty()) {

      // go thru the methods that had the wrong params
      int wn = wrongparamsmathods.getNumItems();
      int wi;
      for (wi=0;wi<wn;wi++) {
        ClassFactoryMethodI *m = wrongparamsmathods.enumItem(wi);
        String name = m->getName();
        // look for a method with that name that hasn't been implemented
        ClassFactoryBlock *best = NULL;
        int bestdiff = 999999;
        int bn = chunk->getNumBlocks();
        int bi;
        for (bi=0;bi<bn;bi++) {
          ClassFactoryBlock *block = chunk->enumBlock(bi);
          if (block->getBlockType() == CF_BLOCKTYPE_METHOD) {
            ClassFactoryMethod *method = block->getMethod();
            if (wrongparamsmathods.searchItem(static_cast<ClassFactoryMethodI*>(method)) == -1) {
              if (STREQL(method->getName(), name)) {
                if (!method->isFlagSet(CF_FLAG_METHOD_FOUND_IN_CPP) && !method->isFlagSet(CF_FLAG_METHOD_FOUND_IN_H)) {
                  int diff = 0;
                  // find best match
                  int pn = method->getNumParams();
                  int n = m->getNumParams();
                  diff += ABS(pn-n)*2;
                  int pi;
                  for (pi=0;pi<pn&&pi<n;pi++) {
                    ClassFactoryParam *param = method->enumParam(pi);
                    ClassFactoryParam *p = m->enumParam(pi);
                    if (!STRCASEEQL(param->getType(), p->getType())) diff++;
                    if (!STRCASEEQL(param->getName(), p->getName())) diff++;
                  }
                  if (diff < bestdiff) {
                    best = block;
                  }
                }
              }
            }
          }
        }
        if (best != NULL) {
          // we found one, remove it
          ClassFactoryMethod *mbest = best->getMethod();
          if (!m->getCodeBlock("cpp") && mbest->getCodeBlock("cpp")) {
            ClassFactoryCodeBlockI *newcb = new ClassFactoryCodeBlockI();
            newcb->copy(mbest->getCodeBlock("cpp"));
            m->setCodeBlock(newcb);
          }
          if (!m->getCodeBlock("h") && mbest->getCodeBlock("h")) {
            ClassFactoryCodeBlockI *newcb = new ClassFactoryCodeBlockI();
            newcb->copy(mbest->getCodeBlock("h"));
            m->setCodeBlock(newcb);
          }
          if (!m->getCodeBlock("inline") && mbest->getCodeBlock("inline")) {
            ClassFactoryCodeBlockI *newcb = new ClassFactoryCodeBlockI();
            newcb->copy(mbest->getCodeBlock("inline"));
            m->setCodeBlock(newcb);
          }
          if (!m->getHeader("cpp") && mbest->getHeader("cpp")) {
            m->setHeader(mbest->getHeader("cpp"), "cpp");
          }
          if (!m->getFooter("cpp") && mbest->getFooter("cpp")) {
            m->setHeader(mbest->getHeader("cpp"), "cpp");
          }
          if (!m->getHeader("h") && mbest->getHeader("h")) {
            m->setHeader(mbest->getHeader("h"), "h");
          }
          if (!m->getFooter("h") && mbest->getFooter("h")) {
            m->setHeader(mbest->getHeader("h"), "h");
          }
          int n = chunk->getBlockNum(mbest->getParentBlock());
//          ASSERT(wrongparamsmathods.searchItem(static_cast<ClassFactoryMethodI*>(best->getMethod())) == -1);
          chunk->removeBlock(best);
          n--;
          ClassFactoryBlock *cb = chunk->enumBlock(n);
          if (cb->getBlockType() == CF_BLOCKTYPE_CODEBLOCK) {
            int newpos = chunk->getBlockNum(m->getParentBlock());
            chunk->moveBlock(cb, newpos);
          }
        } else {
          // we didn't find any, so everything should be fine
        }
      }

      // go thru the functions and remove those that weren't found
      // go thru the data and remove those that weren't found
      for (int fi=0;fi<chunk->getNumBlocks();fi++) {
        ClassFactoryBlock *block = chunk->enumBlock(fi);
        if (block->getBlockType() == CF_BLOCKTYPE_METHOD) {
          ClassFactoryMethod *method = block->getMethod();
          if (!method->isFlagSet(CF_FLAG_METHOD_FOUND_IN_H)) {
            chunk->deleteBlock(block);
            fi--;
          }
        } else if (block->getBlockType() == CF_BLOCKTYPE_DATA) {
          ClassFactoryData *data = block->getData();
          if (!data->isFlagSet(CF_FLAG_DATA_FOUND)) {
            chunk->deleteBlock(block);
            fi--;
          }
        }
      }

      // remember friends
      String frs;
      foreach(friends);
        if (foreach_index > 0) frs += ";";
        frs += friends.getfor()->getValue();
      endfor;
      friends.deleteAll();
      chunk->getParams()->removeItem("friends");
      chunk->getParams()->addItem("friends", frs);

      // process ancestors

#ifdef ODE_IDE
      ASSERT(m_workspace!= NULL);
#endif

      String par;

/*
      // separate the weak from the strong
      for (int a=0;a<ancestors.getNumItems();a++) {
        Ancestor *anc = ancestors.enumItem(a);
#ifdef ODE_IDE
        ClassFactoryChunk *ancchunk = m_workspace->findChunk(anc->classname);
        if (!ancchunk) {
#endif
          if (!par.isempty()) par += ";";
          switch (anc->scope) {
            case SCOPE_PUBLIC:
              par += "public ";
              break;
            case SCOPE_PRIVATE:
              par += "private ";
              break;
            case SCOPE_PROTECTED:
              par += "protected ";
              break;
          }
          par += anc->classname;
#ifdef ODE_IDE
        } else {
          chunk->addSecondaryAncestor(ancchunk->getClassName(), anc->scope);
        }
#endif
      }
      if (!par.isempty()) chunk->getParams()->addItem("ancestors", par);
      ancestors.deleteAll();
*/

      // done with ancestors
    }

    curpass++;
  }
  if (!didstuff)
    chunk->setFlag(CF_FLAG_CLASS_NEW_IMPLEMETATION);
}


int CPPToXHDMerge::startsWith(const char *str, const char *with) {
  const char *p = str;
  while (*p) {
    if (*p != ' ' && *p != '\t') {
      int eq = !STRNCMP(str, with, STRLEN(with));
      return eq;
    }
    p++;
  }
  return 0;
}

char *CPPToXHDMerge::findEndOfPar(const char *line) {
  // todo: make comment-proof
  int pcount = 0;
  int bcount = 0;
  int incmt = 0;
  char *p = (char *)line;

  while (*p) {
    if (*p == '/' && *(p+1) == '/') return NULL;
    if (*p == '/' && *(p+1) == '*') { incmt = 1; p+=2; continue; }
    if (*p == '*' && *(p+1) == '/') { incmt = 1; p+=2; continue; }
    if (*p == '(' && !incmt) pcount++;
    if (*p == '{' && !incmt) bcount++;
    if (*p == '}' && !incmt) bcount++;
    if (*p == ')' && bcount == 0 && !incmt) {
      pcount--;
      if (pcount == 0) return p;
    }
    p++;
  }
  return NULL;
}

