#ifndef _CAST_H
#define _CAST_H

#include <bfc/string/string.h>

class TypeCast;

class Cast {
public:
  Cast(TypeCast *fromtype, const char *totype, const char *function, int ismacro) : m_fromtype (fromtype), m_totype(totype), m_function(function), m_ismacro(ismacro) {}
  virtual ~Cast() {}

  const char *getFrom();
  const char *getTo() { return m_totype; }
  const char *getCast() { return m_function; }
  int isMacro() { return m_ismacro; }

private:
  TypeCast *m_fromtype;
  String m_totype;
  String m_function;
  int m_ismacro;
};

class CastSort {
public:
  static int compareItem(Cast *p1, Cast *p2) { return STRCMP(p1->getTo(), p2->getTo()); }
  static int compareAttrib(const char *attrib, Cast *item) { return STRCMP(attrib, item->getTo()); }
};

class TypeCast {
public:
  TypeCast(const char *fromtype) : m_fromtype(fromtype) {} 
  virtual ~TypeCast() { m_casts.deleteAll(); }
  
  const char *getType() { return m_fromtype; }
  Cast *getCast(const char *totype) { return m_casts.findItem(totype); }
  Cast *addCast(const char *totype, const char *castfunc, int ismacrocast) {
    Cast *c = getCast(totype);
    if (!c) {
      c = new Cast(this, totype, castfunc, ismacrocast);
      m_casts.addItem(c);
    }
    return c;
  }

private:
  String m_fromtype;
  PtrListQuickSorted<Cast, CastSort> m_casts;
};

class TypeCastSort {
public:
  static int compareItem(TypeCast *p1, TypeCast *p2) { return STRCMP(p1->getType(), p2->getType()); }
  static int compareAttrib(const char *attrib, TypeCast *item) { return STRCMP(attrib, item->getType()); }
};

class CastMgr {
public:
  CastMgr() {}
  virtual ~CastMgr() { m_typecasts.deleteAll(); }
  TypeCast *addTypeCast(const char *fromtype, const char *totype, const char *castfunc, int ismacrocast) {
    TypeCast *tc = m_typecasts.findItem(fromtype);
    if (!tc) {
      tc = new TypeCast(fromtype);
      m_typecasts.addItem(tc);
    }
    tc->addCast(totype, castfunc, ismacrocast);
    return tc;
  }
  const char *getCast(const char *from, const char *to) {
    TypeCast *tc = m_typecasts.findItem(from);
    if (tc) {
      Cast *c = tc->getCast(to);
      if (c) return c->getCast();
    }
    return NULL;
  }
  Cast *getCastObject(const char *from, const char *to) {
    TypeCast *tc = m_typecasts.findItem(from);
    if (tc) {
      return tc->getCast(to);
    }
    return NULL;
  }

private:
  PtrListQuickSorted<TypeCast, TypeCastSort> m_typecasts;
};

#endif

