#include "precomp.h"
#include "FactoryTemplate.h"
#include "TemplatizedFactory.h"
#include "../classhierarchy/ClassFactoryChunk.h"

FactoryTemplate::FactoryTemplate(TemplatizedFactory *parent, const char *format) {
  m_parent = parent;
  m_format = format;
  m_savepos = 0;
  m_setcount = 0;
}

FactoryTemplate::~FactoryTemplate() {
  m_chunks.deleteAll();
}

const char *FactoryTemplate::getFormat() { 
  return m_format; 
}

const char *FactoryTemplate::getDescription() {
  return m_desc;
}

void FactoryTemplate::setDescription(const char *desc) {
  m_desc = desc;
}

TemplateChunk *FactoryTemplate::enumChunk(int n) { 
  return m_chunks.enumItem(n); 
}

int FactoryTemplate::getNumChunks() { 
  return m_chunks.getNumItems(); 
}

void FactoryTemplate::addChunk(TemplateChunk *chunk) { 
  m_chunks.addItem(chunk); 
}

void FactoryTemplate::setMainModuleName(const char *m) {
  m_mainModuleName = m;
}

const char *FactoryTemplate::getMainModuleName() {
  return m_mainModuleName;
}

void FactoryTemplate::popSet() {
  VarSet *str;
  m_setStack.pop(&str);
  delete str;
}

void FactoryTemplate::pushSet(const char *varid, int ispersistant) {
  m_setStack.push(new VarSet(varid, ispersistant, m_parent->getChunk()));
}

int FactoryTemplate::isInSet() {
  return m_setStack.peek() > 0;
}

void FactoryTemplate::addToVars(const char *txt) {
  for (int i=0;i<m_setStack.peek() && i<1;i++) {
    VarSet *s = NULL;
    m_setStack.peekAt(&s, i);
    if (s) {
      if (s->ispersistant) {
        String curval = s->m_chunk->getParams()->getItemValue(s->id);
        curval += txt;
        s->m_chunk->getParams()->replaceItem(s->id, curval);
      } else {
        String curval = m_parent->getVars()->getString(s->id);
        curval += txt;
        m_parent->getVars()->setString(s->id, curval);
      }
    }
  }
}
