#include "precomp.h"
#include "ForEachMethod.h"
#include <bfc/parse/paramparser.h>

#include "../classhierarchy/ClassFactoryChunk.h"
#include "../classhierarchy/ClassFactoryBlock.h"
#include "../classhierarchy/ClassFactoryMethod.h"
#include "../classhierarchy/SecondaryAncestor.h"
#include "../outputfactories/TemplatizedFactory.h"
#include "../classhierarchy/Workspace.h"
#include "../classhierarchy/Project.h"

#include <api/xml/xmlparams.h>

ForEachMethod::ForEachMethod(TemplatizedFactory *parent, ClassFactoryChunk *chunk, const char *filter) : ForIterator(parent, FOREACH_METHOD) {
  m_parent = parent;
  m_chunk = chunk;
  m_filter = filter;
  if (m_chunk == NULL) {
    setNumIterations(0);
    return;
  }
  const char *maxdepth = parent->getTemplateChunk()->getParams()->getItemValue("depth");
  if (!maxdepth)
    m_maxdepth = chunk->getParams()->getItemValueInt(StringPrintf("%s_depth", parent->getName()));
  else
    m_maxdepth = ATOI(parent->translateVariables(maxdepth));

  makeMethodList(chunk, 0);

  setNumIterations(m_methods.getNumItems());
}

ForEachMethod::~ForEachMethod() {
}

ClassFactoryMethod *ForEachMethod::getMethod(int n) {
  if (n == -1) n = getIterator();
  return m_methods.enumItem(n);
}

int ForEachMethod::getNumMethods() { 
  return m_methods.getNumItems(); 
}

void ForEachMethod::makeMethodList(ClassFactoryChunk *chunk, int curdepth) {
  int all = m_parent->getTemplateChunk()->getParams()->getItemValueInt("all");
  int bn = chunk->getNumBlocks();
  int b;
  for (b = 0;b < bn; b++) {
    ClassFactoryBlock *block = chunk->enumBlock(b);
    if (block->getBlockType() == CF_BLOCKTYPE_METHOD) {
      ClassFactoryMethod *method = block->getMethod();
      if (!m_filter.isempty()) {
        if (!Std::match(m_filter, method->getName())) continue;
      }
      // check if function is hidden
      if (!all && method->getParams()->getItemValueInt("hide")) continue;
      // check if this function has already been listed via one of its overrides
      int pos = -1, skip = 0;
      ClassFactoryMethod *m = m_methods.findItem(method->getName(), &pos);
      if (m != NULL) do {
        if (overrides(method, m)) skip = 1;
        m = m_methods.enumItem(++pos);
      } while (m && STREQLSAFE(method->getName(), m->getName()));
      if (skip) continue;
      
      m_methods.addItem(method);
    }
  }

  if (curdepth == m_maxdepth) return;
  curdepth++;

  for (int c=0; c < chunk->getNumSecondaryAncestors(); c++) {
    SecondaryAncestor *se = chunk->enumSecondaryAncestor(c);
    ClassFactoryChunk *sec = g_workspace->findChunk(se->getClassName());
    if (sec) {
      makeMethodList(sec, curdepth);
    }
  }
}

int ForEachMethod::overrides(ClassFactoryMethod *possibleancestor, ClassFactoryMethod *possibleoverride) {
  if (!STREQLSAFE(possibleancestor->getName(), possibleoverride->getName())) return 0;
  if (possibleancestor->getNumParams() != possibleoverride->getNumParams()) return 0;
  for (int i=0;i<possibleancestor->getNumParams();i++) {
    ClassFactoryParam *pa_par = possibleancestor->enumParam(i);
    ClassFactoryParam *po_par = possibleoverride->enumParam(i);
    if (!STREQLSAFE(pa_par->getType(), po_par->getType())) return 0;
  }
  return 1;
}

ForEachRetiredMethod::ForEachRetiredMethod(TemplatizedFactory *parent, ClassFactoryChunk *chunk, ClassFactoryMethod *m) :
  ForEachMethod(parent, chunk, StringPrintf("_RETIRED_%s", m->getName())) {
}

