#include "precomp.h"
#include "../classhierarchy/ClassFactoryChunkI.h"
#include "../classhierarchy/Project.h"
#include "../classhierarchy/Workspace.h"
#include "ForEachProjectFolder.h"
#include "TemplatizedFactory.h"
#include <bfc/parse/pathparse.h>

ForEachProjectFolder::ForEachProjectFolder(TemplatizedFactory *parent, const char *rootitem, int nextfolder, int thisfolder) : ForIterator(parent, FOREACH_PROJECTFOLDERS) {
  String root = rootitem;
  if (root.isempty()) root = "/root/";
  if (!Std::isDirChar(root.lastChar())) root += "/";
  int nextfoldern = 2;
  if (nextfolder) {
    ForEachProjectFolder *fepf = parent->getForEachProjectFolder();
    if (fepf) {
      String s = fepf->getFolder();
      PathParser pp(s, "/");
      nextfoldern = pp.getNumStrings()+1;
    }
  }
  // enumerate files from project
  int n = parent->getWorkspace()->getNumChunks();
  int j;
  for (j=0;j<n;j++) {
    ClassFactoryChunk *chunk = parent->getWorkspace()->enumChunk(j);
    String folder = chunk->getParams()->getItemValue("project_folder");
    if (folder.getChar(0) != '/') folder = StringPrintf("/%s", folder.getValue());
    if (folder.lastChar() != '/') folder.cat("/");
    if (!STRCASEEQLSAFE(folder, "")) {
      folder = StringPrintf("/root%s", folder.getValue());
      if (m_folders.findItem(folder)) continue;
      if (root != NULL && *root) {
        if (STRNICMP(root, folder.getValue(), STRLEN(root)))
          continue;
        if (thisfolder && STRCASEEQL(root, folder))
          continue;
      }
      if (nextfolder) {
        PathParser pp(folder, "/");
        if (pp.getNumStrings() != nextfoldern) continue;
      }
      m_folders.addItem(new String(folder));
    }
  }

#ifdef ODE_IDE
  // enumerate packaged files from project
  n = parent->getWorkspace()->getNumFiles();
  for (j=0;j<n;j++) {
    ProjectFile *file = parent->getWorkspace()->enumFile(j);
    String folder = file->getProjectFolder();
    if (folder.getChar(0) != '/') folder = StringPrintf("/%s", folder);
    if (folder.lastChar() != '/') folder.cat("/");
    if (!STRCASEEQLSAFE(folder, "")) {
      folder = StringPrintf("/root%s", folder);
      if (m_folders.findItem(folder)) continue;
      if (root != NULL && *root) {
        if (strnicmp(root, folder, STRLEN(root)))
          continue;
        if (thisfolder && stricmp(root, folder))
          continue;
      }
      if (nextfolder) {
        PathParser pp(folder, "/");
        if (pp.getNumStrings() != nextfoldern) continue;
      }
      m_folders.addItem(new String(folder));
    }
  }
#endif
  setNumIterations(m_folders.getNumItems());
}

ForEachProjectFolder::~ForEachProjectFolder() {
  m_folders.deleteAll();
}

const char *ForEachProjectFolder::getFolder(int n) {
  if (n == -1) n = getIterator();
  return m_folders.enumItem(n)->getValue();
}
