#include "precomp.h"
#include "ForEachTemplateParam.h"
#include "TemplatizedFactory.h"
#include "../classhierarchy/ClassFactoryParser.h"
#include "../classhierarchy/ClassFactoryMethod.h"
#include "../classhierarchy/ClassFactoryData.h"
#include "../classhierarchy/ClassFactoryParam.h"
#include "../classhierarchy/ClassFactoryChunkI.h"

ForEachTemplateParam::ForEachTemplateParam(TemplatizedFactory *parent, ClassFactoryData *data) : ForIterator(parent, FOREACH_TEMPLATEPARAM) {
  m_method = NULL;
  m_chunk = NULL;

  if (data == NULL || !data->getDataType() || !*(data->getDataType())) {
    setNumIterations(0);
    return;
  }

  makeList(data->getDataType());
}

ForEachTemplateParam::ForEachTemplateParam(TemplatizedFactory *parent, ClassFactoryParam *param) : ForIterator(parent, FOREACH_TEMPLATEPARAM) {
  m_chunk = NULL;

  if (param == NULL || !param->getType() || !*(param->getType())) {
    setNumIterations(0);
    return;
  }

  makeList(param->getType());
}

void ForEachTemplateParam::makeList(const char *str) {
  String templatename;
  PtrList<String> templateparams;
  TemplatizedFactory::parseTemplate(str, templatename, templateparams);
  if (templatename.isempty()) {
    setNumIterations(0);
    templateparams.deleteAll();
    return;
  }

  String fakefunctiondecl = "void foo(";
  foreach(templateparams)
    String *s = templateparams.getfor();
    if (foreach_index > 0) fakefunctiondecl += ", ";
    fakefunctiondecl += *s;
    fakefunctiondecl += " ";
    fakefunctiondecl += StringPrintf("%c", 'a' + foreach_index);
  endfor;
  fakefunctiondecl += ");";

  templateparams.deleteAll();

  m_chunk = new ClassFactoryChunkI();
  ClassFactoryParser::parse(fakefunctiondecl, m_chunk);
  m_method = NULL;
  for (int i=0;i<m_chunk->getNumBlocks();i++) {
    ClassFactoryBlock *block = m_chunk->enumBlock(i);
    if (block->getBlockType() == CF_BLOCKTYPE_METHOD) {
      m_method = block->getMethod();
      break;
    }
  }

  if (!m_method) {
    setNumIterations(0);
    return;
  }

  setNumIterations(m_method->getNumParams());
}

ForEachTemplateParam::~ForEachTemplateParam() {
  delete m_chunk;
}

ClassFactoryParam *ForEachTemplateParam::getParam(int n) {
  if (!m_method) return 0;
  if (n == -1) n = getIterator();
  return m_method->enumParam(n);
}


