#include "precomp.h"
#include "Helper.h"
#include <bfc/file/wildcharsenum.h>
#include <bfc/parse/pathparse.h>
#include <bfc/parse/paramparser.h>
#include "../classhierarchy/Project.h"
#include "../classhierarchy/Workspace.h"

BEGIN_SERVICES(HelperSCF_Svc)
DECLARE_SERVICE_DEBUG(SrcClassFactoryCreator<Helper>, "XPST Factory");
END_SERVICES(HelperSCF_Svc, _HelperSCF_Svc);

#ifdef _X86_
extern "C" { int _link_HelperSCFSvc; }
#else
extern "C" { int __link_HelperSCFSvc; }
#endif

extern String g_resourcepath;

Helper::Helper(const char *ext) {
  m_ext = ext;
  String td = g_workspace->getTemplateDir();
  if (!td.isempty()) {
    if (!Std::isDirChar(td.lastChar())) td.cat(DIRCHARSTR);
    td += StringPrintf("*.%s;", ext);
  }

  String res = g_resourcepath;
  if (!res.isempty() && !Std::isDirChar(res.lastChar())) res.cat(DIRCHARSTR);
  td += StringPrintf("%stemplates/*.%s", res.getValue(), ext);

  //printf("searching in \"%s\"\n", td.getValue());

  WildcharsEnumerator en(td);
  for (int i=0;i<en.getNumFiles();i++) {
    PathParser pp(en.enumFile(i));
    String file = pp.getLastString();
    ParamParser pp2(file, ".");
    m_files.addItem(new String(pp2.enumItem(0)));
  }
}

Helper::~Helper() {
  m_files.deleteAll();
}

void Helper::onNewChunk(ClassFactoryChunk *chunk) {
  String format = getCustomFormat(getName());
  loadFormatTemplate(format, NULL, m_ext);
}

int Helper::getNumAvailableFormats() {
  return m_files.getNumItems();
}

const char *Helper::enumAvailableFormat(int i) {
  return m_files.enumItem(i)->getValue();
}

int Helper::tryFormat(const char *format) {
  m_format = format;
  setName(m_format);
  return m_files.findItem(format) != NULL;
}

