#include "precomp.h"
#include "InlineXmlFactory.h"
#include <bfc/string/stringdict.h>
#include <time.h>

#include "../classhierarchy/ClassFactoryChunk.h"
#include "../classhierarchy/ClassFactoryMethod.h"
#include "../classhierarchy/ClassFactoryParam.h"
#include "../classhierarchy/ChunkHierarchy.h"
#include "../classhierarchy/Project.h"

#include <bfc/parse/pathparse.h>

enum {
  IX_ANCESTOR,
  IX_METHOD,
  IX_PARAMS,
  IX_RETTYPE,
  IX_DEFRETVAL,
  IX_CALLUP,
  IX_NOCALLUP,
};

BEGIN_STRINGDICTIONARY(_INLINEXMLTAGS)
SDI("ancestor", IX_ANCESTOR);
SDI("name", IX_METHOD);
SDI("params", IX_PARAMS);
SDI("rettype", IX_RETTYPE);
SDI("defretval", IX_DEFRETVAL);
SDI("callup", IX_CALLUP);
SDI("nocallup", IX_NOCALLUP);
END_STRINGDICTIONARY(_INLINEXMLTAGS, inlinexmltags)

//

InlineXmlFactory::InlineXmlFactory() {
  m_method = NULL;
  incallup = 0;
  innocallup = 0;
  setName("inlinexml");
  m_cpp.setId("inlinexml");
  m_cpp.setName("autocallup");
  registerModule(&m_cpp);
}

InlineXmlFactory::~InlineXmlFactory() {
}

void InlineXmlFactory::setMethod(ClassFactoryMethod *method) {
  incallup = 0;
  innocallup = 0;
  m_method = method;
}

void InlineXmlFactory::setAncestor(const char *ancestor) {
  m_ancestor = ancestor;
}

const char *InlineXmlFactory::getAncestor() {
#ifdef ODE_IDE
  if (m_method != NULL && getChunk() != NULL) {
    ClassFactoryChunk *basechunk = NULL;
    IdeChunkHierarchy h(getChunk());
    String path;
    ClassFactoryMethod *method = h.findMethod(m_method->getName(), path, 0, &basechunk, 0, 0);
    if (method != NULL && basechunk != getChunk()) {
      PathParser pp(path);
      ASSERT(pp.getNumStrings() >= 2);
      String firstclass = pp.enumString(1);
      if (!firstclass.isempty()) {
        basechunk = getWorkspace()->findChunk(firstclass);
      }
      ASSERT(basechunk != NULL);
      tempString = basechunk->getImplementationName();
      return tempString;
    }
  }
#endif
  return m_ancestor; // this can't possibly work, heh
}

ClassFactoryMethod *InlineXmlFactory::getMethod() {
  return m_method;
}

void InlineXmlFactory::onCustomTag(const char *tag, XmlReaderParams *params, int isclose, int discard) {
  int a = inlinexmltags.getId(tag);
  if (discard && a != IX_CALLUP && a != IX_NOCALLUP) return;
  switch (a) {
    case IX_CALLUP:
      if (isclose)
        incallup--;
      else
        incallup++;
      break;
    case IX_NOCALLUP:
      if (isclose)
        innocallup--;
      else
        innocallup++;
      break;
    case IX_ANCESTOR:
      if (getCustomSkip()) break;
      if (isclose) break;
      addToCurData(getAncestor());
      break;
    case IX_METHOD: {
      if (getCustomSkip()) break;
      if (isclose) break;
      if (!m_method) break;
      String s;
      if (getChunk()->isConstructor(m_method))  {
        s += getChunk()->getImplementationName();
      } else if (getChunk()->isDestructor(m_method)) {
        s += StringPrintf("~%s", getChunk()->getImplementationName());
      } else {
        s += m_method->getName();
      }
      addToCurData(s);
      break;
    }
    case IX_PARAMS: {
      if (getCustomSkip()) break;
      if (isclose) break;
      if (!m_method) break;
      String s;
      int i, num = m_method->getNumParams();
      for (i = 0; i < num; i++) {
        ClassFactoryParam *param = m_method->enumParam(i);
        String paramName = param->getName();
        s += paramName;
        if (i != num - 1) {
          s += ", ";
        }
      }
      addToCurData(s);
      break;
    }
    case IX_RETTYPE: {
      if (getCustomSkip()) break;
      if (isclose) break;
      if (!m_method) break;
      String v = m_method->getReturnType();
      if (v.lastChar() != '*' && v.lastChar() != '&') v += " ";
      addToCurData(v);
      break;
    }
    case IX_DEFRETVAL:
      if (getCustomSkip()) break;
      if (isclose) break;
      if (!m_method) break;
      String dv = m_method->getDefaultReturnValue();
      if (!dv || !*dv) {
        getDefaultForType(m_method->getReturnType(), dv);
      }
      addToCurData(dv);
      break;
  }
}

int InlineXmlFactory::getCustomSkip() {
  String anc = getAncestor();
  if (incallup && anc.isempty()) return 1;
  if (innocallup && !anc.isempty()) return 1;
  return 0;
}

void InlineXmlFactory::setChunk(ClassFactoryChunk *chunk) {
  TemplatizedFactory::setChunk(chunk);
  m_cpp.setData(NULL);
}
