#include "precomp.h"
#include "InterfaceFactory.h"
#include <api/xml/xmlparams.h>
#include "../classhierarchy/ClassFactoryChunk.h"

BEGIN_SERVICES(InterfaceSCF_Svc)
DECLARE_SERVICE_DEBUG(SrcClassFactoryCreator<InterfaceFactory>, "Interface Factory");
END_SERVICES(InterfaceSCF_Svc, _InterfaceSCF_Svc);

#ifdef _X86_
extern "C" { int _link_InterfaceSCFSvc; }
#else
extern "C" { int __link_InterfaceSCFSvc; }
#endif


InterfaceFactory::InterfaceFactory() {
  setName("Interface");
  setExt("xpt");
  m_dispatchable = 0;
}

InterfaceFactory::~InterfaceFactory() {
}

void InterfaceFactory::onNewChunk(ClassFactoryChunk *chunk) {
  XmlReaderParams *par = chunk->getParams();
  m_dispatchable = par->getItemValueInt("dispatchable");

  String name;
  name = par->getItemValue("interface_header");
  if (name.isempty()) {
    name = String(chunk->getClassName()) + ".h";
    par->removeItem("interface_header");
    par->addItem("interface_header", name);
  }
  m_h.setName(name);
  m_h.setData("");
  m_h.setId("h");
  registerModule(&m_h);

  name.trunc(0);
  name = par->getItemValue("interface_module");
  if (name.isempty()) {
    name = String(chunk->getClassName()) + ".cpp";
    par->removeItem("interface_module");
    par->addItem("interface_module", name);
  }
  m_cpp.setName(name);
  m_cpp.setData("");
  m_cpp.setId("cpp");
  registerModule(&m_cpp);

  if (m_dispatchable) {
    name.trunc(0);
    name = par->getItemValue("dispatchable_header");
    if (name.isempty()) {
      name = String(StringPrintf("%sX.h", chunk->getClassName()));
      par->removeItem("dispatchable_header");
      par->addItem("dispatchable_header", name);
    }
    m_h_dispatch.setName(name);
    m_h_dispatch.setData("");
    m_h_dispatch.setId("h_dispatch");
    registerModule(&m_h_dispatch);

    name.trunc(0);
    name = par->getItemValue("dispatchable_module");
    if (name.isempty()) {
      name = String(StringPrintf("%sX.cpp", chunk->getClassName()));
      par->removeItem("dispatchable_module");
      par->addItem("dispatchable_module", name);
    }
    m_cpp_dispatch.setName(name);
    m_cpp_dispatch.setData("");
    m_cpp_dispatch.setId("cpp_dispatch");
    registerModule(&m_cpp_dispatch);

    chunk->setAncestor(StringPrintf("%sX", chunk->getClassName()));
  } else
    chunk->setAncestor(chunk->getClassName());
}

const char *InterfaceFactory::getCustomFormat(const char *format) {
  if (m_dispatchable) return "Dispatchable";
  return "PureVirtual";
}

