#include "precomp.h"
#include "TemplateChunk.h"
#include <bfc/string/url.h>

int CachedStringItemComparator::compareItem(CachedStringItem *p1, CachedStringItem *p2) {
  return STRCMP(p1->getValue(), p2->getValue());
}

int CachedStringItemComparator::compareAttrib(const char *attrib, CachedStringItem *item) {
  return STRCMP(attrib, item->getValue());
}

CachedString::CachedString() {
  m_cs = NULL;
}

CachedString::CachedString(const char *str) {
  m_cs = NULL;
  setValue(str);
}

CachedString::~CachedString() {
  unSet();
}

PtrListQuickSorted<CachedStringItem, CachedStringItemComparator> CachedString::m_strings;

void CachedString::unSet() {
  if (m_cs) {
    if (m_cs->getRefCount() == 1) {
      m_strings.removeItem(m_cs);
      delete m_cs;
      m_cs = NULL;
    } else {
      m_cs->delRef();
      m_cs = NULL;
    }
  }
}

const char *CachedString::setValue(const char *newval) {
  if (m_cs != NULL) unSet();
  if (newval == NULL) return NULL;
  CachedStringItem *cs = m_strings.findItem(newval);
  if (cs != NULL) {
    m_cs = cs;
    m_cs->addRef();
  } else {
    m_strings.addItem(new CachedStringItem(newval));
  }
  return newval;
}

TemplateChunkI::TemplateChunkI(int close, int type, const char *name, XmlReaderParams *par, const char *cdata, const char *filename, const char *incpath, int fileline) {
  if (par != NULL) {
    for(int i=0;i<par->getNbItems();i++) {
      m_params.addItem(par->getItemName(i), par->getItemValue(i));
    }
  }
  m_name = name;
  m_cdata = cdata;
  Url::decode(m_cdata);
  m_close = close,
  m_type = type;
  m_filename.setValue(filename);
  m_incpath.setValue(incpath);
  m_line = fileline;
}

TemplateChunkI::~TemplateChunkI() {
}

XmlReaderParams *TemplateChunkI::getParams() { 
  return &m_params; 
}

const char *TemplateChunkI::getName() { 
  return m_name; 
}

const char *TemplateChunkI::getCData() { 
  return m_cdata; 
}

int TemplateChunkI::getType() { 
  return m_type; 
}

int TemplateChunkI::isClosingTag() { 
  return m_close; 
}
