#ifndef __TEMPLETECHUNK_H
#define __TEMPLETECHUNK_H

#include <api/xml/xmlparamsi.h>
#include <bfc/string/string.h>

class TemplateChunk {
public:
  virtual XmlReaderParams *getParams()=0;
  virtual const char *getName()=0;
  virtual const char *getCData()=0;
  virtual int getType()=0;
  virtual int isClosingTag()=0;
  virtual const char *getFilename()=0;
  virtual const char *getIncludePath()=0;
  virtual int getFileLine()=0;
};

class CachedStringItem {
  public:
   CachedStringItem(const char *str) : m_usagecount(1), m_str(str) {}
   virtual ~CachedStringItem() { ASSERT(m_usagecount == 1); }

   const char *getValue() { return m_str; }
   void addRef() { m_usagecount++; }
   void delRef() { m_usagecount--; }
   int getRefCount() { return m_usagecount; }

  private:
    int m_usagecount;
    String m_str;
};

class CachedStringItemComparator {
public:
  static int compareItem(CachedStringItem *p1, CachedStringItem *p2);
  static int compareAttrib(const char *attrib, CachedStringItem *item);
};

class CachedString {
  public:
    CachedString();
    CachedString(const char *str);
    virtual ~CachedString();

    const char *getValue() { if (!m_cs) return NULL; return m_cs->getValue(); }
    const char *setValue(const char *newval);

  private:
    void unSet();
    CachedStringItem *m_cs;
    static PtrListQuickSorted<CachedStringItem, CachedStringItemComparator> m_strings;
};

class TemplateChunkI : public TemplateChunk {
public:
  TemplateChunkI(int close, int type, const char *name, XmlReaderParams *par = NULL, const char *cdata = NULL, const char *filename=NULL, const char *incpath=NULL, int fileline=-1);
  virtual ~TemplateChunkI();
  virtual XmlReaderParams *getParams();
  virtual const char *getName();
  virtual const char *getCData();
  virtual int getType();
  virtual int isClosingTag();
  virtual const char *getFilename() { return m_filename.getValue(); }
  virtual const char *getIncludePath() { return m_incpath.getValue(); }
  virtual int getFileLine() { return m_line; }

private:
  String m_name;
  int m_type;
  int m_close;
  String m_cdata;
  CachedString m_filename;
  CachedString m_incpath;
  int m_line;
  XmlReaderParamsI m_params;
};


#endif
