#include "precomp.h"
#include "TemplatizedFactory.h"
#include <bfc/string/stringdict.h>
#include <bfc/parse/paramparser.h>
#include <bfc/parse/pathparse.h>
#include "CPPToXHDMerge.h"
#include "../classhierarchy/HelperFileEnum.h"
#include "../classhierarchy/ClassFactoryCodeBlock.h"
#include "../classhierarchy/ClassFactoryBlock.h"
#include "../classhierarchy/ClassFactoryChunk.h"
#include "../classhierarchy/ClassFactoryParamI.h"
#include "../classhierarchy/ClassFactoryMethod.h"
#include "../classhierarchy/ClassFactoryParser.h"
#include "../classhierarchy/ClassFactoryData.h"
#include "../classhierarchy/HelperEnum.h"
#include "../classhierarchy/XHDProcessor.h"
#include "../classhierarchy/AncestorEnum.h"
#include "../classhierarchy/Project.h"
#include "../classhierarchy/Workspace.h"

#include "ForIterator.h"
#include "ForEachMethod.h"
#include "ForEachParam.h"
#include "ForEachData.h"
#include "ForEachCodeBlock.h"
#include "ForEachHelper.h"
#include "ForEachHelperFile.h"
#include "ForEachImplFile.h"
#include "ForEachProjectFolder.h"
#include "ForEachTemplateParam.h"
#include "ForEachAncestor.h"
#include "ForEachClass.h"
#include "ForEachMapEntries.h"
#ifdef ODE_IDE
#include "ForEachGroupDefFolder.h"
#include "ForEachPackagedFile.h"
#endif
#include "ImplementationFactory.h"

#include "GlobalStrings.h"
#include <api/xml/parser/xmlparse.h>

#include <time.h>

extern String g_outputDir;

BEGIN_STRINGDICTIONARY(_FACTORYTEMPLATETAGS)
SDI("CDATA", FT_CDATA); // reserved
SDI("XPP:StandaloneTemplate", FT_ROOT);
SDI("XPP:Template", FT_ROOT);
SDI("XPP:ProjectExportTemplate", FT_ROOT);
SDI("XPP:DataTemplate", FT_ROOT);
SDI("Ode:HelperStandaloneTemplate", FT_ROOT);
SDI("Ode:HelperTemplate", FT_ROOT);
SDI("Ode:TreeTemplate", FT_ROOT);
SDI("Ode:ProjectExportTemplate", FT_ROOT);
SDI("Ode:DataServerTemplate", FT_ROOT);
SDI("PROC", FT_PROC);
SDI("CALL", FT_CALL);
SDI("BR", FT_BR);
SDI("OB", FT_OB);
SDI("CB", FT_CB);
SDI("Module", FT_MODULE);
SDI("ModuleData", FT_MODULEDATA);
SDI("Class", FT_CLASS);
SDI("CodeBlock", FT_CODEBLOCK);
SDI("Method", FT_METHOD);
SDI("Param", FT_PARAM);
SDI("TemplateParam", FT_TEMPLATEPARAM);
SDI("Data", FT_DATA);
SDI("Indent", FT_INDENT);
SDI("ForEach", FT_FOREACH);
SDI("If", FT_IF);
SDI("Else", FT_ELSE);
SDI("ForComma", FT_FORCOMMA);
SDI("ForIndex", FT_FORINDEX);
SDI("AddParamComma", FT_ADDPARAMCOMMA);
SDI("DateTime", FT_DATETIME);
SDI("Skip", FT_SKIP);
SDI("Continue", FT_CONTINUE);
SDI("Break", FT_BREAK);
SDI("ResetScope", FT_RESETSCOPE);
SDI("Implementation", FT_IMPLEMENTATION);
SDI("Interface", FT_INTERFACE);
SDI("AutoHeader", FT_AUTOHEADER);
SDI("DeriveAncestor", FT_DERIVEANCESTOR);
SDI("ClassAncestor", FT_ANCESTOR);
SDI("HelperFile", FT_HELPERFILE);
SDI("Project", FT_PROJECT);
SDI("DataTemplate", FT_DATATEMPLATE);
SDI("StdAfx", FT_STDAFX);
SDI("PrecompFile", FT_STDAFX);
SDI("Extern", FT_EXTERN);
SDI("GetGlobal", FT_GETGLOBAL);
SDI("SetGlobal", FT_SETGLOBAL);
SDI("T", FT_T);
SDI("Get", FT_GET);
SDI("Set", FT_SET);
SDI("PGet", FT_PGET);
SDI("PSet", FT_PSET);
SDI("TypeMap", FT_TYPEMAP);
SDI("Map", FT_MAP);
SDI("TypeCast", FT_TYPECAST);
SDI("Cast", FT_CAST);
SDI("ClassDecl", FT_CLASSDECL);
SDI("Guid", FT_GUID);
SDI("CGuid", FT_CGUID);
SDI("Inc", FT_INC);
SDI("Dec", FT_DEC);
SDI("MapEntry", FT_MAPENTRY);
SDI("Year", FT_YEAR);
SDI("Month", FT_MONTH);
SDI("MonthName", FT_MONTHNAME);
SDI("MonthDay", FT_MONTHDAY);
SDI("WeekDay", FT_WEEKDAY);
SDI("Hour", FT_HOUR);
SDI("Minute", FT_MINUTE);
SDI("Second", FT_SECOND);
SDI("Error", FT_ERROR);
SDI("Warning", FT_WARNING);
//SDI("Context", FT_CONTEXT);
SDI("ProjectFolder", FT_PROJECTFOLDER);
#ifdef ODE_IDE
SDI("PackagedFile", FT_PACKAGEDFILE);
#endif
END_STRINGDICTIONARY(_FACTORYTEMPLATETAGS, fttags)

BEGIN_STRINGDICTIONARY(_CLASSMODULEDATATAGS)
SDI("Filename", FT_MODULEDATA_FILENAME);
SDI("Id", FT_MODULEDATA_ID);
END_STRINGDICTIONARY(_CLASSMODULEDATATAGS, classmoduledatatags)

/*BEGIN_STRINGDICTIONARY(_CONTEXTTAGS)
SDI("Class", FT_CONTEXT_CLASS);
SDI("Method", FT_CONTEXT_);
SDI("Param", FT_CONTEXT_);
SDI("", FT_CONTEXT_);
SDI("", FT_CONTEXT_);
SDI("", FT_CONTEXT_);
SDI("", FT_CONTEXT_);
SDI("", FT_CONTEXT_);
SDI("", FT_CONTEXT_);
SDI("", FT_CONTEXT_);
SDI("", FT_CONTEXT_);
SDI("", FT_CONTEXT_);
SDI("", FT_CONTEXT_);
SDI("", FT_CONTEXT_);
SDI("", FT_CONTEXT_);
SDI("", FT_CONTEXT_);
END_STRINGDICTIONARY(_CONTEXTTAGS, contexttags)*/

BEGIN_STRINGDICTIONARY(_CLASSDATATAGS)
SDI("Guid", FT_CLASS_GUID);
SDI("CGuid", FT_CLASS_CGUID);
SDI("LayerGuid", FT_CLASS_LAYERGUID);
SDI("LayerCGuid", FT_CLASS_LAYERCGUID);
SDI("Name", FT_CLASS_NAME);
SDI("ImplName", FT_CLASS_IMPLNAME);
SDI("NumMethods", FT_CLASS_NUMMETHODS);
SDI("InterfaceHeader", FT_CLASS_INTERFACEH);
SDI("InterfaceModule", FT_CLASS_INTERFACECPP);
SDI("ImplementationHeader", FT_CLASS_IMPLEMENTATIONH);
SDI("ImplementationModule", FT_CLASS_IMPLEMENTATIONCPP);
SDI("TargetDir", FT_CLASS_TARGETDIR);
SDI("DispatchHeader", FT_CLASS_DISPATCHH);
SDI("DispatchModule", FT_CLASS_DISPATCHCPP);
SDI("ClassLayer", FT_CLASS_CLASSLAYER);
SDI("LayerAncestor", FT_CLASS_LAYERANCESTOR);
END_STRINGDICTIONARY(_CLASSDATATAGS, classdatatags)

BEGIN_STRINGDICTIONARY(_METHODDATATAGS)
SDI("Name", FT_METHOD_NAME);
SDI("ReturnType", FT_METHOD_RETURNTYPE);
SDI("DefRetVal", FT_METHOD_DEFRETVAL);
SDI("NumParams", FT_METHOD_NUMPARAMS);
SDI("Virtual", FT_METHOD_VIRTUAL);
SDI("Static", FT_METHOD_STATIC);
SDI("Inline", FT_METHOD_INLINE);
SDI("Pure", FT_METHOD_PURE);
SDI("Scope", FT_METHOD_SCOPE);
SDI("Dispatch_Id", FT_METHOD_DISPATCHID);
SDI("Dispatch_SId", FT_METHOD_DISPATCHSID);
SDI("Return", FT_METHOD_RETURN);
SDI("Dispatch_Call", FT_METHOD_DISPATCHCALL);
SDI("Dispatch_DefRetVal", FT_METHOD_DISPATCHDEFRETVAL);
SDI("Dispatch_Cb", FT_METHOD_DISPATCHCB);
SDI("MultipleId", FT_METHOD_MULTIPLEID);
SDI("SourceLine", FT_METHOD_SOURCELINE);
END_STRINGDICTIONARY(_METHODDATATAGS, methoddatatags)

BEGIN_STRINGDICTIONARY(_CODEBLOCKDATATAGS)
SDI("Scope", FT_CODEBLOCK_SCOPE);
SDI("Block", FT_CODEBLOCK_BLOCK);
END_STRINGDICTIONARY(_CODEBLOCKDATATAGS, codeblockdatatags)

BEGIN_STRINGDICTIONARY(_PARAMDATATAGS)
SDI("Name", FT_PARAM_NAME);
SDI("Type", FT_PARAM_TYPE);
SDI("AliasType", FT_PARAM_ALIASTYPE);
SDI("DefVal", FT_PARAM_DEFVAL);
SDI("PtrRef", FT_PARAM_PTRREF);
SDI("IO", FT_PARAM_IO);
END_STRINGDICTIONARY(_PARAMDATATAGS, paramdatatags)

BEGIN_STRINGDICTIONARY(_DATADATATAGS)
SDI("Name", FT_DATA_NAME);
SDI("Type", FT_DATA_TYPE);
SDI("CTorCode", FT_DATA_CTORCODE);
SDI("DTorCode", FT_DATA_DTORCODE);
SDI("AliasType", FT_DATA_ALIASTYPE);
SDI("SourceLine", FT_DATA_SOURCELINE);
END_STRINGDICTIONARY(_DATADATATAGS, datadatatags)

BEGIN_STRINGDICTIONARY(_MAPENTRYDATATAGS)
SDI("From", FT_MAPENTRY_FROM);
SDI("To", FT_MAPENTRY_TO);
SDI("Impl", FT_MAPENTRY_IMPL);
END_STRINGDICTIONARY(_MAPENTRYDATATAGS, mapentrydatatags)

BEGIN_STRINGDICTIONARY(_MAPENTRYTYPETAGS)
END_STRINGDICTIONARY(_MAPENTRYTYPETAGS, mapentrytypetags)

BEGIN_STRINGDICTIONARY(_FOREACHOBJECTTAGS)
SDI("Methods", FOREACH_METHOD);
SDI("Method", FOREACH_METHOD);
SDI("RetiredMethods", FOREACH_RETIREDMETHOD);
SDI("Params", FOREACH_PARAM);
SDI("Param", FOREACH_PARAM);
SDI("Data", FOREACH_DATA);
SDI("TemplateParam", FOREACH_TEMPLATEPARAM);
SDI("CodeBlocks", FOREACH_CODEBLOCK);
SDI("Helpers", FOREACH_HELPERS);
SDI("HelperFiles", FOREACH_HELPERFILES);
SDI("ImplementationFiles", FOREACH_IMPLFILES);
SDI("Class", FOREACH_CLASS);
SDI("Ancestors", FOREACH_ANCESTORS);
SDI("Classes", FOREACH_CLASS);
SDI("ProjectFolders", FOREACH_PROJECTFOLDERS);
#ifdef ODE_IDE
SDI("GroupDefFolders", FOREACH_GROUPDEFFOLDERS);
SDI("PackagedFiles", FOREACH_PACKAGEDFILES);
#endif
SDI("MapEntries", FOREACH_MAPENTRIES);
END_STRINGDICTIONARY(_FOREACHOBJECTTAGS, foreachobjecttags)

BEGIN_STRINGDICTIONARY(_METHODTYPETAGS)
SDI("Ctor", FT_METHODTYPE_CTOR);
SDI("DefaultCtor", FT_METHODTYPE_DEFCTOR);
SDI("Dtor", FT_METHODTYPE_DTOR);
SDI("Virtual", FT_METHODTYPE_VIRTUAL);
SDI("Static", FT_METHODTYPE_STATIC);
SDI("Inline", FT_METHODTYPE_INLINE);
SDI("Pure", FT_METHODTYPE_PURE);
SDI("Raw", FT_METHODTYPE_RAW);
SDI("Dispatch", FT_METHODTYPE_DISPATCH);
SDI("NoDispatch", FT_METHODTYPE_NODISPATCH);
SDI("FirstOfMultiple", FT_METHODTYPE_FIRSTOFMULTIPLE);
SDI("Multiple", FT_METHODTYPE_MULTIPLE);
SDI("HasRetiredVersions", FT_METHODTYPE_HASRETIREDVERSIONS);
SDI("IsRetired", FT_METHODTYPE_ISRETIRED);
SDI("Void", FT_METHODTYPE_ISVOID);
SDI("VoidParam", FT_METHODTYPE_ISVOIDPARAM);
SDI("Script", FT_METHODTYPE_SCRIPT);
SDI("Event", FT_METHODTYPE_EVENT);
SDI("Public", FT_METHODTYPE_PUBLIC);
SDI("Private", FT_METHODTYPE_PRIVATE);
SDI("Protected", FT_METHODTYPE_PROTECTED);
END_STRINGDICTIONARY(_METHODTYPETAGS, methodtypetags)

BEGIN_STRINGDICTIONARY(_CLASSTYPETAGS)
SDI("Interfaced", FT_CLASSTYPE_INTERFACED);
END_STRINGDICTIONARY(_CLASSTYPETAGS, classtypetags)

BEGIN_STRINGDICTIONARY(_HELPERFILETYPETAGS)
SDI("CPP", FT_HELPERFILETYPE_CPP);
SDI("H", FT_HELPERFILETYPE_H);
END_STRINGDICTIONARY(_HELPERFILETYPETAGS, helperfiletypetags)

BEGIN_STRINGDICTIONARY(_HELPERFILEDATATAGS)
SDI("Filename", FT_HELPERFILEDATA_FILENAME);
END_STRINGDICTIONARY(_HELPERFILEDATATAGS, helperfiledatatags)

BEGIN_STRINGDICTIONARY(_PROJECTDATATAGS)
SDI("Name", FT_PROJECT_NAME);
END_STRINGDICTIONARY(_PROJECTDATATAGS, projectdatatags)

BEGIN_STRINGDICTIONARY(_DATATYPETAGS)
SDI("Template", FT_DATATYPE_TEMPLATE);
SDI("Templatized", FT_DATATYPE_TEMPLATIZED);
SDI("Aliased", FT_DATATYPE_ALIASED);
SDI("Class", FT_DATATYPE_CLASS);
SDI("Pointer", FT_DATATYPE_PTR);
SDI("Reference", FT_DATATYPE_REF);
SDI("In", FT_DATATYPE_IN);
SDI("Out", FT_DATATYPE_OUT);
SDI("InOut", FT_DATATYPE_INOUT);
SDI("Script", FT_DATATYPE_SCRIPT);
END_STRINGDICTIONARY(_DATATYPETAGS, datatypetags)

#ifdef ODE_IDE
BEGIN_STRINGDICTIONARY(_PROJECTFOLDERDATATAGS)
SDI("Name", FT_PROJECTFOLDER_NAME);
SDI("Path", FT_PROJECTFOLDER_PATH);
END_STRINGDICTIONARY(_PROJECTFOLDERDATATAGS, projectfolderdatatags)

BEGIN_STRINGDICTIONARY(_PACKAGEDFILEDATATAGS)
SDI("Filename", FT_PACKAGEDFILE_DATA_FILE);
SDI("TargetDir", FT_PACKAGEDFILE_DATA_DIR);
SDI("ProjectFolder", FT_PACKAGEDFILE_DATA_FOLDER);
END_STRINGDICTIONARY(_PACKAGEDFILEDATATAGS, packagedfiledatatags)
#endif

extern String g_resourcepath;

TemplatizedFactory::TemplatizedFactory() {
  m_inherit_parallel = 0;
  m_chunk = NULL;
  m_template = NULL;
  m_ppercent = NULL;
  m_ext = "xpst";
  m_needfriend = 0;
  m_curdata = NULL;
  m_curdataoverride = NULL;
  m_running = 0;
}

TemplatizedFactory::~TemplatizedFactory() {
  delete m_template;
}

void TemplatizedFactory::setExt(const char *ext) {
  m_ext = ext;
}

void TemplatizedFactory::registerModule(FactoryModule *module) {
  m_modules.addItem(module);
}

void TemplatizedFactory::setName(const char *name) {
  m_name = name;
}

int TemplatizedFactory::tryFormat(const char *format) {
  return STRCASEEQLSAFE(format, m_name);
}

const char *TemplatizedFactory::getCustomFormat(const char *format) {
  return format;
}

const char *TemplatizedFactory::getCustomClassData(const char *data, XmlReaderParams *params) {
  return NULL;
}

int TemplatizedFactory::getCustomSkip() {
  return 0;
}

void TemplatizedFactory::onCustomTag(const char *tag, XmlReaderParams *cb, int isclose, int discard) {
}

ForIterator *TemplatizedFactory::getForIterator() {
  if (m_forstack.peek() == 0) return NULL;
  ForIterator *iter = NULL;
  m_forstack.peekAt(&iter, 0);
  return iter;
}

void TemplatizedFactory::addToCurData(const char *str) {
  if (m_curdata == NULL) return;
  //DebugString("+ <%s>\n", str);
  *m_curdata += str;
}

int TemplatizedFactory::continueByType(int type) {
  int n = m_forstack.peek();
  ForIterator *iter = NULL;
  int pos = 0;
  while (pos < n) {
    m_forstack.peekAt(&iter, pos);
    if (iter != NULL && iter->getType() == type) {
      if (iter->isLast()) {
        m_continueskip = pos+1;
        return 1;
      }
      else {
        iter->inc();
        if (m_ppercent) m_ppercent->setPercent((int)(((float)iter->getIterator() / (float)iter->getNumIterations()) * 100));
        m_templatepos = iter->getLoopPos();
      }
      return 0;
    }
    pos++;
  }
  return 0;
}

int TemplatizedFactory::breakByType(int type) {
  int n = m_forstack.peek();
  ForIterator *iter = NULL;
  int pos = 0;
  while (pos < n) {
    m_forstack.peekAt(&iter, pos);
    if (iter != NULL && iter->getType() == type) {
      m_continueskip = pos+1;
      return 1;
    }
    pos++;
  }
  return 0;
}

int TemplatizedFactory::testInherit(ClassFactoryChunk *chunk, const char *inherit) {
  HelperEnum en(chunk);
  int n = en.getNumHelpers();
  for (int i=0;i<n;i++) {
    // optim!
    XHDProcessor processor(0, chunk, en.enumHelper(i));
    if (STREQLSAFE(chunk->getAncestor(), inherit)) return 1;
  }
  return 0;
}

int TemplatizedFactory::testMatchesType(int fortype, const char *type, XmlReaderParams *params) {
  switch (fortype) {
    case FOREACH_METHOD: {
      int a = methodtypetags.getId(type);
      switch (a) {
        case FT_METHODTYPE_CTOR:
          return getChunk() && getCurrentMethod() && getCurrentMethod()->isConstructor();
        case FT_METHODTYPE_DEFCTOR:
          return getChunk() && getCurrentMethod() && getCurrentMethod()->isConstructor() && getCurrentMethod()->getNumParams() == 0;
        case FT_METHODTYPE_DTOR:
          return getChunk() && getCurrentMethod() && getCurrentMethod()->isDestructor();
        case FT_METHODTYPE_VIRTUAL:
          return getCurrentMethod() && getCurrentMethod()->isVirtual();
        case FT_METHODTYPE_STATIC:
          return getCurrentMethod() && getCurrentMethod()->isStatic();
        case FT_METHODTYPE_INLINE:
          return getCurrentMethod() && getCurrentMethod()->isInline();
        case FT_METHODTYPE_PURE: {
          int pure = 0;
          return getCurrentMethod() && getCurrentMethod()->isVirtual(&pure) && pure;
        }
        case FT_METHODTYPE_RAW:
          return getCurrentMethod() && !getCurrentMethod()->isVirtual() && !getCurrentMethod()->isStatic() && !getCurrentMethod()->isInline();
        case FT_METHODTYPE_NODISPATCH:
          return getCurrentMethod() && translateVariablesInt(getCurrentMethod()->getParams()->getItemValue("nodispatch")) == 1;
        case FT_METHODTYPE_DISPATCH:
          return (getCurrentMethod() && translateVariablesInt(getCurrentMethod()->getParams()->getItemValue("dispatch")) == 1 && translateVariablesInt(getCurrentMethod()->getParams()->getItemValue("nodispatch")) == 0);
        case FT_METHODTYPE_MULTIPLE: {
          ClassFactoryMethod *method = getCurrentMethod();
          if (!method) break;
          int pos = -1;
          int multiple = 0;
          getMultipleMethod(method, &pos, &multiple, NULL);
          return multiple;
        }
        case FT_METHODTYPE_FIRSTOFMULTIPLE: {
          ClassFactoryMethod *method = getCurrentMethod();
          if (!method) break;
          int pos = -1;
          int multiple = 0;
          int first = 0;
          getMultipleMethod(method, &pos, &multiple, &first);
          return multiple && first;
        }
        case FT_METHODTYPE_ISVOID:{
          ClassFactoryMethod *method = getCurrentMethod();
          if (method == NULL) return 0;
          return isVoid(method->getReturnType());
        }
        case FT_METHODTYPE_ISVOIDPARAM:{
          ClassFactoryMethod *method = getCurrentMethod();
          if (method == NULL) return 0;
          return method->getNumParams() == 0;
        }
        case FT_METHODTYPE_SCRIPT:{
          ClassFactoryMethod *method = getCurrentMethod();
          if (!method) return 0;
          return method->getParams()->getItemValueInt("scriptable");
        }
        case FT_METHODTYPE_EVENT:{
          ClassFactoryMethod *method = getCurrentMethod();
          if (!method) return 0;
          return method->getParams()->getItemValueInt("isevent");
        }
        case FT_METHODTYPE_PUBLIC:{
          ClassFactoryMethod *method = getCurrentMethod();
          if (!method) return 0;
          return method->getScope() == SCOPE_PUBLIC;
        }
        case FT_METHODTYPE_PRIVATE:{
          ClassFactoryMethod *method = getCurrentMethod();
          if (!method) return 0;
          return method->getScope() == SCOPE_PRIVATE;
        }
        case FT_METHODTYPE_PROTECTED:{
          ClassFactoryMethod *method = getCurrentMethod();
          if (!method) return 0;
          return method->getScope() == SCOPE_PROTECTED;
        }
        case FT_METHODTYPE_ISRETIRED: {
          ClassFactoryMethod *method = getCurrentMethod();
          return (!STRNICMP(method->getName(), "_RETIRED_", 9));
        }
        case FT_METHODTYPE_HASRETIREDVERSIONS: {
          ClassFactoryMethod *method = getCurrentMethod();
          if (!method) break;
          ClassFactoryChunk *chunk = method->getParentBlock()->getParentChunk();
          String name = method->getName();
          name = StringPrintf("_RETIRED_%s", name.getValue());
          for (int i=0;i<chunk->getNumBlocks();i++) {
            if (chunk->enumBlock(i)->getBlockType() == CF_BLOCKTYPE_METHOD) {
              ClassFactoryMethod *m = chunk->enumBlock(i)->getMethod();
              if (STRCASEEQL(name, m->getName()))
                return 1;
            }
          }
          return 0;
        }
      }
    }
    case FOREACH_PROJECTFOLDERS: {
      if (STRCASEEQL(type, "match")) {
        ForEachProjectFolder *fepf = getForEachProjectFolder();
        if (!fepf) return 0;
        String t = params->getItemValue("match");
        if (!Std::isDirChar(t.lastChar())) t += "/";
        return Std::match(t, fepf->getFolder());
      }
    }
    break;
    case FOREACH_TEMPLATEPARAM:
    case FOREACH_PARAM:
    case FOREACH_DATA: {
      String str;
      ClassFactoryParam *param = getCurrentTemplateParam();
      if (!param) param = getCurrentParam();
      ClassFactoryData *data = getCurrentData();
      if (param) str = param->getType();
      else if (data) str = data->getDataType();
      int a = datatypetags.getId(type);
      switch (a) {
        case FT_DATATYPE_TEMPLATE: {
          String templatename;
          PtrList<String> templateparams;
          parseTemplate(str, templatename, templateparams);
          templateparams.deleteAll();
          return (!templatename.isempty());
        }
        case FT_DATATYPE_TEMPLATIZED: {
          String templatename;
          PtrList<String> templateparams;
          parseTemplate(str, templatename, templateparams);
          if (templatename.isempty()) return 0;
          FactoryTemplate *tmpl = loadDataTemplate(templatename, 1);
          if (tmpl != NULL) {
            delete tmpl;
            return 1;
          }
          return 0;
        }
        case FT_DATATYPE_CLASS: {
          String cn;
          return isClass(str, cn);
        }
        case FT_DATATYPE_PTR:
          ClassFactoryParser::trimSpaces(str);
          return (str.lastChar() == '*');
        case FT_DATATYPE_REF:
          ClassFactoryParser::trimSpaces(str);
          return (str.lastChar() == '&');
        case FT_DATATYPE_ALIASED: {// todo: be much smarter
          String s;
          getAliasType(str, s);
          return (!STRCASEEQL(str, s));
        }
        case FT_DATATYPE_IN: {
          ClassFactoryParam *param = getCurrentTemplateParam();
          if (!param) param = getCurrentParam();
          if (!param) break;
          return (param->getInOut() == IO_IN);
        }
        case FT_DATATYPE_OUT: {
          ClassFactoryParam *param = getCurrentTemplateParam();
          if (!param) param = getCurrentParam();
          if (!param) break;
          return (param->getInOut() == IO_OUT);
        }
        case FT_DATATYPE_INOUT: {
          ClassFactoryParam *param = getCurrentTemplateParam();
          if (!param) param = getCurrentParam();
          if (!param) break;
          return (param->getInOut() == IO_INOUT);
        }
        case FT_DATATYPE_SCRIPT:{
          ClassFactoryData *data = getCurrentData();
          if (!data) return 0;
          return data->getParams()->getItemValueInt("scriptable");
        }
      }
    }
    break;
    case FOREACH_CLASS: {
      int a = classtypetags.getId(type);
      switch (a) {
        case FT_CLASSTYPE_INTERFACED: {
          ParamParser ppi(getChunk()->getParams()->getItemValue("xhd_factories"));
          return ppi.hasString("Interface");
        }
      }
    }
    break;
    case FOREACH_MAPENTRIES: {
      int a = mapentrytypetags.getId(type);
/*      switch (a) {
      }*/
    }
    break;
    case FOREACH_HELPERFILES: {
      int a = helperfiletypetags.getId(type);
      switch (a) {
        case FT_HELPERFILETYPE_H: {
          ForEachHelperFile *fehf = getForEachHelperFile();
          return fehf && STRCASESTR(fehf->getFile(), ".h");
        }
        case FT_HELPERFILETYPE_CPP: {
          ForEachHelperFile *fehf = getForEachHelperFile();
          return fehf && STRCASESTR(fehf->getFile(), ".cpp");
        }
      }
    }
    break;
  }
  return 0;
}

void TemplatizedFactory::setProgressCallback(ProgressCallback *cb) { m_ppercent = cb; }

const char *TemplatizedFactory::getName() {
  return m_name;
}

const char *TemplatizedFactory::getDescription() {
  if (!m_template) return "need call to factory->process() to get template description";
  return m_template->getDescription();
}

int TemplatizedFactory::getNumModules() {
  int total=0;
  foreach(m_modules)
    if (m_modules.getfor()->getTemplate() == NULL || m_modules.getfor()->getTemplate() == m_template) total++;
  endfor;
  return total;
}

const char *TemplatizedFactory::enumModuleName(int i) {
  int total=0;
  foreach(m_modules)
    if (m_modules.getfor()->getTemplate() == NULL || m_modules.getfor()->getTemplate() == m_template) {
      if (i == total) return m_modules.getfor()->getName();
      total++;
    }
  endfor;
  return NULL;
}

const char *TemplatizedFactory::enumModuleData(int i) {
  int total=0;
  foreach(m_modules)
    if (m_modules.getfor()->getTemplate() == NULL || m_modules.getfor()->getTemplate() == m_template) {
      if (i == total) return m_modules.getfor()->getData();
      total++;
    }
  endfor;
  return NULL;
}

int TemplatizedFactory::process() {
  if (wantAutoResetTemplates()) {
    foreach(m_auto_modules)
      m_modules.removeItem(m_auto_modules.getfor());
    endfor;
    m_auto_modules.deleteAll();
    m_procs.deleteAll();
  }
  String format = getCustomFormat(getName());
  if (!loadFormatTemplate(format)) return 0;
  if (!STRCASEEQLSAFE(format, m_template->getFormat())) return 0;
  if (!processFormat()) return 0;
  if (wantAutoResetTemplates()) {
    delete m_template;
    m_template = NULL;
  }
  if (m_inherit_parallel) {
    getChunk()->getParams()->replaceItem(StringPrintf("$_%s_inherit_parallel", format.getValue()), getChunk()->getAncestor());
    getChunk()->setAncestor(m_originalAncestor);
  } else
    getChunk()->getParams()->removeItem(StringPrintf("$_%s_inherit_parallel", format.getValue()));
  return 1;
}

void TemplatizedFactory::setChunk(ClassFactoryChunk *chunk) {
  resetFactory();
  m_chunk = chunk;
  m_originalAncestor = chunk ? chunk->getAncestor() : "";
  onNewChunk(m_chunk);
}

void TemplatizedFactory::setChunk_noreset(ClassFactoryChunk *chunk) {
  m_chunk = chunk;
}

ClassFactoryChunk *TemplatizedFactory::getChunk() {
  int classpos=-1;
  int ancestorpos=-1;
  ForEachClass *fec = static_cast<ForEachClass*>(getLastIteratorByType(FOREACH_CLASS, &classpos));
  ForEachAncestor *fea = static_cast<ForEachAncestor*>(getLastIteratorByType(FOREACH_ANCESTORS, &ancestorpos));
  if (fec && !fea) return fec->getClass();
  if (!fec && fea) return fea->getAncestor()->m_chunk;
  if (fec && fea) {
    if (classpos < ancestorpos) return fec->getClass();
    if (classpos > ancestorpos) return fea->getAncestor()->m_chunk;
    ASSERT(0);
  }
  return m_chunk;
}

void TemplatizedFactory::pushForeach(ForIterator *iter) {
  m_forstack.push(iter);
}

ForIterator *TemplatizedFactory::popForeach() {
  ForIterator *iter = NULL;
  m_forstack.pop(&iter);
  return iter;
}

ForIterator *TemplatizedFactory::createForEach(const char *objects, XmlReaderParams *params, ClassFactoryChunk *chunk, ClassFactoryMethod *method, ClassFactoryParam *param, String &txt) {
  switch (foreachobjecttags.getId(objects)) {
    case FOREACH_METHOD: {
      String filter = params->getItemValue("mask");
      ForEachMethod *m = new ForEachMethod(this, getChunk(), filter);
      return m;
    }
    case FOREACH_MAPENTRIES: {
      String map = params->getItemValue("map");
      ForEachMapEntries *m = new ForEachMapEntries(this, getChunk(), map);
      return m;
    }
    case FOREACH_RETIREDMETHOD: {
      ClassFactoryMethod *method = getCurrentMethod();
      if (!method) return NULL;
      return new ForEachRetiredMethod(this, getChunk(), method);
    }
    case FOREACH_CODEBLOCK:
      return new ForEachCodeBlock(this, getChunk(), translateVariables(params->getItemValue("type")));
    case FOREACH_PARAM: {
      ClassFactoryMethod *method = getCurrentMethod();
      if (method == NULL) {
        txt = "*** <foreach objects=\"params\"/> tag found outside a <foreach object=\"methods\"/> ***";
        return NULL;
      }
      return new ForEachParam(this, method);
    }
    case FOREACH_TEMPLATEPARAM: {
      ClassFactoryParam *param = getCurrentParam();
      ClassFactoryData *data = getCurrentData();
      if (data) return new ForEachTemplateParam(this, data);
      if (param) return new ForEachTemplateParam(this, param);
      return NULL;
    }
    case FOREACH_DATA:
      return new ForEachData(this, getChunk());
    case FOREACH_ANCESTORS:
      return new ForEachAncestor(this, getChunk());
    case FOREACH_CLASS:
      return new ForEachClass(this);
    case FOREACH_IMPLFILES:
      return new ForEachImplFile(this, getChunk());
    case FOREACH_HELPERS:
      return new ForEachHelper(this, getChunk());
    case FOREACH_HELPERFILES: {
      ForEachHelper *feh = getForEachHelper();
      if (!feh) {
        txt = "*** <foreach objects=\"helperfiles\"/> tag found outside a <foreach objects=\"helpers\"/> ***";
        return NULL;
      }
      return new ForEachHelperFile(this, getChunk(), feh->getIterator());
    }
    case FOREACH_PROJECTFOLDERS: {
      String root = translateVariables(params->getItemValue("root"));
      String s = params->getItemValue("level");
      int nextlevel = s.iscaseequal("next");
      int thislevel = s.iscaseequal("root");
      return new ForEachProjectFolder(this, root, nextlevel, thislevel);
    }
#ifdef ODE_IDE
    case FOREACH_GROUPDEFFOLDERS: {
      String root = translateVariables(params->getItemValue("root"));
      String s = params->getItemValue("level");
      int nextlevel = s.iscaseequal("next");
      int thislevel = s.iscaseequal("root");
      return new ForEachGroupDefFolder(this, root, nextlevel, thislevel);
    }
    case FOREACH_PACKAGEDFILES:
      return new ForEachPackagedFile(this);
#endif
  }
  return NULL;
}

ForIterator *TemplatizedFactory::getLastIteratorByType(int type, int *_pos) {
  int n = m_forstack.peek();
  ForIterator *iter = NULL;
  int pos = 0;
  while (pos < n) {
    m_forstack.peekAt(&iter, pos);
    if (iter != NULL && iter->getType() == type) {
      if (_pos != NULL) *_pos = pos;
      return iter;
    }
    pos++;
  }
  return NULL;
}

ForIterator *TemplatizedFactory::getLastIterator() {
  if (m_forstack.peek() == 0) return NULL;
  ForIterator *iter = NULL;
  m_forstack.peekAt(&iter, 0);
  return iter;
}

void TemplatizedFactory::resetFactory() {
  delete m_template;
  m_template = NULL;
  m_chunk = NULL;
  m_ppercent = NULL;
  m_modules.removeAll();
  m_auto_modules.deleteAll();
  m_procs.deleteAll();
}

int TemplatizedFactory::needFriend() {
  return m_needfriend;
}

ForEachMethod *TemplatizedFactory::getForEachMethod() {
  return static_cast<ForEachMethod*>(getLastIteratorByType(FOREACH_METHOD));
}

ForEachMapEntries *TemplatizedFactory::getForEachMapEntries() {
  return static_cast<ForEachMapEntries*>(getLastIteratorByType(FOREACH_MAPENTRIES));
}

ForEachParam *TemplatizedFactory::getForEachParam() {
  return static_cast<ForEachParam*>(getLastIteratorByType(FOREACH_PARAM));
}

ForEachData *TemplatizedFactory::getForEachData() {
  return static_cast<ForEachData *>(getLastIteratorByType(FOREACH_DATA));
}

ForEachCodeBlock *TemplatizedFactory::getForEachCodeBlock() {
  return static_cast<ForEachCodeBlock*>(getLastIteratorByType(FOREACH_CODEBLOCK));
}

ForEachClass *TemplatizedFactory::getForEachClass() {
  return static_cast<ForEachClass*>(getLastIteratorByType(FOREACH_CLASS));
}

ForEachHelper *TemplatizedFactory::getForEachHelper() {
  return static_cast<ForEachHelper *>(getLastIteratorByType(FOREACH_HELPERS));
}

ForEachHelperFile *TemplatizedFactory::getForEachHelperFile() {
  return static_cast<ForEachHelperFile *>(getLastIteratorByType(FOREACH_HELPERFILES));
}

ForEachAncestor *TemplatizedFactory::getForEachAncestor() {
  return static_cast<ForEachAncestor *>(getLastIteratorByType(FOREACH_ANCESTORS));
}

ForEachImplFile *TemplatizedFactory::getForEachImplFile() {
  return static_cast<ForEachImplFile *>(getLastIteratorByType(FOREACH_IMPLFILES));
}

ForEachProjectFolder *TemplatizedFactory::getForEachProjectFolder() {
  return static_cast<ForEachProjectFolder *>(getLastIteratorByType(FOREACH_PROJECTFOLDERS));
}

#ifdef ODE_IDE
ForEachGroupDefFolder *TemplatizedFactory::getForEachGroupDefFolder() {
  return static_cast<ForEachGroupDefFolder *>(getLastIteratorByType(FOREACH_GROUPDEFFOLDERS));
}
#endif

ClassFactoryCodeBlock *TemplatizedFactory::getCurrentCodeBlock() {
  ForIterator *iter = getLastIteratorByType(FOREACH_CODEBLOCK);
  if (iter) return static_cast<ForEachCodeBlock*>(iter)->getCodeBlock();
  return NULL;
}

ClassFactoryMethod *TemplatizedFactory::getCurrentMethod() {
  ForIterator *iter = getLastIteratorByType(FOREACH_METHOD);
  if (iter) return static_cast<ForEachMethod*>(iter)->getMethod();
  return getRootMethod();
}

int TemplatizedFactory::getCurrentMapEntry() {
  ForIterator *iter = getLastIteratorByType(FOREACH_MAPENTRIES);
  if (iter) return iter->getIterator();
  return -1;
}

ClassFactoryParam *TemplatizedFactory::getCurrentParam() {
  ForIterator *iter = getLastIteratorByType(FOREACH_PARAM);
  if (iter) return static_cast<ForEachParam*>(iter)->getParam();
  return NULL;
}

ClassFactoryParam *TemplatizedFactory::getCurrentTemplateParam() {
  ForIterator *iter = getLastIteratorByType(FOREACH_TEMPLATEPARAM);
  if (iter) return static_cast<ForEachTemplateParam*>(iter)->getParam();
  return NULL;
}

ClassFactoryData *TemplatizedFactory::getCurrentData() {
  if (m_curdataoverride) return m_curdataoverride;
  ForIterator *iter = getLastIteratorByType(FOREACH_DATA);
  if (iter) return static_cast<ForEachData*>(iter)->getData();
  return NULL;
}

#ifdef ODE_IDE
ProjectFile *TemplatizedFactory::getCurrentPackagedFile() {
  ForIterator *iter = getLastIteratorByType(FOREACH_PACKAGEDFILES);
  if (iter) return static_cast<ForEachPackagedFile*>(iter)->getFile();
  return NULL;
}
#endif

int TemplatizedFactory::getDiscardCondition() {
  if (m_continueskip > 0) return 1;
  if (m_skip > 0) return 1;
  if (m_control_discard) return 1;
  if (getCustomSkip()) return 1;
  int n = m_controlstack.peek();
  int pos = 0;
  int discard;
  while (pos < n) {
    m_controlstack.peekAt(&discard, pos);
    if (discard) return 1;
    pos++;
  }
  return 0;
}

void TemplatizedFactory::xmlReaderOnStartElementCallback(const char *xmlpath, const char *xmltag, XmlReaderParams *params) {
  int a = fttags.getId(xmltag);
  if (a == FT_ROOT) {
    if (m_intemplate == 0) {
      if (getChunk() && STRCASEEQLSAFE(xmltag, "XPP:StandaloneTemplate")) {
        String inherit = translateVariables(params->getItemValue("inherit"));
        String cn = translateVariables(params->getItemValue("classname"));
        m_inherit_parallel = 0;
        if (!cn.isempty()) {
          m_inherit_parallel = !inherit.iscaseequal("serial");
          char *p = STRCHR(cn, '*');
          if (p != NULL) {
            String s;
            s.catn(cn, p-cn.getValue());
            if (getChunk()) s.cat(getChunk()->getClassName());
            s.cat(p+1);
            getChunk()->setAncestor(s);
          }
        }
      }
      int a = translateVariablesInt(params->getItemValue("friend"), -1);
      if (a != -1) m_needfriend = a;
      m_template->setDescription(params->getItemValue("desc"));
    }
    m_intemplate++;
    return;
  }
  if (a == FT_MODULE) {
    m_parserinmodule++;
    if (!m_curformat.iscaseequal("[runvar]")) {
      String moduleid = translateVariables(params->getItemValue("id"));
      String moduledefname = translateVariables(params->getItemValue("filename"));
      FactoryModule *module = getModule(moduleid, m_template);
      if (module == NULL) {
        module = new FactoryModule();
        String name;
        String savename = StringPrintf("%s_%s", m_curformat.getValue(), moduleid.getValue());
        name = translateVariables(m_chunk->getParams()->getItemValue(savename));
        if (name.isempty()) {
          char *p = STRCHR(moduledefname, '*');
          if (p != NULL) {
            String s;
            s.catn(moduledefname, p-moduledefname);
            if (m_chunk) s.cat(m_chunk->getClassName());
            s.cat(p+1);
            name = s;
            if (m_chunk) m_chunk->getParams()->replaceItem(savename, name);
          }
        }
        module->setName(name);
        module->setId(moduleid);
        module->setSubModuleOf(m_template);
        m_modules.addItem(module);
        m_auto_modules.addItem(module);
      }
    }
  }
  if (!m_parserinmodule && a == FT_PROC) {
    m_parserinmodule++;
    FactoryModule *fm = new FactoryModule();
    fm->setId(params->getItemValue("id"));
    fm->setStartPos(m_curTemplate->getNumChunks());
    m_procs.addItem(fm);
  }
  if (!m_parserinmodule && a == FT_TYPEMAP) {
    if (m_curTypeMap == NULL) {
      const char *id = translateVariables(params->getItemValue("id"));
      m_curTypeMap = g_typeMaps->getTypeMap(id);
      if (!m_curTypeMap) {
        m_curTypeMap = new TypeMap(this, id);
        g_typeMaps->addTypeMap(m_curTypeMap);
      }
    }
  }
  if (!m_parserinmodule && a == FT_MAP && m_curTypeMap != NULL) {
    String from = translateVariables(params->getItemValue("from"));
    String to = translateVariables(params->getItemValue("to"));
    String impl = translateVariables(params->getItemValue("to"));
    String _default =translateVariables(params->getItemValue("default"));
    if (_default.isempty())
      m_curTypeMap->addMapping(from, to, impl);
    else
      m_curTypeMap->setDefaultMapping(_default, impl);
  }
  if (!m_parserinmodule && a == FT_TYPECAST) {
    String from = translateVariables(params->getItemValue("from"));
    String to = translateVariables(params->getItemValue("to"));
    String cast = translateVariables(params->getItemValue("cast"));
    int ismacro = translateVariablesInt(params->getItemValue("ismacro"));
    g_casts->addTypeCast(from, to, cast, ismacro);
  }
  if (!m_intemplate) return;
  if (m_parserinmodule) {
    TemplateChunkI *chunk = new TemplateChunkI(0, a, xmltag, params, NULL, xmlGetFileName(), xmlGetIncludePath(), xmlGetFileLine());
    m_curTemplate->addChunk(chunk);
  }
}

void TemplatizedFactory::xmlReaderOnEndElementCallback(const char *xmlpath, const char *xmltag) {
  int a = fttags.getId(xmltag);
  if (a == FT_ROOT) { m_intemplate--; return; }
  if (a == FT_TYPEMAP && !m_parserinmodule) { m_curTypeMap = NULL; }
  if (a == FT_MODULE || a == FT_PROC) {
    m_parserinmodule--;
  }
  if (!m_intemplate) return;
  TemplateChunkI *chunk = new TemplateChunkI(1, a, xmltag, NULL, NULL, xmlGetFileName(), xmlGetIncludePath(), xmlGetFileLine());
  m_curTemplate->addChunk(chunk);
}

void TemplatizedFactory::xmlReaderOnCharacterDataCallback(const char *xmlpath, const char *xmltag, const char *str) {

  if (!m_intemplate) return;

  char *s = STRDUP(str);
  char *p = s;
  int done = 0;
  int hadleadingcr = 0;

  // eat all leading empty lines first, don't discard lines without CR
  if (*p) {
    char *lastcr = s-1;
    while (*p) {
      if (*p != ' ' && *p != '\r' && *p != '\n') {
        p = lastcr+1;
        hadleadingcr = 1;
        break;
      } else if (*p == '\r' || *p == '\n') lastcr = p;
      p++;
    }
    if (lastcr == s-1) p = s;
  }

  char *t = p;

  if (*p) {
    while (p && p >= t && *p && !done) {
      // find first non-space/cr char
      if (*p != ' ' && *p != '\t' && *p != '\n' && *p != '\r') {
        // go back to last cr or beginning of string
        while (p >= s)
          if (*p != '\r' && *p != '\n')
            p--;
          else
            break;
        break;
      }
      p--;
    }
    p++;
  }

  t = p;
  p = t+STRLEN(t)-1;
  while (p >= t) {
    if (*p != ' ' && *p != '\t' && *p != '\n' && *p != '\r') {
      p++;
      break;
    }
    p--;
  }
  if (p < t) {
    // not found
    p = t;
    *p = 0;
  }
  while (*p) {
    if (*p == '\r' || *p == '\n') {
      *p = 0;
      break;
    }
    p++;
  }

  // if there was a CR at the beginning of the string, ie:
  // <tag/>
  //     txt<tag/>
  // ... then we need to eliminate leading spaces
  //
  // if there was no CR (ie: <tag/> <tag/>) then we should keep them
  if (hadleadingcr) {
    while (*t && *t == ' ') t++;
  }

  TemplateChunkI *chunk = NULL;
  if (*t) {
    chunk = new TemplateChunkI(0, FT_CDATA, "CDATA", NULL, t, xmlGetFileName(), xmlGetIncludePath(), xmlGetFileLine());
    m_curTemplate->addChunk(chunk);
  }

  FREE(s);
}

void TemplatizedFactory::unloadAllTemplates() {
  delete m_template;
  m_template = NULL;
}

int TemplatizedFactory::processFormat() {

  int l_return = 0;

  if (!m_running) {
    m_tempTemplateChunk = NULL;
    m_curdata = NULL;
    m_curdataoverride = NULL;
    m_templatepos = 0;
    m_gotcr = 0;
    m_skip = 0;
    m_continueskip = 0;
    m_inelse = 0;
    m_curmoduledesc = NULL;
    m_curscope = SCOPE_NEUTRAL;
    m_control_discard = 0;
    m_curTemplate = m_template;
  } else {
    m_gotcrstack.push(m_gotcr);
    m_gotcr = 0;
  }

  m_running++;
  ClassFactoryParam *param = NULL;
  String val;

  int instruction_count = 0;
  n_template_chunks = m_curTemplate->getNumChunks();
  while (!l_return) {
    if ((instruction_count++ % 5) == 0) xpp_updateProgress();
    while (!l_return && m_templatepos < n_template_chunks) {
      //int setcr = 0;
      val.trunc(0);
      TemplateChunk *templatechunk = m_curTemplate->enumChunk(m_templatepos);
      m_tempTemplateChunk = templatechunk;
      int isclose = templatechunk->isClosingTag();
      int xmltagtype = templatechunk->getType();
      int discard_condition = getDiscardCondition();
      if (discard_condition && !(xmltagtype == FT_FOREACH || xmltagtype == FT_IF || xmltagtype == FT_ELSE || xmltagtype == FT_UNKNOWN || xmltagtype == FT_PROC)) {
        m_templatepos++;
        continue;
      }
      switch (xmltagtype) {
        case FT_UNKNOWN:
          onCustomTag(templatechunk->getName(), templatechunk->getParams(), isclose, discard_condition);
          break;
        case FT_CDATA:
          if (isclose) break;
          if (getCustomSkip()) break;
          val = templatechunk->getCData();
          break;
        case FT_BR:
          if (isclose) break;
          val = CRLF;
          //setcr = 1;
          break;
        case FT_OB:
          if (isclose) break;
          val = OBRC;
          break;
        case FT_CB:
          if (isclose) break;
          val = CBRC;
          break;
        case FT_DATETIME:
          if (isclose) break;
          time_t now;
          time(&now);
          struct tm *ntime;
          ntime = localtime(&now);
          val = asctime(ntime);
          if (val.lastChar() == '\n') val.trunc(val.len()-1);
          break;
        case FT_MODULE:
          if (!isclose) {
            m_outputstack.push(m_curdata);
            m_moduledescstack.push(m_curmoduledesc);
            m_curdata = new BigString();
            m_curmoduledesc = new ModuleDescription();
            m_curmoduledesc->m_moduleId = translateVariables(templatechunk->getParams()->getItemValue("id"));
            m_curmoduledesc->m_moduleDefName = translateVariables(templatechunk->getParams()->getItemValue("filename"));
            FactoryModule *module = NULL;
            module = getModule(m_curmoduledesc->m_moduleId, m_curTemplate);
            if (module == NULL) {
              module = new FactoryModule();
              String name;
              String savename = StringPrintf("%s_%s", m_curformat.getValue(), m_curmoduledesc->m_moduleId.getValue());
              XmlReaderParams *par = templatechunk->getParams();
              name = translateVariables(getChunk()->getParams()->getItemValue(savename));
              if (name.isempty()) {
                char *p = STRCHR(m_curmoduledesc->m_moduleDefName, '*');
                if (p != NULL) {
                  String s;
                  s.catn(m_curmoduledesc->m_moduleDefName, p-m_curmoduledesc->m_moduleDefName);
                  if (getChunk()) s.cat(getChunk()->getClassName());
                  s.cat(p+1);
                  name = s;
                  if (!m_curformat.iscaseequal("[runvar]")) {
                    getChunk()->getParams()->replaceItem(savename, name);
                  }
                }
              }
              module->setName(name);
              module->setId(m_curmoduledesc->m_moduleId);
              module->setSubModuleOf(m_curTemplate);
              m_modules.addItem(module);
              m_auto_modules.addItem(module);
            }
          } else {
            // set data !
            FactoryModule *mod = getModule(m_curmoduledesc->m_moduleId, m_curTemplate);
            if (mod != NULL)
                mod->setData(*m_curdata);
            delete m_curdata;
            delete m_curmoduledesc;
            m_curdata = NULL;
            m_outputstack.pop(&m_curdata);
            m_curmoduledesc = NULL;
            m_moduledescstack.pop(&m_curmoduledesc);
          }
          break;
        case FT_MODULEDATA:
          if (!isclose) {
            FactoryModule *module = NULL;
            int j;
            int jn = m_modules.getNumItems();
            for (j=0;j<jn;j++) {
              if (STRCASEEQL(m_curmoduledesc->m_moduleId, m_modules.enumItem(j)->getId())) {
                module = m_modules.enumItem(j);
                break;
              }
            }
            if (!module) break;
            const char *data = translateVariables(templatechunk->getParams()->getItemValue("data"));
            int b = classmoduledatatags.getId(data);
            switch (b) {
              case FT_MODULEDATA_FILENAME:
                val = module->getName();
                break;
              case FT_MODULEDATA_ID:
                val = module->getId();
                break;
            }
          }
          break;
        case FT_CLASSDECL: {
          if (!isclose) {
            String classname = translateVariables(templatechunk->getParams()->getItemValue("name"));
            String implementationname = translateVariables(templatechunk->getParams()->getItemValue("implname"));
            String exclude = translateVariables(templatechunk->getParams()->getItemValue("exclude"));
            String factory = translateVariables(templatechunk->getParams()->getItemValue("factory"));
            String inherit = translateVariables(templatechunk->getParams()->getItemValue("inherit"));
            if (implementationname.isempty()) implementationname = classname;

            for (int d=0;d<templatechunk->getParams()->getNbItems();d++) {
              const char *n = templatechunk->getParams()->getItemName(d);
              if (STRCASEEQL(n, "name")) continue;
              if (STRCASEEQL(n, "implname")) continue;
              if (STRCASEEQL(n, "exclude")) continue;
              if (STRCASEEQL(n, "inherit")) continue;
              if (STRCASEEQL(n, "factory")) continue;
              getChunk()->getParams()->replaceItem(n, templatechunk->getParams()->getItemValue(d));
            }

            if (templatechunk->getParams()->getItemValue("factory") != NULL)
              getChunk()->getParams()->replaceItem("xhd_factories", factory);

            val = "class ";
            val += implementationname;

            PtrList<ancestordef> ancestors;

            String serialancestor = getChunk()->getAncestor();
            if (!serialancestor.isempty()) ancestors.addItem(new ancestordef(serialancestor, "public"));

            int i;
            ParamParser ip(inherit);

            while (getChunk()->getNumSecondaryAncestors()>0)
              getChunk()->removeSecondaryAncestorNum(0);

            for (i=0;i<ip.getNumItems();i++) {
              ParamParser ipp(ip.enumItem(i), " ");
              if (ipp.getNumItems() > 1) {
                ancestors.addItem(new ancestordef(ipp.enumItem(1), ipp.enumItem(0)));
                int scope = SCOPE_PUBLIC;
                if (STRCASEEQLSAFE(ipp.enumItem(0), "public")) scope = SCOPE_PUBLIC;
                else if (STRCASEEQLSAFE(ipp.enumItem(0), "private")) scope = SCOPE_PRIVATE;
                else if (STRCASEEQLSAFE(ipp.enumItem(0), "protected")) scope = SCOPE_PROTECTED;
                getChunk()->addSecondaryAncestor(ipp.enumItem(1), scope);
              } else if (ipp.getNumItems() == 1) {
                ancestors.addItem(new ancestordef(ip.enumItem(i), "public"));
                getChunk()->addSecondaryAncestor(ip.enumItem(i), SCOPE_PUBLIC);
              }
            }

            HelperEnum en(getChunk());

            for (int h=0;h<en.getNumHelpers();h++) {
              String helper = en.enumHelper(h);
              String parallel_ancestor = getChunk()->getParams()->getItemValue(StringPrintf("$_%s_inherit_parallel", helper.getValue()));
              if (!parallel_ancestor.isempty()) ancestors.addItem(new ancestordef(parallel_ancestor, "public"));
            }

            ParamParser pp(exclude);
            for (i=0;i<pp.getNumItems();i++) {
              for (int j=0;j<ancestors.getNumItems();j++) {
                if (STRCASEEQL(ancestors.enumItem(j)->name.getValue(), pp.enumItem(i))) {
                  ancestors.removeByPos(j);
                  j--;
                }
              }
            }
            if (ancestors.getNumItems() > 0) val += " :";
            int col = getLastLinePos(val);
            foreach(ancestors)
              if (foreach_index > 0) {
                val += ",";
                if (foreach_index < ancestors.getNumItems()) {
                  val += CRLF;
                  int n = col;
                  while (n) { val += " "; n--; }
                }
              }

              String scope = ancestors.getfor()->scope;
              if (scope.isempty()) val += " public";
              else { val += " "; val += scope; }
              val += " ";
              val += ancestors.getfor()->name;
            endfor
            ancestors.deleteAll();
            val += openBrace();
            val += CRLF;
          }
        break;
        }
      case FT_GUID:
        if (!isclose) {
          if (getChunk()) {
            String guidid = translateVariables(templatechunk->getParams()->getItemValue("id"));
            if (guidid.isempty()) break;
            String classGuid = getChunk()->getParams()->getItemValue(StringPrintf("guid_%s", guidid.getValue()));
            GUID cguid;
            char sguid[256];
            if (classGuid.isempty()) {
              // not yet set, make one up
              nsGUID::createGuid(&cguid);
              nsGUID::toChar(cguid, sguid);
              classGuid = sguid;
              getChunk()->getParams()->replaceItem(StringPrintf("guid_%s", guidid.getValue()), classGuid);
            } else {
              cguid = nsGUID::fromChar(classGuid);
            }
            nsGUID::toChar(cguid, sguid);
            char *_sguid=sguid+1;
            if (*_sguid) _sguid[STRLEN(_sguid)-1]=0;
            val = StringPrintf("%s", _sguid);
          }
        }
        break;
      case FT_CGUID:
        if (!isclose) {
          if (getChunk()) {
            String guidid = translateVariables(templatechunk->getParams()->getItemValue("id"));
            int flat = translateVariablesInt(templatechunk->getParams()->getItemValue("flat"));
            if (guidid.isempty()) break;
            String classGuid = getChunk()->getParams()->getItemValue(StringPrintf("guid_%s", guidid.getValue()));
            GUID cguid;
            char sguid[256];
            if (classGuid.isempty()) {
              // not yet set, make one up
              nsGUID::createGuid(&cguid);
              nsGUID::toChar(cguid, sguid);
              classGuid = sguid;
              getChunk()->getParams()->replaceItem(StringPrintf("guid_%s", guidid.getValue()), classGuid);
            } else {
              cguid = nsGUID::fromChar(classGuid);
            }
            val = StringPrintf("%s0x%08x, 0x%04x, 0x%04x, %s0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, %2d, %2d%s",
                                flat ? "" : "{ ",
                                cguid.Data1,
                                cguid.Data2,
                                cguid.Data3,
                                flat ? "" : "{ ",
                                cguid.Data4[0],
                                cguid.Data4[1],
                                cguid.Data4[2],
                                cguid.Data4[3],
                                cguid.Data4[4],
                                cguid.Data4[5],
                                cguid.Data4[6],
                                cguid.Data4[7],
                                flat ? "" : " } }");
          }
        }
        break;
        case FT_CLASS: {
          if (isclose) break;
          String data = translateVariables(templatechunk->getParams()->getItemValue("data"));
          switch (classdatatags.getId(data)) {
            case FT_CLASS_NAME:
              if (getChunk()) {
                val = getChunk()->getClassName();
              }
              break;
            case FT_CLASS_GUID:
              if (getChunk()) {
                String classGuid = getChunk()->getParams()->getItemValue("class_guid");
                GUID cguid;
                char sguid[256];
                if (classGuid.isempty()) {
                  // not yet set, make one up
                  nsGUID::createGuid(&cguid);
                  nsGUID::toChar(cguid, sguid);
                  classGuid = sguid;
                  getChunk()->getParams()->replaceItem("class_guid", classGuid);
                } else {
                  cguid = nsGUID::fromChar(classGuid);
                }
                nsGUID::toChar(cguid, sguid);
                char *_sguid=sguid+1;
                if (*_sguid) _sguid[STRLEN(_sguid)-1]=0;
                val = StringPrintf("%s", _sguid);
              }
              break;
            case FT_CLASS_CGUID:
              if (getChunk()) {
                String classGuid = getChunk()->getParams()->getItemValue("class_guid");
                int flat = translateVariablesInt(templatechunk->getParams()->getItemValue("flat"));
                GUID cguid;
                char sguid[256];
                if (classGuid.isempty()) {
                  // not yet set, make one up
                  nsGUID::createGuid(&cguid);
                  nsGUID::toChar(cguid, sguid);
                  classGuid = sguid;
                  getChunk()->getParams()->replaceItem("class_guid", classGuid);
                } else {
                  cguid = nsGUID::fromChar(classGuid);
                }
                val = StringPrintf("%s0x%08x, 0x%04x, 0x%04x, %s0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, %2d, %2d%s",
                                    flat ? "" : "{ ",
                                    cguid.Data1,
                                    cguid.Data2,
                                    cguid.Data3,
                                    flat ? "" : "{ ",
                                    cguid.Data4[0],
                                    cguid.Data4[1],
                                    cguid.Data4[2],
                                    cguid.Data4[3],
                                    cguid.Data4[4],
                                    cguid.Data4[5],
                                    cguid.Data4[6],
                                    cguid.Data4[7],
                                    flat ? "" : " } }");
              }
              break;
            case FT_CLASS_LAYERGUID:
              if (getChunk()) {
                String layer = getChunk()->getAncestor();
                String classGuid = getChunk()->getParams()->getItemValue(StringPrintf("%s_guid", layer.getValue()));
                GUID cguid;
                char sguid[256];
                if (classGuid.isempty()) {
                  // not yet set, make one up
                  nsGUID::createGuid(&cguid);
                  nsGUID::toChar(cguid, sguid);
                  classGuid = sguid;
                  getChunk()->getParams()->replaceItem(StringPrintf("%s_guid", layer.getValue()), classGuid);
                } else {
                  cguid = nsGUID::fromChar(classGuid);
                }
                nsGUID::toChar(cguid, sguid);
                char *_sguid=sguid+1;
                if (*_sguid) _sguid[STRLEN(_sguid)-1]=0;
                val = StringPrintf("%s", _sguid);
              }
              break;
            case FT_CLASS_LAYERCGUID:
              if (getChunk()) {
                String layer = getChunk()->getAncestor();
                String classGuid = getChunk()->getParams()->getItemValue(StringPrintf("%s_guid", layer.getValue()));
                int flat = translateVariablesInt(templatechunk->getParams()->getItemValue("flat"));
                GUID cguid;
                char sguid[256];
                if (classGuid.isempty()) {
                  // not yet set, make one up
                  nsGUID::createGuid(&cguid);
                  nsGUID::toChar(cguid, sguid);
                  classGuid = sguid;
                  getChunk()->getParams()->replaceItem(StringPrintf("%s_guid", layer.getValue()), classGuid);
                } else {
                  cguid = nsGUID::fromChar(classGuid);
                }
                val = StringPrintf("%s0x%08x, 0x%04x, 0x%04x, %s0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, %2d, %2d%s",
                                    flat ? "" : "{ ",
                                    cguid.Data1,
                                    cguid.Data2,
                                    cguid.Data3,
                                    flat ? "" : "{ ",
                                    cguid.Data4[0],
                                    cguid.Data4[1],
                                    cguid.Data4[2],
                                    cguid.Data4[3],
                                    cguid.Data4[4],
                                    cguid.Data4[5],
                                    cguid.Data4[6],
                                    cguid.Data4[7],
                                    flat ? "" : " } }");
              }
              break;
            case FT_CLASS_IMPLNAME:
              if (getChunk()) {
                val = getChunk()->getImplementationName();
              }
              break;
            case FT_CLASS_CLASSLAYER:
              if (getChunk()) {
                val = getClassLayer();
              }
              break;
            case FT_CLASS_LAYERANCESTOR: {
              if (isclose) break;
              val = m_originalAncestor;
              break;
            }
            case FT_CLASS_NUMMETHODS: {
              if (isclose) break;
              val = StringPrintf("%d", ForEachMethod(this, getChunk(), NULL).getNumMethods());
              break;
            }
            case FT_CLASS_INTERFACEH: {
              if (getChunk()) {
                HelperFileEnum ife(getChunk(), "Interface");
                if (ife.getNumFiles() > 0)
                  val = ife.enumFile(0);
              }
              break;
            }
            case FT_CLASS_IMPLEMENTATIONCPP: {
              if (getChunk()) {
                ImplementationFactory fact;
                fact.setWorkspace(g_workspace);
                fact.setChunk(getChunk());
                val = fact.enumModuleName(1);
                PathParser pp(g_outputDir);
                for (int i=0;i<pp.getNumStrings();i++)
                  val.prepend("../");
              }
              break;
            }
            case FT_CLASS_TARGETDIR: {
              if (getChunk()) {
                String dir = getChunk()->getParams()->getItemValue("target_dir");
                if (dir.isempty()) dir = ".";
                if (Std::isDirChar(dir.lastChar())) dir.trunc(dir.len()-1);
                dir.replace("\\", "/");
                val = dir;
              }
              break;
            }
            case FT_CLASS_IMPLEMENTATIONH: {
              if (getChunk()) {
                ImplementationFactory fact;
                fact.setWorkspace(g_workspace);
                fact.setChunk(getChunk());
                val = fact.enumModuleName(0);
                PathParser pp(g_outputDir);
                for (int i=0;i<pp.getNumStrings();i++)
                  val.prepend("../");
              }
              break;
            }
            case FT_CLASS_INTERFACECPP: {
              if (getChunk()) {
                HelperFileEnum ife(getChunk(), "Interface");
                val = ife.enumFile(1);
              }
              break;
            }
            case FT_CLASS_DISPATCHH: {
              if (getChunk()) {
                HelperFileEnum ife(getChunk(), "Interface");
                val = ife.enumFile(2);
              }
              break;
            }
            case FT_CLASS_DISPATCHCPP: {
              if (getChunk()) {
                HelperFileEnum ife(getChunk(), "Interface");
                val = ife.enumFile(3);
              }
              break;
            }
            case -1:
              val = getCustomClassData(data, templatechunk->getParams());
              break;
          }
          break;
        }
        case FT_IMPLEMENTATION: {
          if (isclose) break;
          String data = translateVariables(templatechunk->getParams()->getItemValue("class"));
          ClassFactoryChunk *chunk = findChunk(data);
          if (chunk != NULL) {
            val = chunk->getImplementationName();
          }
        }
        break;
        case FT_INTERFACE: {
          if (isclose) break;
          String data = translateVariables(templatechunk->getParams()->getItemValue("class"));
          ClassFactoryChunk *chunk = findChunk(data);
          if (chunk != NULL) {
            val = chunk->getClassName();
          }
        }
        break;
        case FT_RESETSCOPE:
          if (isclose) break;
          m_curscope = SCOPE_NEUTRAL;
          break;
        case FT_CODEBLOCK: {
          if (isclose) break;
          ClassFactoryCodeBlock *codeblock = getCurrentCodeBlock();
          if (codeblock == NULL) {
            val = "*** <codeblock/> tag found outside a <foreach objects=\"codeblocks\"/> ***";
            break;
          }
          String data = translateVariables(templatechunk->getParams()->getItemValue("data"));
          switch (codeblockdatatags.getId(data)) {
            case FT_CODEBLOCK_SCOPE: {
              String x = codeblock->getCodeBlock();
              int sc = m_curscope;
              if (!x.isempty()) {
                processInlineXml(getChunk(), NULL, x, getWorkspace());
                CPPToXHDMerge::updateScope(x, &sc);
                //if (STRCHR(x, '\n')) setcr=1;
              }
              if (sc != m_curscope) {
                //setcr = 1;
                m_curscope = sc;
                switch (sc) {
                  case SCOPE_PUBLIC:
                    val = "public:\n";
                    break;
                  case SCOPE_PRIVATE:
                    val = "private:\n";
                    break;
                  case SCOPE_PROTECTED:
                    val = "protected:\n";
                    break;
                }
              }
              break;
            }
            case FT_CODEBLOCK_BLOCK: {
              String x = codeblock->getCodeBlock();
              if (!x.isempty()) {
                processInlineXml(getChunk(), NULL, x, getWorkspace());
              }
              val = x;
              break;
            }
          }
        }
        break;
        case FT_METHOD: {
          if (isclose) break;
          ClassFactoryMethod *method = getCurrentMethod();
          if (method == NULL) {
            val = "*** <method/> tag found outside a <foreach objects=\"methods\"/> ***";
            break;
          }
          String data = translateVariables(templatechunk->getParams()->getItemValue("data"));
          switch (methoddatatags.getId(data)) {
            case FT_METHOD_NAME:
              val = method->getName();
              break;
            case FT_METHOD_RETURNTYPE: {
              val = method->getReturnType();
              char c = val.lastChar();
              if (!val.isempty() && c != '*' && c != '&') val.cat(" ");
              break;
            }
            case FT_METHOD_DEFRETVAL:
              val = method->getDefaultReturnValue();
              break;
            case FT_METHOD_NUMPARAMS:
              val = StringPrintf("%d", method->getNumParams());
              break;
            case FT_METHOD_VIRTUAL:
              if (method->isVirtual(NULL)) val = "virtual ";
              break;
            case FT_METHOD_PURE: {
              int pure = 0;
              if (method->isVirtual(&pure) && pure) val = "=0";
              break;
            }
            case FT_METHOD_STATIC:
              if (method->isStatic()) val = "static ";
              break;
            case FT_METHOD_INLINE:
              if (method->isInline()) val = "inline ";
              break;
            case FT_METHOD_RETURN:
              if (isVoid(method->getReturnType())) break;
              val = "return ";
              break;
            case FT_METHOD_SOURCELINE:
              val = StringPrintf("%d", method->getSourceLine());
              break;
            case FT_METHOD_DISPATCHID: {
              int dispid = translateVariablesInt(method->getParams()->getItemValue("dispatch_id"));
              if (dispid == 0) {
                dispid = newDispatchableId(getChunk());
                method->getParams()->addItem("dispatch_id", StringPrintf("%d", dispid));
              }
              val = StringPrintf("%d", dispid);
              break;
            }
            case FT_METHOD_DISPATCHSID: {
              if (getChunk()) {
                String s = getChunk()->getClassName();
                s += "_";
                s += method->getName();
                s.toupper();
                val = s;
              }
              break;
            }
            case FT_METHOD_DISPATCHCALL:
              if (isVoid(method->getReturnType())) val = "_voidcall";
              else val = "_call";
              break;
            case FT_METHOD_DISPATCHDEFRETVAL: {
              if (isVoid(method->getReturnType())) {
                break;
              }
              String s;
              getDefaultForType(method->getReturnType(), s);
              s = String(", ") + s;
              val = s;
              break;
            }
            case FT_METHOD_DISPATCHCB:
              if (isVoid(method->getReturnType())) val = "VCB";
              else val = "CB";
              break;
            case FT_METHOD_MULTIPLEID: {
              ClassFactoryMethod *method = getCurrentMethod();
              if (method != NULL) {
                int pos=-1, multiple=0, isfirst=0;
                getMultipleMethod(method, &pos, &multiple, &isfirst);
                if (pos != -1 && multiple && !isfirst) {
                  val = StringPrintf("%d", multiple);
                }
              break;
              }
            }
            case FT_METHOD_SCOPE: {
              int sc = method->getScope();
              if (sc != m_curscope) {
                //setcr = 1;
                m_curscope = sc;
                switch (sc) {
                  case SCOPE_PUBLIC:
                    val = "public:\n";
                    break;
                  case SCOPE_PRIVATE:
                    val = "private:\n";
                    break;
                  case SCOPE_PROTECTED:
                    val = "protected:\n";
                    break;
                }
              }
              break;
            }
          }
          break;
        }
        case FT_TEMPLATEPARAM: {
          if (isclose) break;
          param = getCurrentTemplateParam();
          if (param == NULL) {
            val = "*** <templateparam/> tag found outside a <foreach objects=\"templateparam\"/> ***";
            break;
          }
          goto paramhandler;
        }
        case FT_PARAM: {
          if (isclose) break;
          param = getCurrentParam();
          if (param == NULL) {
            val = "*** <param/> tag found outside a <foreach objects=\"params\"/> ***";
            break;
          }
          paramhandler:
          String data = translateVariables(templatechunk->getParams()->getItemValue("data"));
          switch (paramdatatags.getId(data)) {
            case FT_PARAM_NAME:
              val = param->getName();
              break;
            case FT_PARAM_TYPE: {
              val = param->getType();
              char c = val.lastChar();
              if (!val.isempty() && c != '*' && c != '&') val.cat(" ");
              break;
            }
            case FT_PARAM_ALIASTYPE: {
              String str;
              getAliasType(param->getType(), str);
              val = str;
              char c = val.lastChar();
              if (!val.isempty() && c != '*' && c != '&') val.cat(" ");
              break;
            }
            case FT_PARAM_PTRREF: {
              ClassFactoryParam *param = getCurrentParam();
              if (param) {
                String str=param->getType();
                ClassFactoryParser::trimSpaces(str);
                if (str.lastChar() == '&') val = "&";
                else if (str.lastChar() == '*') val = "*";
              }
              break;
            }
            case FT_PARAM_IO: {
              ClassFactoryParam *param = getCurrentParam();
              switch (param->getInOut()) {
                case IO_IN:
                  val = "in ";
                  break;
                case IO_OUT:
                  val = "out ";
                  break;
                case IO_INOUT:
                  val = "inout ";
                  break;
              }
              break;
            }
            case FT_PARAM_DEFVAL: {
              const char *def = param->getDefaultValue();
              if (def != NULL && *def != 0)
                val = StringPrintf(" = %s", def);
              break;
            }
          }
          break;
        }
        case FT_DATA: {
          if (isclose) break;
          ClassFactoryData *data = getCurrentData();
          if (data == NULL) {
            val = "*** <data/> tag found outside a <foreach objects=\"data\"/> ***";
            break;
          }
          String datadata = translateVariables(templatechunk->getParams()->getItemValue("data"));
          switch (datadatatags.getId(datadata)) {
            case FT_DATA_NAME:
              val = data->getName();
              break;
            case FT_PARAM_TYPE: {
              val = data->getDataType();
              char c = val.lastChar();
              if (!val.isempty() && c != '*' && c != '&') val.cat(" ");
              break;
            }
            case FT_DATA_CTORCODE: {
              val = data->getCtorCode();
              break;
            }
            case FT_DATA_DTORCODE: {
              val = data->getDtorCode();
              break;
            }
            case FT_DATA_SOURCELINE: {
              val = StringPrintf("%d", data->getSourceLine());
              break;
            }
            case FT_DATA_ALIASTYPE: {
              String str;
              getAliasType(data->getDataType(), str);
              val = str;
              char c = val.lastChar();
              if (!val.isempty() && c != '*' && c != '&') val.cat(" ");
              break;
            }
          }
          break;
        }
        case FT_INDENT:
          if (isclose) {
            POPI;
          } else {
            PUSHI;
          }
          break;
        case FT_FORINDEX: {
          if (isclose) break;
          ForIterator *forit = getForIterator();
          if (forit == NULL) {
            val = "*** <forcomma/> tag found outside a <foreach/> ***";
            break;
          }
          val.printf("%d", forit->getIterator());
        }
        break;
        case FT_FORCOMMA: {
          if (isclose) break;
          ForIterator *forit = getForIterator();
          if (forit == NULL) {
            val = "*** <forcomma/> tag found outside a <foreach/> ***";
            break;
          }
          if (forit->getIterator() == 0) break;
          val = ", ";
          break;
        }
        case FT_ADDPARAMCOMMA: {
          if (isclose) break;
          ClassFactoryMethod *m = getCurrentMethod();
          if (m && m->getNumParams() > 0) val = ", ";
          break;
        }
        case FT_SKIP: {
          if (isclose) break;
          int fortype = foreachobjecttags.getId(translateVariables(templatechunk->getParams()->getItemValue("objects")));
          ForIterator *iter = getLastIteratorByType(fortype);
          if (iter == NULL) {
            val = "*** <skip /> tag found outside the context it wants to skip ***";
            break;
          }
          const char *type = translateVariables(templatechunk->getParams()->getItemValue("type"));
          ParamParser pp(type, "|");
          const char *_op = translateVariables(templatechunk->getParams()->getItemValue("op"));
          int op_and = 0; // OR by default
          if (STRCASEEQL(_op, "and")) op_and = 1;
          // todo: xor
          int match = op_and ? 1 : 0;
          for (int p=0;p<pp.getNumItems();p++) {
            if (!op_and && testMatchesType(fortype, pp.enumItem(p), templatechunk->getParams())) match = 1;
            else if (op_and && !testMatchesType(fortype, pp.enumItem(p), templatechunk->getParams())) match = 0;
          }
          int op_not = translateVariablesInt(templatechunk->getParams()->getItemValue("not"));
          if (op_not) match = !match;
          if (match) {
            if (!continueByType(fortype))
              continue;
          }
        }
        break;
        case FT_BREAK: {
          if (isclose) break;
          int fortype = foreachobjecttags.getId(translateVariables(templatechunk->getParams()->getItemValue("objects")));
          if (fortype == -1) {
            ForIterator *fi = getLastIterator();
            if (fi) fortype = fi->getType();
            else val = "*** <break /> found outside of a for loop ***";
          }
          ForIterator *iter = getLastIteratorByType(fortype);
          if (iter == NULL) {
            val = "*** <break /> tag found outside of a for loop with the requested context ***";
            break;
          } else
            if (!breakByType(fortype)) continue;
        }
        break;
        case FT_CONTINUE: {
          if (isclose) break;
          int fortype = foreachobjecttags.getId(translateVariables(templatechunk->getParams()->getItemValue("objects")));
          if (fortype == -1) {
            ForIterator *fi = getLastIterator();
            if (fi) fortype = fi->getType();
            else val = "*** <continue /> found outside of a for loop ***";
          }
          ForIterator *iter = getLastIteratorByType(fortype);
          if (iter == NULL) {
            val = "*** <continue /> tag found outside of a for loop with the requested context ***";
            break;
          } else
            if (!continueByType(fortype)) continue;
        }
        break;
        case FT_DERIVEANCESTOR: {
          if (isclose) break;
          if (!getChunk()) break;

          HelperEnum en(getChunk());
          int n = en.findHelper(m_curformat) - 1;
          while (n >= 0) {
            String anc_class = getChunk()->getParams()->getItemValue(StringPrintf("$_%s_inherit_parallel", en.enumHelper(n)));
            if (anc_class.isempty()) break;
            n--;
          }

          if (n >= 0) {
            XHDProcessor processor(0, getChunk(), en.enumHelper(n));
            ProcessorEntry *pe = processor.getFactoryEntry(en.enumHelper(n));
            if (pe != NULL) {
              val = String(" : public ") + pe->m_classname;
            }
          }

          if (templatechunk->getParams()->getItemValueInt("moreancestors")) {
            if (val.isempty()) val += " : "; else val += ", ";
          }

          break;
        }
        case FT_PROJECT: {
          if (isclose) break;
          const char *data= translateVariables(templatechunk->getParams()->getItemValue("data"));
          int b = projectdatatags.getId(data);
          switch (b) {
            case FT_PROJECT_NAME:
              val = g_workspaceName ? g_workspaceName->getValue() : "application";
              break;
          }
        }
        break;
#ifdef ODE_IDE
        case FT_PROJECTFOLDER: {
          if (isclose) break;
          const char *data = translateVariables(templatechunk->getParams()->getItemValue("data"));
          int b = projectfolderdatatags.getId(data);
          switch (b) {
            case FT_PROJECTFOLDER_NAME: {
              ForEachProjectFolder *fepf = getForEachProjectFolder();
              if (fepf) {
                PathParser pp(fepf->getFolder());
                val = pp.getLastString();
              }
              break;
            }
            case FT_PROJECTFOLDER_PATH:
              ForEachProjectFolder *fepf = getForEachProjectFolder();
              if (fepf) {
                val = fepf->getFolder();
                if (strnicmp(val, "/root", 5))
                  val = String(val.getValue() + 5);
              }
              break;
          }
        }
        break;
#endif
        case FT_STDAFX: {
          if (isclose) break;
          String pf = g_workspace->getPrecompFile();
          if (!pf.isempty()) { // fucko !
            val = "#include \"";
            val += g_workspace->getPrecompFile();
            val += "\"\n";
          }
        }
        break;
        case FT_INC: {
          if (!isclose) {
            String id = translateVariables(templatechunk->getParams()->getItemValue("id"));
            String v = StringPrintf("%d", ATOI(m_Vars.getString(id)) + 1);
            m_Vars.setString(id, v);
          }
        }
        break;
        case FT_DEC: {
          if (!isclose) {
            String id = translateVariables(templatechunk->getParams()->getItemValue("id"));
            String v = StringPrintf("%d", ATOI(m_Vars.getString(id)) - 1);
            m_Vars.setString(id, v);
          }
        }
        break;
        case FT_PGET: {
          if (!isclose) {
            String id = translateVariables(templatechunk->getParams()->getItemValue("id"));
            const char *s = getChunk()->getParams()->getItemValue(id);
            if (!s) {
              val = translateVariables(templatechunk->getParams()->getItemValue("default"));
              int setdef = translateVariablesInt(templatechunk->getParams()->getItemValue("setdefault"));
              if (setdef) {
                getChunk()->getParams()->replaceItem(id, val);
              }
            }
            else val = s;
          }
        }
        break;
        case FT_GET: {
          if (!isclose) {
            String id = translateVariables(templatechunk->getParams()->getItemValue("id"));
            const char *s = m_Vars.getString(id);
            if (!s) val = translateVariables(templatechunk->getParams()->getItemValue("default"));
            else val = s;
          }
        }
        break;
        case FT_PSET:
        case FT_SET: {
          if (isclose) {
            m_curTemplate->popSet();
          } else {
            String id = translateVariables(templatechunk->getParams()->getItemValue("id"));
            if (templatechunk->getParams()->getItemValue("value")) {
              String value = translateVariables(templatechunk->getParams()->getItemValue("value"));
              if (xmltagtype == FT_PSET)
                getChunk()->getParams()->replaceItem(id, value);
              else
                m_Vars.setString(id, value);
            } else {
              if (xmltagtype == FT_PSET)
                getChunk()->getParams()->replaceItem(id, "");
              else
                m_Vars.setString(id, "");
            }
            m_curTemplate->pushSet(id, (xmltagtype == FT_PSET));
          }
        }
        break;
        case FT_T: {
          if (!isclose) {
            val = translateVariables(templatechunk->getParams()->getItemValue("t"));
          }
        }
        break;
#ifdef ODE_IDE
        case FT_GETGLOBAL: {
          String id = translateVariables(templatechunk->getParams()->getItemValue("id"));
          val = g_globalVars->getString(id);
        }
        break;
        case FT_SETGLOBAL: {
          String id = translateVariables(templatechunk->getParams()->getItemValue("id"));
          String value = translateVariables(templatechunk->getParams()->getItemValue("value"));
          g_globalVars->setString(id, value);
        }
        break;
        case FT_PACKAGEDFILE: {
          const char *data = translateVariables(templatechunk->getParams()->getItemValue("data"));
          ProjectFile *pf = getCurrentPackagedFile();
          if (pf) {
            int a = packagedfiledatatags.getId(data);
            switch (a) {
              case FT_PACKAGEDFILE_DATA_FILE:
                val = pf->getFilename();
                break;
              case FT_PACKAGEDFILE_DATA_DIR:
                val = pf->getTargetDir();
                break;
              case FT_PACKAGEDFILE_DATA_FOLDER:
                val = pf->getProjectFolder();
                break;
            }
          }
        break;
        }
#endif
        case FT_EXTERN: { // obsolete, here for compatibility, should go away as pget does the same thing better
          if (isclose || !getChunk()) break;
          const char *data= translateVariables(templatechunk->getParams()->getItemValue("data"));
          val = getChunk()->getParams()->getItemValue(data);
          break;
        }
        case FT_DATATEMPLATE: {
          if (isclose) break;
          ClassFactoryData *data = getCurrentData();
          if (!data) break;
          String templatename;
          PtrList<String> templateparams;
          parseTemplate(data->getDataType(), templatename, templateparams);
          if (templatename.isempty()) break;
          FactoryTemplate *tmpl = loadDataTemplate(templatename);
          if (tmpl == NULL) break;
          m_curTemplate->savePos(m_templatepos+1); // +1 = next tag
          m_templatestack.push(m_curTemplate);
          m_curTemplate = tmpl;
          tmpl->setMainModuleName(translateVariables(templatechunk->getParams()->getItemValue("module")));
          m_templatepos = 0;
          n_template_chunks = m_curTemplate->getNumChunks();
          m_dataoverridestack.push(m_curdataoverride);
          m_curdataoverride = data;
          continue;
        }
        case FT_CALL: {
          if (isclose) break;
          const char *id = translateVariables(templatechunk->getParams()->getItemValue("id"));
          FactoryModule *fm = getProc(id);
          if (!fm) break;

          //DebugString("Call %s\n", id);

          PtrList<GlobalString> varsave;
          for (int i=0;i<templatechunk->getParams()->getNbItems();i++) {
            if (!STRCASEEQL(templatechunk->getParams()->getItemName(i), "id")) {
              varsave.addItem(new GlobalString(templatechunk->getParams()->getItemName(i), m_Vars.getString(templatechunk->getParams()->getItemName(i))));
              m_Vars.setString(templatechunk->getParams()->getItemName(i), translateVariables(templatechunk->getParams()->getItemValue(i)));
            }
          }

          m_outputstack.push(m_curdata);
          m_curdata = new BigString();

          m_posstack.push(m_templatepos);

          m_templatepos = fm->getStartPos()+1; // avoid calling FT_PROC

          processFormat();

          String ret = m_curdata->getValue();
          delete m_curdata;

          m_outputstack.pop(&m_curdata);
          m_posstack.pop(&m_templatepos);
          val = ret;

          foreach(varsave)
            m_Vars.setString(varsave.getfor()->getId(), varsave.getfor()->getValue());
          endfor;
          varsave.deleteAll();
          break;
        }
        case FT_PROC: {
          if (!isclose) m_skip++;
          if (isclose)
            if (!m_skip) {
              l_return = 1;
              if (m_gotcr) {
                //addToCurData(INDT);
                //DebugString("leaving proc with cr\n");
              }
            }
            else
              m_skip--;
          break;
        }
        case FT_HELPERFILE: {
          if (isclose) break;
          ForEachHelper *h = getForEachHelper();
          ForEachHelperFile *hf = getForEachHelperFile();
          if (h == NULL) {
            val = "*** <helperfile/> tag found outside a foreach helperfiles context ***";
            break;
          }
          int helpernum = h->getIterator();
          int helperfilenum = hf->getIterator();
          HelperFileEnum en(getChunk(), helpernum);
          val = en.enumFile(helperfilenum);
          break;
        }
        case FT_AUTOHEADER: {
          if (isclose) break;
          if (!getChunk()) break;
          String s;

/*
          String classGuid = getChunk()->getParams()->getItemValue("class_guid");
          if (!classGuid.isempty()) {
            s += String("// ") + classGuid + "\n";
            s += "static const GUID "; s += getChunk()->getClassName(); s += " = \n";
            GUID cguid = nsGUID::fromChar(classGuid);
            s += StringPrintf("{ 0x%08x, 0x%04x, 0x%04x, { 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x } }\n",
                                cguid.Data1,
                                cguid.Data2,
                                cguid.Data3,
                                cguid.Data4[0],
                                cguid.Data4[1],
                                cguid.Data4[2],
                                cguid.Data4[3],
                                cguid.Data4[4],
                                cguid.Data4[5],
                                cguid.Data4[6],
                                cguid.Data4[7]);

          }
*/
          if (!m_inherit_parallel) {

            HelperEnum en(getChunk());
            int n = en.getNumHelpers()-1;

            int isinline = m_curformat.iscaseequal("inlinexml");

            if (!isinline) {
              n = en.findHelper(m_curformat)-1;
            } else {
              // use last
            }

            int gottail = 0;
            while (n >= 0) {
              String anc_class = getChunk()->getParams()->getItemValue(StringPrintf("$_%s_inherit_parallel", en.enumHelper(n)));
              if ((isinline && (!anc_class.isempty() || !gottail)) || (!isinline && !gottail && anc_class.isempty())) {
                if (anc_class.isempty()) gottail = 1;
                String prevancestor = getChunk()->getAncestor();
                XHDProcessor processor(0, getChunk(), en.enumHelper(n));
                svc_srcClassFactory *fact = processor.getFactory(en.enumHelper(n));
                if (fact != NULL) {
                  int mn = fact->getNumModules();
                  int j;
                  for (j=0;j<mn;j++) {
                    const char *filename = fact->enumModuleName(j);
                    if (STRSTR(filename, ".h")) {
                      s += "#include \"";
                      if (!g_outputDir.isempty()) s += g_outputDir;
                      s += filename;
                      s += "\"\n";
                    }
                  }
                }
                getChunk()->setAncestor(prevancestor);
                if (!m_curformat.iscaseequal("inlinexml")) break;
              }
              n--;
            }
          }

          PtrListInsertSorted<String, TSortString> forwdecls;

          ForEachMethod fem(this, getChunk(), NULL);
          int nb = fem.getNumMethods();
          int ni;
          for (ni=0;ni<nb;ni++) {
            ClassFactoryMethod *method = fem.getMethod(ni);
            String rettype = method->getReturnType();
            String classname;
            if (!isVoid(rettype) && isClass(rettype, classname)) {
              if (STRCHR(rettype, '*')) {
                if (forwdecls.findItem(classname)) continue;
                String uc = classname; uc.toupper();
                if (classname.isequal(uc)) continue; // disallow all caps objects
                  // disallow this class
                if (STREQL(classname, getClassLayer())) continue;
                // disallow unknown types
                if (g_workspace->findChunk(classname))
                  forwdecls.addItem(new String(classname));
              }
            }
            int np = method->getNumParams();
            int ip;
            for (ip=0;ip<np;ip++) {
              ClassFactoryParam *param = method->enumParam(ip);
              String type = param->getType();
              if (!isVoid(type) && isClass(type, classname)) {
                if (STRCHR(type, '*')) {
                  if (forwdecls.findItem(classname)) continue;
                  String uc = classname; uc.toupper();
                  if (classname.isequal(uc)) continue; // disallow all caps objects
                  // disallow this class
                  if (STREQL(classname, getClassLayer())) continue;
                  // disallow unknown types
                  if (g_workspace->findChunk(classname))
                    forwdecls.addItem(new String(classname));
                }
              }
            }
          }

          nb = getChunk()->getNumBlocks();
          ni;
          for (ni=0;ni<nb;ni++) {
            ClassFactoryBlock *block = getChunk()->enumBlock(ni);
            if (block->getBlockType() == CF_BLOCKTYPE_DATA) {
              ClassFactoryData *data = block->getData();
              String type = data->getDataType();
              String classname;
              if (!isVoid(type) && isClass(type, classname)) {
                if (STRCHR(type, '*')) {
                  if (forwdecls.findItem(classname)) continue;
                  String uc = classname; uc.toupper();
                  if (classname.isequal(uc)) continue; // disallow all caps objects
                  // disallow this class
                  if (STREQL(classname, getClassLayer())) continue;
                  // disallow unknown types
                  if (g_workspace->findChunk(classname))
                    forwdecls.addItem(new String(classname));
                }
              }
            }
          }
          int mcount = 0;
          foreach(forwdecls)
            if (mcount++ == 0 && !s.isempty()) s += "\n";
            s += "class ";
            s += *forwdecls.getfor();
            s += ";\n";
          endfor;
          forwdecls.deleteAll();
          val = s;
          break;
        }
        case FT_FOREACH: {
          if (isclose) {
            ForIterator *iter = getForIterator();
            if (m_continueskip > 0) {
              m_continueskip--;
              if (m_continueskip == 0) {
                delete iter;
                popForeach();
              }
            }
            if (discard_condition) {
              m_controlstack.pop(&m_control_discard);
              if (m_skip > 0) m_skip--;
              break;
            }
            if (iter == NULL) {
              val = "*** internal error with </foreach> ***";
              break;
            }
            if (iter->isLast()) {
              m_controlstack.pop(&m_control_discard);
              delete iter;
              popForeach();
              break;
            } else {
              iter->inc();
              m_templatepos = iter->getLoopPos();
              continue;
            }
          } else {
            if (m_continueskip > 0) {
              m_continueskip++;
            }
            if (discard_condition) {
              m_controlstack.push(m_control_discard);
              m_control_discard = 1;
              if (m_skip > 0) m_skip++;
              break;
            }
            String objects = translateVariables(templatechunk->getParams()->getItemValue("objects"));
            if (m_ppercent) m_ppercent->setPercent(0);
            String s;
            ForIterator *iter = createForEach(objects, templatechunk->getParams(), getChunk(), getCurrentMethod(), getCurrentParam(), s);
            if (!iter || iter->getNumIterations() == 0) {
              val = s;
              m_controlstack.push(m_control_discard);
              m_control_discard = 1;
              delete iter;
              break;
            }
            m_controlstack.push(m_control_discard);
            m_control_discard = 0;
            iter->setLoopPos(m_templatepos+1);
            pushForeach(iter);
          }
          break;
        }
        case FT_IF: {
          if (isclose) {
            m_controlstack.pop(&m_control_discard);
            break;
          } else {
            if (discard_condition) {
              m_controlstack.push(m_control_discard);
              m_control_discard = 1;
              break;
            }
            int istrue = 0;
            // do test
            int objecttype = foreachobjecttags.getId(translateVariables(templatechunk->getParams()->getItemValue("object")));
            const char *_op = translateVariables(templatechunk->getParams()->getItemValue("op"));
            int op_and = 0; // OR by default
            if (STRCASEEQL(_op, "and")) op_and = 1;
            // todo: xor
            if (objecttype == -1) {
              if (templatechunk->getParams()->getItemValue("text") != NULL) {
                // string tests
                String text = translateVariables(templatechunk->getParams()->getItemValue("text"));
                String match = translateVariables(templatechunk->getParams()->getItemValue("match"));
                String equal = translateVariables(templatechunk->getParams()->getItemValue("equal"));
                if (match.isempty() && equal.isempty()) {
                  if (op_and) istrue &= text.isempty();
                  else istrue |= text.isempty();
                } else if (!match.isempty() && equal.isempty()) {
                  if (op_and) istrue &= Std::match(match, text);
                  else istrue |= Std::match(match, text);
                } else if (!equal.isempty()) {
                  if (op_and) istrue &= STRCASEEQLSAFE(equal, text);
                  else istrue |= STRCASEEQLSAFE(equal, text);
                }
              }
              if (templatechunk->getParams()->getItemValue("bool") != NULL) {
                // boolean test
                String booleanval = translateVariables(templatechunk->getParams()->getItemValue("bool"));
                if (op_and) istrue = 0; // err that doesn't look right... FUCKO CHECK!
                else istrue |= (ATOI(booleanval) != 0);
              }
              if (templatechunk->getParams()->getItemValue("procdef") != NULL) {
                // is proc defined ?
                String proc = translateVariables(templatechunk->getParams()->getItemValue("procdef"));
                istrue = getProc(proc) != NULL; // todo: add multiple procdefs params support as well as op_and
              }
            } else {
              String type = translateVariables(templatechunk->getParams()->getItemValue("isoftype"));
              String inherits = translateVariables(templatechunk->getParams()->getItemValue("inherits"));
              if (inherits.isempty()) inherits = translateVariables(templatechunk->getParams()->getItemValue("inherit"));
              int match = op_and ? 1 : 0;
              if (!type.isempty()) {
                ParamParser pp(type, "|");
                for (int p=0;p<pp.getNumItems();p++) {
                  if (!op_and && testMatchesType(objecttype, pp.enumItem(p), templatechunk->getParams())) match = 1;
                  else if (op_and && !testMatchesType(objecttype, pp.enumItem(p), templatechunk->getParams())) match = 0;
                }
              }
              if (!inherits.isempty()) {
                ParamParser pp(inherits, "|");
                for (int p=0;p<pp.getNumItems();p++) {
                  if (!op_and && testInherit(getChunk(), pp.enumItem(p))) match = 1;
                  else if (op_and && !testInherit(getChunk(), pp.enumItem(p))) match = 0;
                }
              }
              istrue = match;
            }

            int op_not = translateVariablesInt(templatechunk->getParams()->getItemValue("not"));
            if (op_not) istrue = !istrue;
            if (!istrue) {
              m_controlstack.push(m_control_discard);
              m_control_discard = 1;
              break;
            }
            m_controlstack.push(m_control_discard);
            m_control_discard = 0;
          }
          break;
        }
        case FT_ELSE: {
          m_control_discard = !m_control_discard;
          break;
        }
        case FT_TYPEMAP: {
          if (!isclose) {
            if (m_curTypeMap == NULL) {
              const char *id = translateVariables(templatechunk->getParams()->getItemValue("id"));
              m_curTypeMap = g_typeMaps->getTypeMap(id);
              if (!m_curTypeMap) {
                m_curTypeMap = new TypeMap(this, id);
                g_typeMaps->addTypeMap(m_curTypeMap);
              }
            }
          } else m_curTypeMap = NULL;
        }
        break;
        case FT_MAPENTRY:
          if (!isclose) {
            ForEachMapEntries *fema = getForEachMapEntries();
            if (!fema)
              val = "*** <mapentry/> tag found outside a <foreach object=\"mapentries\"/> ***";
            else {
              int n = fema->getIterator();
              TypeMap *tm = fema->getMap();
              if (tm) {
                String mapentrydata = translateVariables(templatechunk->getParams()->getItemValue("data"));
                switch (mapentrydatatags.getId(mapentrydata )) {
                  case FT_MAPENTRY_FROM:
                    val = tm->enumMappingFrom(fema->getIterator());
                    break;
                  case FT_MAPENTRY_TO:
                    val = tm->enumMappingTo(fema->getIterator());
                    break;
                  case FT_MAPENTRY_IMPL:
                    val = tm->enumMappingImpl(fema->getIterator());
                    break;
                }
              }
            }
          }
        break;
        case FT_MAP: {
          if (!isclose) {
            if (m_curTypeMap) {
              String from = translateVariables(templatechunk->getParams()->getItemValue("from"));
              String to = translateVariables(templatechunk->getParams()->getItemValue("to"));
              String impl = translateVariables(templatechunk->getParams()->getItemValue("impl"));
              String _default =translateVariables(templatechunk->getParams()->getItemValue("default"));
              if (_default.isempty())
                m_curTypeMap->addMapping(from, to, impl);
              else
                m_curTypeMap->setDefaultMapping(_default, impl);
            } else {
              String from_type;
              String typemap;
              int nodef = translateVariablesInt(templatechunk->getParams()->getItemValue("nodefault"));
              int impl = translateVariablesInt(templatechunk->getParams()->getItemValue("impl"));
              int basemapping = translateVariablesInt(templatechunk->getParams()->getItemValue("basemapping"));
              from_type = translateVariables(templatechunk->getParams()->getItemValue("type"));
              typemap = translateVariables(templatechunk->getParams()->getItemValue("map"));
              ClassFactoryParser::trimSpaces(from_type);
              if (from_type.isempty() || typemap.isempty()) break;
              char c = from_type.lastChar();
              if ((c == '*' || c == '&') && !endsIn(from_type, "char *"))
                from_type.trunc(from_type.numCharacters()-1);
              else c = 0;
              ClassFactoryParser::trimSpaces(from_type);
              String _default;
              foreach(g_typeMaps->getMaps())
                TypeMap *map = g_typeMaps->getMaps()->getfor();
                if (STRCASEEQLSAFE(map->getId(), typemap)) {
                  _default = map->getDefaultMapping();
                  TypeMapEntry *e = map->getMap(from_type);
                  if (e != NULL) {
                    val = e->getMapping(impl);
                    if (val.lastChar() != '&')
                      if (!basemapping && c) val += StringPrintf("%s%c", val.lastChar() == '*' ? "" : " ", c);
                      else val += " ";
                    break;
                  }
                }
              endfor;
              if (!nodef && val.isempty()) {
                val = _default.isempty() ? from_type : _default;
                if (val.lastChar() != '&')
                  if (!basemapping && c) val += StringPrintf("%s%c", val.lastChar() == '*' ? "" : " ", c);
                  else val += " ";
              }
            }
          }
        }
        break;
        case FT_TYPECAST: {
          if (!isclose) {
            String from = translateVariables(templatechunk->getParams()->getItemValue("from"));
            String to = translateVariables(templatechunk->getParams()->getItemValue("to"));
            String cast = translateVariables(templatechunk->getParams()->getItemValue("cast"));
            int ismacro = translateVariablesInt(templatechunk->getParams()->getItemValue("ismacro"));
            g_casts->addTypeCast(from, to, cast, ismacro);
          }
          break;
        }
        case FT_CAST: {
          if (!isclose) {
            String from;
            String to;
            from = translateVariables(templatechunk->getParams()->getItemValue("from"));
            to = translateVariables(templatechunk->getParams()->getItemValue("to"));
            String test = translateVariables(templatechunk->getParams()->getItemValue("test"));
            ClassFactoryParser::trimSpaces(from);
            ClassFactoryParser::trimSpaces(to);
            if (from.isempty() || to.isempty()) break;
            if (test.isempty())
              val = g_casts->getCast(from, to);
            else {
              if (test.iscaseequal("ismacro")) {
                Cast *c = g_casts->getCastObject(from, to);
                if (c && c->isMacro()) val = "1";
                else val = "0";
              }
            }
          }
          break;
        }
        case FT_YEAR: {
          if (isclose) break;
          time_t now;
          time(&now);
          struct tm *ntime;
          ntime = localtime(&now);
          val = StringPrintf(ntime->tm_year + 1900);
          break;
        }
        case FT_MONTH: {
          if (isclose) break;
          time_t now;
          time(&now);
          struct tm *ntime;
          ntime = localtime(&now);
          val = StringPrintf(ntime->tm_mon + 1);
          break;
        }
        case FT_MONTHNAME: {
          if (isclose) break;
          time_t now;
          time(&now);
          struct tm *ntime;
          ntime = localtime(&now);
          char *monthname[12] = {"january", "february", "march", "april", "may", "june", "july", "august", "september", "october", "november", "december"};
          val = monthname[ntime->tm_mon];
          break;
        }
        case FT_MONTHDAY: {
          if (isclose) break;
          time_t now;
          time(&now);
          struct tm *ntime;
          ntime = localtime(&now);
          val = StringPrintf(ntime->tm_mday);
          break;
        }
        case FT_WEEKDAY: {
          if (isclose) break;
          time_t now;
          time(&now);
          struct tm *ntime;
          ntime = localtime(&now);
          char *weekday[7] = {"sunday", "monday", "thuesday", "wednesday", "thrusday", "friday", "saturday"};
          val = weekday[ntime->tm_wday];
          break;
        }
        case FT_HOUR: {
          if (isclose) break;
          time_t now;
          time(&now);
          struct tm *ntime;
          ntime = localtime(&now);
          val = StringPrintf("%d", ntime->tm_hour);
          break;
        }
        case FT_MINUTE: {
          if (isclose) break;
          time_t now;
          time(&now);
          struct tm *ntime;
          ntime = localtime(&now);
          val = StringPrintf("%02d", ntime->tm_min);
          break;
        }
        case FT_SECOND: {
          if (isclose) break;
          time_t now;
          time(&now);
          struct tm *ntime;
          ntime = localtime(&now);
          val = StringPrintf("%02d", ntime->tm_sec);
          break;
        }
        case FT_WARNING:
        case FT_ERROR: {
          if (isclose) break;
          String text = translateVariables(templatechunk->getParams()->getItemValue("text"));
          int code = translateVariablesInt(templatechunk->getParams()->getItemValue("code"), -1);
          int errline = translateVariablesInt(templatechunk->getParams()->getItemValue("line"), -1);
          String file = translateVariables(templatechunk->getParams()->getItemValue("file"));

          String incpath;
          String filename;
          if (errline == -1 && file.isempty()) {
            errline = templatechunk->getFileLine();
            incpath = templatechunk->getIncludePath();
            filename = templatechunk->getFilename();
          } else {
            // todo: split file into incpath & filename
            filename = file;
          }

          if (xmltagtype == FT_ERROR)
            xpp_errorMessage(filename, errline, incpath, code, text);
          else
            xpp_warningMessage(filename, errline, incpath, code, text);

          break;
        }
      }

      if (!val.isempty()) {

        int v;
        String delprefix = translateVariables(templatechunk->getParams()->getItemValue("delprefix"));
        if (!delprefix.isempty()) {
          if (!STRNICMP(val.getValue(), delprefix.getValue(), delprefix.len()))
            val = String(val.getValue()+delprefix.len());
        }
        String delsuffix = translateVariables(templatechunk->getParams()->getItemValue("delsuffix"));
        if (!delsuffix.isempty()) {
          ParamParser pp(delsuffix, "|");
          for (int i=0;i<pp.getNumItems();i++) {
            String ds = pp.enumItem(i);
            int vallen = val.len();
            int srclen = ds.len();
            if (STRCASEEQL(val.getValue()+vallen-srclen, ds)) val.trunc(vallen-srclen);
          }
        }
        String enforcesuffix = translateVariables(templatechunk->getParams()->getItemValue("enforcesuffix"));
        if (!enforcesuffix.isempty()) {
          int vallen = val.len();
          int srclen = delsuffix.len();
          if (!STRCASEEQL(val.getValue()+vallen-srclen, delsuffix)) val.cat(enforcesuffix);
        }
        String enforceprefix = translateVariables(templatechunk->getParams()->getItemValue("enforceprefix"));
        if (!enforceprefix.isempty()) {
          if (STRNICMP(val.getValue(), enforceprefix, enforceprefix.len()))
            val = String(enforceprefix) + val;
        }
        int trim = translateVariablesInt(templatechunk->getParams()->getItemValue("trim"));
        if (trim) {
          ClassFactoryParser::trimSpaces(val);
        }
        String replace_from = translateVariables(templatechunk->getParams()->getItemValue("replacefrom"));
        String replace_to = translateVariables(templatechunk->getParams()->getItemValue("replaceto"));
        if (!replace_from.isempty() && !replace_to.isempty()) {
          val.replace(replace_from, replace_to);
        }
        int tv = translateVariablesInt(templatechunk->getParams()->getItemValue("uclead"));
        if (tv) {
          if (val.len() == 1) val.toupper();
          else if (val.len() > 1) {
            if (tv == 2) val.tolower();
            val = StringPrintf("%c%s", TOUPPER(val.getChar(0)), val.getValue()+1);
          }
        }
        if (translateVariablesInt(templatechunk->getParams()->getItemValue("lclead"))) {
          if (val.len() == 1) val.tolower();
          else if (val.len() > 1)
            val = StringPrintf("%c%s", TOLOWER(val.getChar(0)), val.getValue()+1);
        }
        if (translateVariablesInt(templatechunk->getParams()->getItemValue("ucase"))) val.toupper();
        if (translateVariablesInt(templatechunk->getParams()->getItemValue("lcase"))) val.tolower();
        v = translateVariablesInt(templatechunk->getParams()->getItemValue("word"), -1);
        if (v >= 0) {
          const char *sep = translateVariables(templatechunk->getParams()->getItemValue("sep"));
          if (!sep) sep = " ";
          ParamParser pp(val, sep);
          val = pp.enumItem(v);
        }
        if (m_gotcr) {
          addToCurData(INDT);
        }
        m_gotcr = ((val.lastChar() == '\r') || (val.lastChar() == '\n'));
        m_curTemplate->addToVars(val);
        if (!m_curTemplate->isInSet()) addToCurData(val);
      }

      //if (setcr) m_gotcr = 1;
      m_templatepos++;
    }

    if (!m_templatestack.isempty() && !(STRCASEEQLSAFE(m_curTemplate->getFormat(), "[runvar]"))) {
      ASSERT(m_curTemplate != m_template);
      FactoryModule *mod = getModule(m_curTemplate->getMainModuleName(), m_curTemplate);
      if (mod != NULL)
        val = mod->getData();
      for (int i = 0;i < m_auto_modules.getNumItems(); i++) {
        FactoryModule *mod = m_auto_modules.enumItem(i);
        if (mod->getTemplate() == m_curTemplate) {
          m_auto_modules.removeByPos(i);
          m_modules.removeItem(mod);
          delete mod;
        }
      }
      delete m_curTemplate;
      m_templatestack.pop(&m_curTemplate);
      m_templatepos = m_curTemplate->getSavedPos();
      n_template_chunks = m_curTemplate->getNumChunks();
      m_dataoverridestack.pop(&m_curdataoverride);

      if (m_gotcr) {
        addToCurData(INDT);
        m_gotcr = 0;
      }
      addToCurData(val);

      continue;
    }

    break;
  }

/*
  ASSERT(m_forstack.isempty());
  ASSERT(m_templatestack.isempty());
  ASSERT(m_outputstack.isempty());
  ASSERT(m_dataoverridestack.isempty());
  ASSERT(m_moduledescstack.isempty());
  ASSERT(m_controlstack.isempty());
*/

  // done !

  m_running--;

  if (m_running) {
    m_gotcrstack.pop(&m_gotcr);
  } else {
    ASSERT(m_forstack.isempty());
  }

  return 1;
}

FactoryTemplate *TemplatizedFactory::insertFormatTemplate(const char *format, const char *tplt, const char *wantnew_format) {
  FactoryTemplate *loaded = NULL;
  m_curformat = format;
  m_intemplate = 0;
  m_parserinmodule = 0;

  FactoryTemplate *old = m_curTemplate;

  if (!wantnew_format) {
    if (m_template != NULL) resetFactory();

    delete m_template;
    m_template = new FactoryTemplate(this, m_curformat);

   m_curTemplate = m_template;
  } else {
    m_curTemplate = new FactoryTemplate(this, wantnew_format);
  }

  XmlReader::registerCallback("*", static_cast<XmlReaderCallback*>(this));
  BigString t;
  t += "buf:";
  t += "<Ode:HelperTemplate>";
  t += "<module id=\"";
  t += format;
  t += "\">";
  t += tplt;
  t += "</module>";
  t += "</Ode:HelperTemplate>";
  XmlReader::loadFile(t, NULL, static_cast<XmlReaderCallback*>(this));
  XmlReader::unregisterCallback(static_cast<XmlReaderCallback*>(this));

  loaded = m_curTemplate;

  m_curTemplate = old;
  return loaded;
}

int TemplatizedFactory::loadFormatTemplate(const char *format, const char *subdir, const char *ext) {
  if (m_template && STRCASEEQLSAFE(format, m_template->getFormat())) return 1;

  String file;

  file = g_workspace->getTemplateDir();
  if (!file.isempty()) {
    if (!Std::isDirChar(file.lastChar())) file.cat(DIRCHARSTR);
    file += StringPrintf("%s.%s", format, ext);
  }

  if (access(file, 0)) {

    if (subdir == NULL) subdir = "templates";
    if (ext == NULL) ext = m_ext;

    file = g_resourcepath;
    if (!file.isempty() && !Std::isDirChar(file.lastChar())) file.cat(DIRCHARSTR);
    if (file.isempty()) file = "";

    file = StringPrintf("%s%s/%s.%s", file.getValue(), subdir, format, ext);
    //duh! file.tolower();
    //printf("loading \"%s\"\n", file.getValue());
    if (access(file, 0)) {
      printf("Couldn't load template for format \"%s\"", format);
      return 0;
    }
  }

  m_curformat = format;
  m_intemplate = 0;
  m_parserinmodule = 0;

  m_curTypeMap = NULL;

  delete m_template;
  m_template = new FactoryTemplate(this, m_curformat);

  FactoryTemplate *old = m_curTemplate;
  m_curTemplate = m_template;

  XmlReader::registerCallback("*", static_cast<XmlReaderCallback*>(this));
  String path;
  PathParser pp(file);
  if (Std::isDirChar(*file.getValue())) path += DIRCHARSTR;
  for (int i=0;i<pp.getNumStrings()-1;i++) {
    path += pp.enumString(i);
    path += Std::dirCharStr();
  }
  file = pp.getLastString();
  XmlReader::loadFile(file, path, static_cast<XmlReaderCallback*>(this));
  XmlReader::unregisterCallback(static_cast<XmlReaderCallback*>(this));

  m_curTemplate = old;

  return 1;
}

FactoryTemplate *TemplatizedFactory::loadDataTemplate(const char *format, int testonly) {
  String file;
  String subdir = "templates/data";
  String ext = "xpdt";

  file = g_workspace->getTemplateDir();
  if (!file.isempty()) {
    if (!Std::isDirChar(file.lastChar())) file.cat(DIRCHARSTR);
    file += StringPrintf("data/%s.%s", format, ext.getValue());
  }

  if (access(file, 0)) {

    file = g_resourcepath;
    if (!file.isempty() && !Std::isDirChar(file.lastChar())) file.cat(DIRCHARSTR);
    if (file.isempty()) file = "";

    file = StringPrintf("%s%s/%s.%s", file.getValue(), subdir.getValue(), format, ext.getValue());
    //duh! file.tolower();
    if (access(file, 0)) {
      printf("Couldn't load data template for format \"%s\"", format);
      return 0;
    }
  }

  m_curformat = format;
  m_intemplate = 0;
  m_parserinmodule = 0;

  m_curTypeMap = NULL;

  FactoryTemplate *tmpl = new FactoryTemplate(this, m_curformat);
  if (testonly) return tmpl;

  FactoryTemplate *old = m_curTemplate;
  m_curTemplate = tmpl;

  XmlReader::registerCallback("*", static_cast<XmlReaderCallback*>(this));
  String path;
  PathParser pp(file);
  if (Std::isDirChar(*file.getValue())) path += DIRCHARSTR;
  for (int i=0;i<pp.getNumStrings()-1;i++) {
    path += pp.enumString(i);
    path += Std::dirCharStr();
  }
  file = pp.getLastString();
  XmlReader::loadFile(file, path, static_cast<XmlReaderCallback*>(this));
  XmlReader::unregisterCallback(static_cast<XmlReaderCallback*>(this));

  m_curTemplate = old;

  return tmpl;
}

int TemplatizedFactory::newDispatchableId(ClassFactoryChunk *chunk) {
  int i, n = chunk->getNumBlocks();
  TList<int> ids;
  for (i=0;i<n;i++) {
    ClassFactoryBlock *block = chunk->enumBlock(i);
    if (block->getBlockType() == CF_BLOCKTYPE_METHOD) {
      ClassFactoryMethod *method = block->getMethod();
      int id = translateVariablesInt(method->getParams()->getItemValue("dispatch_id"));
      if (id != 0)
        ids.addItem(id);
    }
  }
  int newid = 10;
  n = ids.getNumItems();
  for (i=0;i<n;i++) {
    if (newid == ids.enumItem(i)) {
      newid += 10;
      i = -1;
      continue;
    }
  }
  return newid;
}

FactoryModule *TemplatizedFactory::getModule(const char *format, FactoryTemplate *parent) {
  for (int i=0;i<m_modules.getNumItems();i++) {
    FactoryModule *module = m_modules.enumItem(i);
    if (module->getTemplate() == parent || (module->getTemplate() == NULL && parent == m_template)) {
      if (STRCASEEQLSAFE(format, module->getId()))
        return module;
    }
  }
  return NULL;
}

FactoryModule *TemplatizedFactory::getProc(const char *id) {
  for (int i=0;i<m_procs.getNumItems();i++) {
    FactoryModule *module = m_procs.enumItem(i);
    if (STRCASEEQLSAFE(id, module->getId()))
      return module;
  }
  return NULL;
}

const char *TemplatizedFactory::translateVariables(const char *str) {
  if (str == NULL) return "";
  char *p = (char *)str;
  String varname;
  String val;

  String rec="";

  while (*p) {
    if (((*p == '$' && *(p+1) != '$') || (*p == '#' && *(p+1) != '#')) && *(p+1)) {
      int persist = *p == '#';
      varname.trunc(0);
      val.trunc(0);
      if (*(p+1) == '(') {
        char *d = STRCHR(p, ')');
        if (d != NULL) {
          varname.catn(p+2, d-p-2);
          p = d;
        }
        if (!varname.isempty()) {
          int nullparamcount = 0;
          PtrListQuickSorted<String, StringComparator> paramlist;
          ParamParser pp(varname, ".");
          int n = pp.getNumItems();
          String xml;
          xml = "<";
          xml += String(pp.enumItem(0)) + " ";
          if (n > 1) {
            xml += String("data=\"") + pp.enumItem(1) + "\" ";
            paramlist.addItem(new String("data"));
            for (int g=2;g<n;g+=2) {
              String paramname = pp.enumItem(g);
              String paramvalue = pp.enumItem(g+1);
              if (paramname.isempty()) {
                paramname = StringPrintf("nullparam%d", nullparamcount++);
              }
              if (paramlist.findItem(paramname)) {
                if (paramname.isequal("data"))
                  xpp_errorMessage(getTemplateChunk()->getFilename(), getTemplateChunk()->getFileLine(), getTemplateChunk()->getIncludePath(), XML_ERROR_DUPLICATE_ATTRIBUTE, "duplication of default attribute 'data'");
                else
                  xpp_errorMessage(getTemplateChunk()->getFilename(), getTemplateChunk()->getFileLine(), getTemplateChunk()->getIncludePath(), XML_ERROR_DUPLICATE_ATTRIBUTE, StringPrintf("duplicate attribute '%s'", paramname.getValue()));
              }
              paramlist.addItem(new String(paramname));
              xml += String(paramname) + "=\"" + paramvalue + "\" ";
            }
            paramlist.deleteAll();

            xml += "/>";

            FactoryTemplate *tmpl = insertFormatTemplate("[runvar]", xml, "[runvar]");
            if (tmpl != NULL) {
              m_curTemplate->savePos(m_templatepos);
              m_templatestack.push(m_curTemplate);
              m_curTemplate = tmpl;
              m_templatepos = 0;
              n_template_chunks = m_curTemplate->getNumChunks();

              processFormat();

              FactoryModule *mod = getModule("[runvar]", m_curTemplate);
              if (mod != NULL)
                val = mod->getData();

              for (int i = 0;i < m_auto_modules.getNumItems(); i++) {
                FactoryModule *mod = m_auto_modules.enumItem(i);
                if (mod->getTemplate() == m_curTemplate) {
                  m_auto_modules.removeByPos(i);
                  m_modules.removeItem(mod);
                  delete mod;
                }
              }
              delete m_curTemplate;
              m_templatestack.pop(&m_curTemplate);
              m_templatepos = m_curTemplate->getSavedPos();
              n_template_chunks = m_curTemplate->getNumChunks();
              rec.cat(val);
            }
          }
        }
      } else {
        // walk til next separator
        char *pz = p+1;
        while (*pz) {
          if (!((*pz >= 'A' && *pz <= 'Z') || (*pz >= 'a' && *pz<= 'z') || (*pz >= '0' && *pz <= '9') || *pz == '_'))
            break;
          pz++;
        }
        varname.catn(p+1, pz-p-1);
        p = pz-1;
        if (*p && *(p+1) == '.')
          p++;
        if (persist) {
          rec.cat(getChunk()->getParams()->getItemValue(varname));
        } else {
          if (g_globalVars->haveString(varname))
            rec.cat(g_globalVars->getString(varname));
          else
            rec.cat(m_Vars.getString(varname));
        }
      }
    } else {
      if (*p == '$' && *(p+1) == '$') p++;
      else if (*p == '#' && *(p+1) == '#') p++;
      rec.catn(p, 1);
    }
    p++;
  }

  if (rec.len() > 0) {
    char *pp = rec.getNonConstVal() + rec.len()-1;
    while (*pp == ' ' && pp > rec.getNonConstVal()) *pp = 0;
    pp = rec.getNonConstVal();
    while (*pp == ' ') STRCPY(pp, pp+1);
  }

  m_translatereturn = rec;
  return m_translatereturn;
}

int TemplatizedFactory::translateVariablesInt(const char *str, int def) {
  String s = translateVariables(str);
  if (s.isempty()) return def;
  return ATOI(s.getValue());
}

void TemplatizedFactory::parseTemplate(const char *datatype, String &templatename, PtrList < String > &paramlist) {

  templatename = "";
  paramlist.deleteAll();
  if (!datatype) return;

  char *bracket = STRCHR(datatype, '<');
  if (bracket == NULL) {
    // not a template
    return;
  }

  templatename.catn(datatype, bracket-datatype);

  char *p = bracket;
  String curparam;
  // count brackets to find closing one
  int bcount = 0;
  while (*p) {
    if (*p == '>') {
      bcount--;
      if (bcount == 0) {
        if (curparam.len() > 0)
          paramlist.addItem(new String(curparam));
        p++;
        break;
      }
      curparam.catn(p, 1);
    } else if (*p == '<') {
      if (bcount > 0) curparam.catn(p, 1);
      bcount++;
    } else if (*p == ',') {
      if (bcount == 1) {
        // next parameter
        paramlist.addItem(new String(curparam));
        curparam.trunc(0);
      } else {
        curparam.catn(p, 1);
      }
    } else {
      curparam.catn(p, 1);
    }
    p++;
  }
}

// todo: be much smarter
void TemplatizedFactory::getAliasType(const char *type, String &str) {
  if (STRCASEEQL("String", type)) str = "const char *";
  else str = type;
}

void TemplatizedFactory::getMultipleMethod(ClassFactoryMethod *method, int *pos, int *multiple, int *first) {
  int namematch = 0;
  int found = -1;
  if (first) *first = 0;
  ClassFactoryChunk *chunk = method->getParentBlock()->getParentChunk();
  for (int i=0;i<chunk->getNumBlocks();i++) {
    if (chunk->enumBlock(i)->getBlockType() == CF_BLOCKTYPE_METHOD) {
      ClassFactoryMethod *m = chunk->enumBlock(i)->getMethod();
      if (!STREQL(m->getName(), method->getName())) continue;
      namematch++;
      if (m == method) {
        found = i;
        if (first) *first = (namematch == 1);
      }
    }
  }
  if (pos) *pos = found;
  if (multiple) *multiple = namematch;
}

int TemplatizedFactory::endsIn(const char *s, const char *txt) {
  return STRCASEEQL(s+STRLEN(s)-STRLEN(txt), txt);
}

int TemplatizedFactory::getLastLinePos(const char *s) {
  int count = 0;
  const char *p = s + STRLEN(s)-1;
  // find last CR
  while (p >= s) {
    if (ClassFactoryParser::isOneOf(*p, "\r\n")) break;
    count++;
    p--;
  }
  return count;
}

const char *TemplatizedFactory::getClassLayer() {
  static String val;
  if (m_curformat.iscaseequal("inlinexml")) val = getChunk()->getImplementationName();
  else val = getChunk()->getAncestor();
  return val;
}
