#ifndef ____TEMPLETIZEDFACTORY_H
#define ____TEMPLETIZEDFACTORY_H

/*<?<autoheader/>*/
class BigString;
class ClassFactoryChunk;
class ClassFactoryCodeBlock;
class ClassFactoryData;
class ClassFactoryMethod;
class ClassFactoryParam;
class FactoryModule;
class FactoryTemplate;
class ForEachAncestor;
class ForEachCodeBlock;
class ForEachData;
class ForEachHelper;
class ForEachHelperFile;
class ForEachImplFile;
class ForEachMethod;
class ForEachParam;
class ForEachClass;
class ForEachMapEntries;
class ForEachProjectFolder;
#ifdef ODE_IDE
class ForEachGroupDefFolder;
#endif
class ForIterator;
class ModuleDescription;
class ProgressCallback;
class XmlReaderParams;
/*?>*/

class ProjectFile;

#include "FactoryTemplate.h"
#include "FactoryModule.h"

#include "OutputFactoryBase.h"
#include <bfc/stack.h>
#include <bfc/ptrlist.h>
#include <bfc/string/string.h>
#include <bfc/string/bigstring.h>
#include "../inputfactories/XPPXmlReaderI.h"
#include "TypeMap.h"
#include "GlobalStrings.h"
#include "Cast.h"

using namespace wasabi;

enum {
  FT_UNKNOWN=-1,
  FT_CDATA,
  FT_ROOT,
  FT_PROC,
  FT_CALL,
  FT_BR,
  FT_OB,
  FT_CB,
  FT_MODULE,
  FT_MODULEDATA,
  FT_CLASS,
  FT_CLASSDECL,
  FT_CODEBLOCK,
  FT_METHOD,
  FT_PARAM,
  FT_TEMPLATEPARAM,
  FT_DATA,
  FT_INDENT,
  FT_FOREACH,
  FT_IF,
  FT_ELSE,
  FT_FORCOMMA,
  FT_FORINDEX,
  FT_ADDPARAMCOMMA,
  FT_DATETIME,
  FT_SKIP,
  FT_BREAK,
  FT_CONTINUE,
  FT_RESETSCOPE,
  FT_IMPLEMENTATION,
  FT_INTERFACE,
  FT_AUTOHEADER,
  FT_DERIVEANCESTOR,
  FT_ANCESTOR,
  FT_THISCLASS,
  FT_HELPERFILE,
  FT_PROJECT,
  FT_DATATEMPLATE,
  FT_STDAFX,
  FT_EXTERN,
  FT_GETGLOBAL,
  FT_SETGLOBAL,
  FT_T,
  FT_GET,
  FT_SET,
  FT_PGET,
  FT_PSET,
  FT_TYPEMAP,
  FT_MAP,
  FT_TYPECAST,
  FT_CAST,
  FT_GUID,
  FT_CGUID,
  FT_PROJECTFOLDER,
  FT_INC,
  FT_DEC,
  FT_MAPENTRY,
  FT_YEAR,
  FT_MONTH,
  FT_MONTHDAY,
  FT_MONTHNAME,
  FT_WEEKDAY,
  FT_HOUR,
  FT_MINUTE,
  FT_SECOND,
  FT_ERROR,
  FT_WARNING,
#ifdef ODE_IDE
  FT_PACKAGEDFILE,
#endif
};

enum {
  FT_MODULEDATA_FILENAME,
  FT_MODULEDATA_ID,
};

enum {
  FT_CLASS_GUID,
  FT_CLASS_CGUID,
  FT_CLASS_LAYERGUID,
  FT_CLASS_LAYERCGUID,
  FT_CLASS_NAME,
  FT_CLASS_IMPLNAME,
  FT_CLASS_INTERFACEH,
  FT_CLASS_INTERFACECPP,
  FT_CLASS_IMPLEMENTATIONH,
  FT_CLASS_IMPLEMENTATIONCPP,
  FT_CLASS_TARGETDIR,
  FT_CLASS_DISPATCHH,
  FT_CLASS_DISPATCHCPP,
  FT_CLASS_CLASSLAYER,
  FT_CLASS_LAYERANCESTOR,
  FT_CLASS_NUMMETHODS,
};

enum {
  FT_METHOD_NAME,
  FT_METHOD_RETURNTYPE,
  FT_METHOD_DEFRETVAL,
  FT_METHOD_NUMPARAMS,
  FT_METHOD_VIRTUAL,
  FT_METHOD_STATIC,
  FT_METHOD_INLINE,
  FT_METHOD_PURE,
  FT_METHOD_SCOPE,
  FT_METHOD_DISPATCHID,
  FT_METHOD_DISPATCHSID,
  FT_METHOD_RETURN,
  FT_METHOD_DISPATCHCALL,
  FT_METHOD_DISPATCHDEFRETVAL,
  FT_METHOD_DISPATCHCB,
  FT_METHOD_MULTIPLEID,
  FT_METHOD_SOURCELINE,
};

enum {
  FT_CODEBLOCK_BLOCK,
  FT_CODEBLOCK_SCOPE,
};

enum {
  FT_PARAM_NAME,
  FT_PARAM_TYPE,
  FT_PARAM_ALIASTYPE,
  FT_PARAM_PTRREF,
  FT_PARAM_IO,
  FT_PARAM_DEFVAL,
};

enum {
  FT_DATA_NAME,
  FT_DATA_TYPE,
  FT_DATA_CTORCODE,
  FT_DATA_DTORCODE,
  FT_DATA_ALIASTYPE,
  FT_DATA_SOURCELINE,
};

enum {
  FT_MAPENTRY_FROM,
  FT_MAPENTRY_TO,
  FT_MAPENTRY_IMPL,
};

enum {
  FT_DATATYPE_TEMPLATE,
  FT_DATATYPE_TEMPLATIZED,
  FT_DATATYPE_CLASS,
  FT_DATATYPE_PTR,
  FT_DATATYPE_REF,
  FT_DATATYPE_ALIASED,
  FT_DATATYPE_IN,
  FT_DATATYPE_OUT,
  FT_DATATYPE_INOUT,
  FT_DATATYPE_SCRIPT,
};

enum {
  FT_METHODTYPE_CTOR,
  FT_METHODTYPE_DEFCTOR,
  FT_METHODTYPE_DTOR,
  FT_METHODTYPE_VIRTUAL,
  FT_METHODTYPE_STATIC,
  FT_METHODTYPE_INLINE,
  FT_METHODTYPE_PURE,
  FT_METHODTYPE_RAW,
  FT_METHODTYPE_DISPATCH,
  FT_METHODTYPE_NODISPATCH,
  FT_METHODTYPE_FIRSTOFMULTIPLE,
  FT_METHODTYPE_MULTIPLE,
  FT_METHODTYPE_HASRETIREDVERSIONS,
  FT_METHODTYPE_ISRETIRED,
  FT_METHODTYPE_ISVOID,
  FT_METHODTYPE_ISVOIDPARAM,
  FT_METHODTYPE_SCRIPT,
  FT_METHODTYPE_EVENT,
  FT_METHODTYPE_PUBLIC,
  FT_METHODTYPE_PRIVATE,
  FT_METHODTYPE_PROTECTED,
};

enum {
  FT_CLASSTYPE_INTERFACED,
};

enum {
  FT_HELPERFILETYPE_CPP,
  FT_HELPERFILETYPE_H,
};

enum {
  FT_HELPERFILEDATA_FILENAME,
};

enum {
  FT_PROJECT_NAME,
};

#ifdef ODE_IDE
enum {
  FT_PROJECTFOLDER_NAME,
  FT_PROJECTFOLDER_PATH,
};

enum {
  FT_PACKAGEDFILE_DATA_FILE,
  FT_PACKAGEDFILE_DATA_DIR,
  FT_PACKAGEDFILE_DATA_FOLDER,
};
#endif

enum {
  FOREACH_RETIREDMETHOD=1000,
};

class TSortString {
public:
  static int compareItem(String *p1, String *p2) {
    return STRICMP(p1->getValue(), p2->getValue());
  }
  static int compareAttrib(const char *attrib, String *item) {
    return STRICMP(attrib, item->getValue());
  }
};

class ModuleDescription {
  public:
    String m_moduleId;
    String m_moduleDefName;
};


class TemplatizedFactory : public OutputFactoryBase, public XPP_XmlReaderCallbackI {
  friend OutputFactoryBase;
  public:
  TemplatizedFactory();
  virtual ~TemplatizedFactory();

  virtual void onNewChunk(ClassFactoryChunk *chunk) {};

  virtual void setExt(const char *ext);
  virtual void registerModule(FactoryModule *module);
  virtual void setName(const char *name);
  virtual int tryFormat(const char *format);
  virtual const char *getCustomFormat(const char *format);
  virtual const char *getCustomClassData(const char *data, XmlReaderParams *params);
  virtual int getCustomSkip();

  virtual void onCustomTag(const char *tag, XmlReaderParams *cb, int isclose, int discard);
  virtual ForIterator *getForIterator();
  virtual void addToCurData(const char *str);
  virtual int continueByType(int type);
  virtual int breakByType(int type);
  virtual int testMatchesType(int fortype, const char *type, XmlReaderParams *params);

  virtual void setProgressCallback(ProgressCallback *cb);
  virtual const char *getName();
  virtual const char *getDescription();
  virtual int getNumModules();
  virtual const char *enumModuleName(int i);
  virtual const char *enumModuleData(int i);
  virtual int process();
  virtual void setChunk(ClassFactoryChunk *chunk);
  virtual void setChunk_noreset(ClassFactoryChunk *chunk);
  virtual ClassFactoryChunk *getChunk();

  virtual void pushForeach(ForIterator *iter);
  virtual ForIterator *popForeach();
  virtual ForIterator *createForEach(const char *objects, XmlReaderParams *params, ClassFactoryChunk *chunk, ClassFactoryMethod *method, ClassFactoryParam *param, String &txt);
  ForIterator *getLastIteratorByType(int type, int *_pos = NULL);
  ForIterator *getLastIterator();
  virtual void resetFactory();
  virtual int wantAutoResetTemplates() { return 0; }
  virtual int needFriend();

  ForEachMethod *getForEachMethod();
  ForEachMapEntries *getForEachMapEntries();
  ForEachParam *getForEachParam();
  ForEachData *getForEachData();
  ForEachCodeBlock *getForEachCodeBlock();
  ForEachHelper *getForEachHelper();
  ForEachHelperFile *getForEachHelperFile();
  ForEachAncestor *getForEachAncestor();
  ForEachImplFile *getForEachImplFile();
  ForEachClass *getForEachClass();
  ForEachProjectFolder *getForEachProjectFolder();
#ifdef ODE_IDE
  ForEachGroupDefFolder *getForEachGroupDefFolder();
#endif

  virtual ClassFactoryCodeBlock *getCurrentCodeBlock();
  virtual ClassFactoryMethod *getCurrentMethod();
  virtual ClassFactoryParam *getCurrentParam();
  virtual ClassFactoryParam *getCurrentTemplateParam();
  virtual ClassFactoryData *getCurrentData();
  virtual int getCurrentMapEntry();
#ifdef ODE_IDE
  virtual ProjectFile *getCurrentPackagedFile();
#endif

  int getDiscardCondition();


  public:

  virtual void xmlReaderOnStartElementCallback(const char *xmlpath, const char *xmltag, XmlReaderParams *params);
  virtual void xmlReaderOnEndElementCallback(const char *xmlpath, const char *xmltag);
  virtual void xmlReaderOnCharacterDataCallback(const char *xmlpath, const char *xmltag, const char *str);
  virtual void unloadAllTemplates();
  virtual int processFormat();
  virtual FactoryTemplate *insertFormatTemplate(const char *format, const char *tplt, const char *wantnew_format = NULL);
  virtual int loadFormatTemplate(const char *format, const char *subdir = NULL, const char *ext = NULL);
  virtual FactoryTemplate *loadDataTemplate(const char *format, int testonly = 0);
  virtual int newDispatchableId(ClassFactoryChunk *chunk);
  virtual FactoryModule *getModule(const char *format, FactoryTemplate *parent);
  virtual FactoryModule *getProc(const char *id);

  const char *translateVariables(const char *str);
  int translateVariablesInt(const char *str, int def = 0);

  static void parseTemplate(const char *datatype, String &templatename, PtrList < String > &paramlist);
  static void getAliasType(const char *type, String &str);
  static void getMultipleMethod(ClassFactoryMethod *method, int *pos, int *multiple, int *first);

  virtual GlobalStrings *getVars() { return &m_Vars; }
  int testInherit(ClassFactoryChunk *chunk, const char *inherit);

  TemplateChunk *getTemplateChunk() { return m_tempTemplateChunk; }

  protected:

  private:

  int endsIn(const char *s, const char *txt);
  int getLastLinePos(const char *s);
  const char *getClassLayer();
  virtual ClassFactoryMethod *getRootMethod() {return NULL;}

  ClassFactoryChunk * m_chunk;
  PtrList < FactoryModule > m_modules;
  PtrList < FactoryModule > m_auto_modules;
  PtrList < FactoryModule > m_procs;
  FactoryTemplate * m_template;
  FactoryTemplate * m_curTemplate;
  Stack<ForIterator*> m_forstack;
  Stack<FactoryTemplate*> m_templatestack;
  Stack<BigString *> m_outputstack;
  Stack<ClassFactoryData *> m_dataoverridestack;
  Stack < int > m_gotcrstack;
  Stack < int > m_posstack;
  ClassFactoryData * m_curdataoverride;
  TypeMap * m_curTypeMap;
  int m_templatepos;
  String m_curformat;
  int m_intemplate;
  BigString * m_curdata;
  int m_gotcr;
  int m_skip;
  int m_continueskip;
  Stack<ModuleDescription*> m_moduledescstack;
  ModuleDescription * m_curmoduledesc;
  int m_curscope;
  ProgressCallback * m_ppercent;
  String m_ext;
  String m_originalAncestor;
    int m_needfriend;
  String m_name;
  int m_inelse;
  Stack < int > m_controlstack;
  int m_control_discard;
  String m_translatereturn;
  int n_template_chunks;
  int m_running;
  int m_proc;
  GlobalStrings m_Vars;
  int m_parserinmodule;
  int m_inherit_parallel;
  TemplateChunk *m_tempTemplateChunk;
};

class ancestordef {
public:
  ancestordef(const char *_name, const char *_scope) : name(_name), scope(_scope) {};
  String scope;
  String name;
};

#endif
