#ifndef _TYPEMAP_H
#define _TYPEMAP_H

#include <bfc/string/string.h>
#include <api/xml/xmlparamsi.h>

class TypeMap;
class TypeMapMapping;
class TemplatizedFactory;

class TypeMapEntry {
public:
  TypeMapEntry(TypeMap *parent, const char *type, const char *mapping, const char *impl) {
    m_type = type;
    m_parent = parent;
    m_mapping = mapping;
    m_impl = impl;
  }
  virtual ~TypeMapEntry() { }
  const char *getType() { return m_type; }
  void setType(const char *type);
  const char *getMapping(int impl) { return (impl && !m_impl.isempty()) ? m_impl : m_mapping; }
  void setMapping(const char *mapping, const char *impl) { m_mapping = mapping; m_impl = impl; }
private:
  String m_type;
  String m_mapping;
  String m_impl;
  TypeMap *m_parent;
};

class TypeMapEntrySort {
public:
  static int compareItem(TypeMapEntry *p1, TypeMapEntry* p2) { return STRCMP(p1->getType(), p2->getType()); }
  static int compareAttrib(const char *attrib, TypeMapEntry *item) { return STRCMP(attrib, item->getType()); }
};

class TypeMap {
public:
	TypeMap(TemplatizedFactory *parent, const char *id) { m_parent = parent; m_id = id; }
	virtual ~TypeMap() { m_entries.deleteAll(); }

  TypeMapEntry *getMap(const char *type) {
    return m_entries.findItem(type);
  }
  const char *getId() { return m_id; }
  void addMapping(const char *type, const char *mapping, const char *impl);
  void setDefaultMapping(const char *mapping, const char *impl) { m_default = mapping; m_defimpl = impl; }
  const char *getDefaultMapping() { return m_default; }
  void resort() { m_entries.sort(); }
  int getNumMappings() { return m_entries.getNumItems(); }
  const char *enumMappingFrom(int n);
  const char *enumMappingTo(int n);
  const char *enumMappingImpl(int n);

private:
  PtrListQuickMultiSorted<TypeMapEntry, TypeMapEntrySort> m_entries;
  TemplatizedFactory *m_parent;
  String m_id;
  String m_default;
  String m_defimpl;
};

class TypeMapSort {
public:
  static int compareItem(TypeMap *p1, TypeMap *p2) { return STRCMP(p1->getId(), p2->getId()); }
  static int compareAttrib(const char *attrib, TypeMap *item) { return STRCMP(attrib, item->getId()); }
};

class TypeMapMgr {
public:
  TypeMapMgr() { }
  virtual ~TypeMapMgr() { m_typeMaps.deleteAll(); }
  virtual TypeMap *getTypeMap(const char *id) { return m_typeMaps.findItem(id); }
  virtual void addTypeMap(TypeMap *tm) { if (getTypeMap(tm->getId())) return; m_typeMaps.addItem(tm); }
  PtrListQuickSorted<TypeMap, TypeMapSort> *getMaps() { return &m_typeMaps; }
private:
  PtrListQuickSorted<TypeMap, TypeMapSort> m_typeMaps;
};

#endif
