/* ---------------------------------------------------------------------------
                           Nullsoft Database Engine
                             --------------------
                        codename: Near Death Experience
--------------------------------------------------------------------------- */

/* ---------------------------------------------------------------------------

 Virtual File System

--------------------------------------------------------------------------- */

#include "precomp.h"
//#include <stdio.h>
#ifdef WIN32
#include "io.h"
#else
#include <sys/types.h>
#include <sys/stat.h>
#endif
#include "Vfs.h"
#include "bfc/std.h"

//----------------------------------------------------------------------------
VFILE *Vfopen_mem(const char *mem, const char *mode, int size) {
  return _Vfopen(NULL, mode, TRUE, mem, size);
}

//----------------------------------------------------------------------------
VFILE *Vfopen(const char *fl, const char *mode, BOOL Cached) {
  return _Vfopen(fl, mode, Cached, 0, 0);
}

//----------------------------------------------------------------------------
VFILE *_Vfopen(const char *fl, const char *mode, BOOL Cached, const char *mem, const int size) {
  VFILE *f = (VFILE *)MALLOC(sizeof(VFILE));
  if (!f)
    return NULL;
  f->mode = 0;
  f->data = NULL;
  f->filesize = 0;
  f->maxsize = 0;
  f->ptr = 0;
  f->cached = Cached;
  if (!strchr(mode, '+'))  {
    if (strchr(mode, 'r'))
      f->mode = VFS_READ | VFS_MUSTEXIST;
    if (strchr(mode, 'w'))
      f->mode = VFS_WRITE | VFS_CREATE | VFS_NEWCONTENT;
    if (strchr(mode, 'a'))
      f->mode = VFS_WRITE | VFS_CREATE | VFS_SEEKEOF;
  } else {
    if (strstr(mode, "r+"))
      f->mode = VFS_WRITE | VFS_MUSTEXIST;
    if (strstr(mode, "w+"))
      f->mode = VFS_WRITE | VFS_CREATE | VFS_NEWCONTENT;
    if (strstr(mode, "a+"))
      f->mode = VFS_WRITE | VFS_CREATE | VFS_SEEKEOF;
  }
  if (f->mode == 0 || ((f->mode & VFS_READ) && (f->mode & VFS_WRITE))) {
    FREE(f);
    return NULL;
  }

  if (!fl) {
    f->data = (unsigned char *)CALLOC(size+1, 1);
    if (mem != NULL) MEMCPY(f->data, mem, size);
    f->filesize = size;
    f->maxsize = size;
    f->rfile = NULL;
    f->filename = NULL;
  } else {
    if (!f->cached) {
      f->rfile = fopen(fl, mode);
      if (f->rfile)
        f->filename = STRDUP(fl);
      else {
        FREE(f);
        return NULL;
      }
    return f;
    }

    if (f->mode & VFS_MUSTEXIST)
      if (access(fl, 0)) {
        FREE(f);
        return NULL;
    }

    if (!(f->mode & VFS_NEWCONTENT)) {
      FILE *pf;
      pf = fopen(fl, "rb");
      if (!pf) {
        f->data = (unsigned char *)CALLOC(VFILE_INC, 1);
        if (f->data == NULL) {
          FREE(f);
          return NULL;
        }
        f->filesize = 0;
        f->maxsize = VFILE_INC;
      } else {
        fseek(pf, 0, SEEK_END);
        f->filesize = ftell(pf);
        fseek(pf, 0, SEEK_SET);
        f->data = (unsigned char *)CALLOC(f->filesize+1, 1);
        if (f->data == NULL) {
          FREE(f);
          return NULL;
        }
        f->maxsize = f->filesize;
        fread(f->data, f->filesize, 1, pf);
        fclose(pf);
      }
    }

    if (f->mode & VFS_SEEKEOF)
      f->ptr = f->filesize;

    f->filename = STRDUP(fl);
  }

  return f;
}

//----------------------------------------------------------------------------
void Vfclose(VFILE *f) {
  if (!f) return;

  if (f->filename == NULL) {
    FREE(f->data);
    FREE(f);
    return;
  }

  if (!f->cached) {
    if (f->filename)
    FREE(f->filename);
    fclose(f->rfile);
    FREE(f);
    return;
  }

  if (!(f->mode & VFS_WRITE)) {
  if (f->filename)
    FREE(f->filename);
    FREE(f->data);
    FREE(f);
    return;
  }

  FILE *pf = fopen(f->filename, "wb");

  if (!pf)
    return;

  fwrite(f->data, f->filesize, 1, pf);
  fclose(pf);

  if (f->filename) {
    FREE(f->filename);
    FREE(f->data);
    FREE(f);
  }
}

//----------------------------------------------------------------------------
void Vfread(void *ptr, size_t size, size_t n, VFILE *f) {
  if (!ptr || !f) return;
  if (!f->cached) {
  fread(ptr, size, n, f->rfile);
    return;
  }
  int s = size*n;
  if (s + f->ptr > f->filesize) {
    //FUCKO: remove this
    if(! (f->ptr < f->filesize)) {
      #ifdef WIN32
      MessageBox(NULL,"NDB bug: vfread in vfs.cpp","uhh",0);
      #else
      printf("vfs bug: vfread in vfs.cpp\n");
      #endif
    }
    s = f->filesize - f->ptr;
  }
  MEMCPY(ptr, f->data+f->ptr, s);
  f->ptr += s;
}

//----------------------------------------------------------------------------
void Vfwrite(void *ptr, size_t size, size_t n, VFILE *f) {
  if (!ptr || !f) return;
  if (!f->cached) {
    fwrite(ptr, size, n, f->rfile);
    return;
  }
  int s = size*n;
  if (s + f->ptr > f->maxsize) {
    // grow f->data,f->maxsize to be (s + f->ptr + VFILE_INC-1)&~(VFILE_INC-1)
    // instead of calling Vgrow again which gets kinda slow
    int newsize=(s + f->ptr + VFILE_INC-1)&~(VFILE_INC-1);
    f->data=(unsigned char *)REALLOC(f->data,newsize);
    if (f->data == NULL) return;
    MEMSET(f->data+f->maxsize,0,newsize-f->maxsize);
    f->maxsize=newsize;
  }
  MEMCPY(f->data + f->ptr, ptr, s);
  f->ptr += s;
  if (f->ptr > f->filesize)
    f->filesize = f->ptr;
}

//----------------------------------------------------------------------------
void Vgrow(VFILE *f) {
  if (!f || !f->cached)
    return;
  unsigned char *newBlock = (unsigned char *)CALLOC(f->maxsize + VFILE_INC, 1);
  if (f->data)
    MEMCPY(newBlock, f->data, f->maxsize);
  f->maxsize += VFILE_INC;
  if (f->data)
    FREE(f->data);
  f->data = newBlock;
}

//----------------------------------------------------------------------------
void Vfputs(char *str, VFILE *f) {
  if (!f) return;
  if (!f->cached) {
    fputs(str, f->rfile);
    return;
  }
  Vfwrite(str, strlen(str), 1, f);
}

//----------------------------------------------------------------------------
void Vfputc(char c, VFILE *f) {
  if (!f) return;
  if (!f->cached) {
    fputc(c, f->rfile);
    return;
  }
  Vfwrite(&c, 1, 1, f);
}

// not mac comliant
//----------------------------------------------------------------------------
char *Vfgets(char *dest, int n, VFILE *f) {
  if (!f) return NULL;
  if (!f->cached)
    return fgets(dest, n, f->rfile);

  unsigned char c=0;
  char *p;
  int l=0;

  p = dest;
  while (l < n && !Vfeof(f)) {
    c = f->data[f->ptr];
    f->ptr++;
    *p = c;
    p++;
    l++;
    if (c == '\n') {
      if (!Vfeof(f) && f->data[f->ptr] == '\r') {
        f->ptr++;
      }
    break;
    }
  }
  *p=0;
  return dest;
}

//----------------------------------------------------------------------------
char Vfgetc(VFILE *f) {
  if (!f) return EOF;
  if (!f->cached)
    return fgetc(f->rfile);
  if (!Vfeof(f))
    return f->data[f->ptr++];
  return EOF;
}

//----------------------------------------------------------------------------
unsigned long Vftell(VFILE *f) {
  if (!f) return (unsigned long)-1;
  if (!f->cached)
    return ftell(f->rfile);
  return f->ptr;
}

//----------------------------------------------------------------------------
void Vfseek(VFILE *f, long i, int whence) {
  if (!f) return;
  if (!f->cached) {
    fseek(f->rfile, i, whence);
    return;
  }
  switch (whence) {
    case SEEK_SET:
      f->ptr = i;
      break;
    case SEEK_CUR:
      f->ptr += i;
      break;
    case SEEK_END:
      f->ptr = f->filesize+i;
      break;
  }
}

//----------------------------------------------------------------------------
int Vfeof(VFILE *f) {
  if (!f) return -1;
  if (!f->cached)
  return feof(f->rfile);
  return (f->ptr >= f->filesize);
}

//----------------------------------------------------------------------------
void Vsync(VFILE *f) {
  if (!f) return;
  if (f->mode & VFS_WRITE) {
    if (!f->cached) {
      int p=ftell(f->rfile);
      fclose(f->rfile);
      f->rfile = fopen(f->filename, "r+b");
      if (!f->rfile)
        return;
      fseek(f->rfile, p, SEEK_SET);
      return;
    }
    FILE *pf = fopen(f->filename, "wb");
    if (!pf)
      return;
    fwrite(f->data, f->filesize, 1, pf);
    fclose(pf);
  }
}

//----------------------------------------------------------------------------
const char *Vfgetmem(VFILE *f) {
  if (!f) return NULL;
  if (!f->cached) return NULL;
  return (const char *)f->data;
}
