#include "precomp.h"
#include "XHDFactory.h"

#include "../classhierarchy/ClassFactoryChunkI.h"
#include "../outputfactories/XHDWriter.h"

// we need to work on the UI format a bit more to make
// this a pair of checkboxes.
static const char *FORMAT_NAME = "XHD";

// The service factory is asking us if we can handle a specific format.
// Return true or false.
int XHDFactory::hasFormat(const char *format) {
  if (STRCASEEQLSAFE(format, FORMAT_NAME)) return 1;
  return 0;
}

// We are being asked how many formats we support
int XHDFactory::getNumFormats() {
  return 1;
}

// We are being asked for a human readable string for our supported formats
const char *XHDFactory::enumFormatName(int i) {
  switch (i) {
    case 0:
      return FORMAT_NAME;
  }
  return NULL;
}

// We are being asked for how many output files we created
int XHDFactory::getNumModules() {
  if (!xhdData.isempty()) return 1;
  return 0;
}

// We are being asked to return the filename for a specific module.
const char *XHDFactory::enumModuleName(int i) {
  switch (i) {
    case 0:
      return xhdName;
  }
  return NULL;
}

// We are being asked to return the generated data for a specific module.
const char *XHDFactory::enumModuleData(int i) {
  switch (i) {
    case 0:
      return xhdData;
  }
  return NULL;
}

void XHDFactory::setChunk(ClassFactoryChunk *_chunk) {
  OutputFactoryBase::setChunk(_chunk);
  m_ppercent = NULL;
  chunk = _chunk;
}

// We are being told to process.  Call the proper process method
// based on what format we're being told to create.
int XHDFactory::process(const char *format) {
  if (STRCASEEQLSAFE(format, FORMAT_NAME)) return processXHD(chunk);
  return 0;
}

int XHDFactory::processXHD(ClassFactoryChunk *chunk) {
  xhdData = XHDWriter::convert(chunk, NULL, m_ppercent);
  xhdName = StringPrintf("%s.xhd", chunk->getClassName());
  return 1;
}

