#include "precomp.h"
#include "XHDWriter.h"
#include "../classhierarchy/ClassFactoryChunk.h"
#include "../classhierarchy/ClassFactoryMethod.h"
#include "../classhierarchy/ClassFactoryParam.h"
#include "../classhierarchy/ClassFactoryBlock.h"
#include "../classhierarchy/ClassFactoryData.h"
#include "../classhierarchy/ClassFactoryCodeBlock.h"
#include "../classhierarchy/HelpSection.h"
#include "../classhierarchy/HelpElement.h"
#include <api/xml/xmlparams.h>
#include <bfc/string/url.h>
#include <bfc/string/stringdict.h>
#include "../classhierarchy/Project.h"
#include "../classhierarchy/SecondaryAncestor.h"

const char *XHDWriter::paramToText(const char *name, const char *value, String &dest, const char *prfx) {
  dest = "";
  String s = name;
  if (*s.getValue() == '$') return NULL;
  dest += prfx;
  if (!prfx) dest += " ";
  dest += s;
  dest += "=\"";
  String ss = value;
  Url::encode(ss, FALSE, URLENCODE_EXCLUDE_7BIT_ABOVEEQ32|URLENCODE_ENCODEXML);
  dest += ss;
  dest += "\"";
  return dest;
}


const char *XHDWriter::paramsToText(XmlReaderParams *p, const char *prfx, XmlReaderParams *raw_params) {
  String tmp;
  PtrListQuickSorted<String, StringComparator> written_params;

  pttstr.reset();
  if (raw_params) {
    for (int j=0;j<raw_params->getNbItems();j++) {
      const char *cs = paramToText(raw_params->getItemName(j), raw_params->getItemValue(j), tmp, prfx);
      if (!cs) continue;
      pttstr += cs;
      written_params.addItem(new String(raw_params->getItemName(j)));
    }
  }
  for (int j=0;j<p->getNbItems();j++) {
    if (raw_params && written_params.findItem(p->getItemName(j))) continue;
    pttstr += paramToText(p->getItemName(j), p->getItemValue(j), tmp, prfx);
  }
  written_params.deleteAll();
  return pttstr;
}

extern void clear_stdin(void);
extern int xpp_confirmRewrite(const char *filename);

const char *XHDWriter::convert(ClassFactoryChunk *c, const char *file, ProgressCallback *percent, ChunkCollection *collection) {
  m_chunkcollection = collection;
  m_ppercent = percent;
  xmlData = "";
  if (xpp_confirmRewrite(file)) {
    FILE *o = NULL;
    if (file != NULL)
      o = fopen(file, "wt");
    BigString s;
    s += "<xhd>\n";

    if (c->getHeader("h")) {
      s += "  <cpp type=\"h\">";//\n";
      s += "<![CDATA[";//\n";
      String cb = c->getHeader("h")->getCodeBlock();
      fixupCData(cb);
      s += cb;
      s += "]]>";//\n";
      s += "</cpp>\n";
    }
    if (c->getHeader("cpp")) {
      s += "  <cpp type=\"cpp\">";//\n";
      s += "<![CDATA[";//\n";
      String cb = c->getHeader("cpp")->getCodeBlock();
      fixupCData(cb);
      s += cb;
      s += "]]>";//\n";
      s += "</cpp>\n";
    }

    s += "  <class";
    String cn = c->getClassName();
    Url::encode(cn, FALSE, URLENCODE_EXCLUDE_7BIT_ABOVEEQ32|URLENCODE_ENCODEXML);
    s += " name=\""; s += cn; s += "\"";
    String impl = c->getImplementationName();
    if (!impl.isempty()) {
      Url::encode(impl, FALSE, URLENCODE_EXCLUDE_7BIT_ABOVEEQ32|URLENCODE_ENCODEXML);
      s += "\n         implementation=\""; s += impl; s += "\"";
    }
    String ance = c->getAncestor();
    if (!ance.isempty()) {
      Url::encode(ance, FALSE, URLENCODE_EXCLUDE_7BIT_ABOVEEQ32|URLENCODE_ENCODEXML);
      s += "\n         ancestor=\""; s += ance; s += "\"";
    }
    s += paramsToText(c->getParams(), "\n         ", c->getRawParams());
    int _c, nc = c->getNumSecondaryAncestors();
    String ancs;
    for (_c=0;_c<nc;_c++) {
      if (!ancs.isempty()) ancs += ";";
      switch (c->enumSecondaryAncestor(_c)->getScope()) {
        case SCOPE_PUBLIC:
          ancs += "public ";
          break;
        case SCOPE_PRIVATE:
          ancs += "private ";
          break;
        case SCOPE_PROTECTED:
          ancs += "protected ";
          break;
      }
      ancs += c->enumSecondaryAncestor(_c)->getClassName();
    }
    if (!ancs.isempty()) {
      Url::encode(ancs, FALSE, URLENCODE_EXCLUDE_7BIT_ABOVEEQ32|URLENCODE_ENCODEXML);
      s += " inherit=\"";
      s += ancs;
      s += "\"";
    }
    s += ">\n";
    int nhs = c->getNumHelpSections();
    int hs;
    for (hs=0;hs<nhs;hs++) {
      HelpSection *section = c->enumHelpSection(hs);
      if (section == NULL) continue;
      s += "    <helpsection name=\"";
      String sn = section->getName();
      Url::encode(sn, FALSE, URLENCODE_EXCLUDE_7BIT_ABOVEEQ32|URLENCODE_ENCODEXML);
      s += sn;
      s += "\">\n";
      int nhe = section->getNumElements();
      int he;
      for (he=0;he<nhe;he++) {
        HelpElement *element = section->getElement(he);
        if (element == NULL) continue;
        s += "      <helpelement name=\"";
        String en = element->getName();
        Url::encode(en, FALSE, URLENCODE_EXCLUDE_7BIT_ABOVEEQ32|URLENCODE_ENCODEXML);
        s += en;
        s += "\">";
        s += "<![CDATA[";
        s += element->getHtml();
        s += "]]>";
        s += "</helpelement>\n";
      }
      s += "    </helpsection>\n";
    }

    if (c->getHeader("class")) {
      s += "    <cpp type=\"class\">";//\n";
      s += "<![CDATA[";//\n";
      String cb = c->getHeader("class")->getCodeBlock();
      fixupCData(cb);
      s += cb;
      s += "]]>";//\n";
      s += "</cpp>\n";
    }

    int nb = c->getNumBlocks();
    for (int i=0;i<nb;i++) {
      if (m_ppercent) m_ppercent->setPercent((int)(((float)i / (float)nb) * 100));
      ClassFactoryBlock *b = c->enumBlock(i);
      switch (b->getBlockType()) {
        case CF_BLOCKTYPE_METHOD: {
          ClassFactoryMethod *m = b->getMethod();
          s += "    <method";
          String mn = m->getName();
          Url::encode(mn, FALSE, URLENCODE_EXCLUDE_7BIT_ABOVEEQ32|URLENCODE_ENCODEXML);
          s += " name=\""; s += mn; s += "\"";
          const char *rtype = m->getReturnType();
          if (rtype && *rtype) {
            String rt = rtype;
            Url::encode(rt, FALSE, URLENCODE_EXCLUDE_7BIT_ABOVEEQ32|URLENCODE_ENCODEXML);
            s += " type=\""; s += rt; s += "\"";
          }
          const char *def = m->getDefaultReturnValue();
          if (def && *def) {
            String d = def;
            Url::encode(d, FALSE, URLENCODE_EXCLUDE_7BIT_ABOVEEQ32|URLENCODE_ENCODEXML);
            s += " default=\""; s += d; s += "\"";
          }
          s += paramsToText(m->getParams());
          s += ">\n";
          int nhs = c->getNumHelpSections();
          int hs;
          for (hs=0;hs<nhs;hs++) {
            HelpSection *section = m->enumHelpSection(hs);
            if (section == NULL) continue;
            s += "    <helpsection name=\"";
            String sn = section->getName();
            Url::encode(sn, FALSE, URLENCODE_EXCLUDE_7BIT_ABOVEEQ32|URLENCODE_ENCODEXML);
            s += sn;
            s += "\">\n";
            int nhe = section->getNumElements();
            int he;
            for (he=0;he<nhe;he++) {
              HelpElement *element = section->getElement(he);
              if (element == NULL) continue;
              s += "      <helpelement name=\"";
              String en = element->getName();
              Url::encode(en, FALSE, URLENCODE_EXCLUDE_7BIT_ABOVEEQ32|URLENCODE_ENCODEXML);
              s += en;
              s += "\">";
              s += "<![CDATA[";
              s += element->getHtml();
              s += "]]>";
              s += "</helpelement>\n";
            }
            s += "    </helpsection>\n";
          }
          for (int k=0;k<m->getNumParams();k++) {
            ClassFactoryParam *p = m->enumParam(k);
            s += "      <param";
            String pn = p->getName();
            Url::encode(pn, FALSE, URLENCODE_EXCLUDE_7BIT_ABOVEEQ32|URLENCODE_ENCODEXML);
            String pt = p->getType();
            Url::encode(pt, FALSE, URLENCODE_EXCLUDE_7BIT_ABOVEEQ32|URLENCODE_ENCODEXML);
            s += " name=\""; s += pn; s += "\"";
            s += " type=\""; s += pt; s += "\"";
            def = p->getDefaultValue();
            if (def && *def) {
              String d = def;
              Url::encode(d, FALSE, URLENCODE_EXCLUDE_7BIT_ABOVEEQ32|URLENCODE_ENCODEXML);
              s += " default=\""; s += d; s += "\"";
            }
            s += paramsToText(p->getParams());
            int nhs = c->getNumHelpSections();
            if (nhs == 0) {
              s += "/>\n";
            } else {
              s += ">\n";
              int hs;
              for (hs=0;hs<nhs;hs++) {
                HelpSection *section = p->enumHelpSection(hs);
                if (section == NULL) continue;
                s += "        <helpsection name=\"";
                String sn = section->getName();
                Url::encode(sn, FALSE, URLENCODE_EXCLUDE_7BIT_ABOVEEQ32|URLENCODE_ENCODEXML);
                s += sn;
                s += "\">\n";
                int nhe = section->getNumElements();
                int he;
                for (he=0;he<nhe;he++) {
                  HelpElement *element = section->getElement(he);
                  if (element == NULL) continue;
                  s += "        <helpelement name=\"";
                  String en = element->getName();
                  Url::encode(en, FALSE, URLENCODE_EXCLUDE_7BIT_ABOVEEQ32|URLENCODE_ENCODEXML);
                  s += en;
                  s += "\">";
                  s += "<![CDATA[";
                  s += element->getHtml();
                  s += "]]>";
                  s += "</helpelement>\n";
                }
                s += "        </helpsection>\n";
              }
              s += "      </param>\n";
            }
          }
          const char *hcpp = m->getHeader("cpp");
          if (hcpp != NULL) {
            s += "      <cpp type=\"hcpp\">";//\n";
            s += "<![CDATA[";//\n";
            s += hcpp;
            s += "]]>";//\n";
            s += "</cpp>\n";
          }
          const char *hh = m->getHeader("h");
          if (hh != NULL) {
            s += "      <cpp type=\"hh\">";//\n";
            s += "<![CDATA[";//\n";
            s += hh;
            s += "]]>";//\n";
            s += "</cpp>\n";
          }
          const char *fcpp = m->getFooter("cpp");
          if (fcpp != NULL) {
            s += "      <cpp type=\"fcpp\">";//\n";
            s += "<![CDATA[";//\n";
            s += fcpp;
            s += "]]>";//\n";
            s += "</cpp>\n";
          }
          const char *fh = m->getFooter("h");
          if (fh != NULL) {
            s += "      <cpp type=\"fh\">";//\n";
            s += "<![CDATA[";//\n";
            s += fh;
            s += "]]>";//\n";
            s += "</cpp>\n";
          }
          const char *fi = m->getFooter("inline");
          if (fi != NULL) {
            s += "      <cpp type=\"finline\">";//\n";
            s += "<![CDATA[";//\n";
            s += fi;
            s += "]]>";//\n";
            s += "</cpp>\n";
          }
          const char *hi = m->getHeader("inline");
          if (hi != NULL) {
            s += "      <cpp type=\"hinline\">";//\n";
            s += "<![CDATA[";//\n";
            s += hi;
            s += "]]>";//\n";
            s += "</cpp>\n";
          }
          ClassFactoryCodeBlock *cb = m->getCodeBlock("cpp");
          if (cb != NULL) {
            s += "      <cpp type=\"";
            const char *t = cb->getParams()->getItemValue("type");
            if (t == NULL || *t == 0) t = "cpp";
            s += t;
            s += "\">";//\n";
            s += "<![CDATA[";//\n";
            String scb = cb->getCodeBlock();
            fixupCData(scb);
            s += scb;
            s += "]]>";//\n";
            s += "</cpp>\n";
          }
          cb = m->getCodeBlock("h");
          if (cb != NULL) {
            s += "      <cpp type=\"";
            const char *t = cb->getParams()->getItemValue("type");
            if (t == NULL || *t == 0) t = "h";
            s += t;
            s += "\">";//\n";
            s += "<![CDATA[";//\n";
            String scb = cb->getCodeBlock();
            fixupCData(scb);
            s += scb;
            s += "]]>";//\n";
            s += "</cpp>\n";
          }
          cb = m->getCodeBlock("inline");
          if (cb != NULL) {
            s += "      <cpp type=\"inline\">";//\n";
            s += "<![CDATA[";//\n";
            String scb = cb->getCodeBlock();
            fixupCData(scb);
            s += scb;
            s += "]]>";//\n";
            s += "</cpp>\n";
          }
          s += "    </method>\n";
          break;
        }
        case CF_BLOCKTYPE_CODEBLOCK: {
          ClassFactoryCodeBlock *cb = b->getCodeBlock();
          s += "    <cpp type=\"";
          const char *t = cb->getParams()->getItemValue("type");
          if (t == NULL || *t == 0) t = "cpp";
          s += t;
          s += "\">";//\n";
          s += "<![CDATA[";//\n";
          String scb = cb->getCodeBlock();
          fixupCData(scb);
          s += scb;
          s += "]]>";//\n";
          s += "</cpp>\n";
          break;
        }
        case CF_BLOCKTYPE_DATA: {
          ClassFactoryData *d = b->getData();
          s += "    <data ";
          String dn = d->getName();
          Url::encode(dn, FALSE, URLENCODE_EXCLUDE_7BIT_ABOVEEQ32|URLENCODE_ENCODEXML);
          s += " name=\""; s += dn; s += "\"";
          String dtype = d->getDataType();
          Url::encode(dtype, FALSE, URLENCODE_EXCLUDE_7BIT_ABOVEEQ32|URLENCODE_ENCODEXML);
          s += " type=\""; s += dtype; s += "\"";
          s += paramsToText(d->getParams());
          s += ">\n";
          const char *ct = d->getCtorCode();
          if (ct != NULL && *ct) {
            s += "      <cpp type=\"ctor\">";//\n";
            s += "<![CDATA[";//\n";
            s += ct;
            s += "]]>";//\n";
            s += "</cpp>\n";
          }
          const char *dt = d->getDtorCode();
          if (dt != NULL && *dt) {
            s += "      <cpp type=\"dtor\">";//\n";
            s += "<![CDATA[";//\n";
            s += dt;
            s += "]]>";//\n";
            s += "</cpp>\n";
          }
          s += "    </data>\n";
        }
      }
    }

    s += "  </class>\n";

    if (c->getFooter("class")) {
      s += "    <cpp type=\"class\">";//\n";
      s += "<![CDATA[";//\n";
      String cb = c->getFooter("class")->getCodeBlock();
      fixupCData(cb);
      s += cb;
      s += "]]>";//\n";
      s += "</cpp>\n";
    }
    if (c->getFooter("h")) {
      s += "  <cpp type=\"h\">";//\n";
      s += "<![CDATA[";//\n";
      String cb = c->getFooter("h")->getCodeBlock();
      fixupCData(cb);
      s += cb;
      s += "]]>";//\n";
      s += "</cpp>\n";
    }
    if (c->getFooter("cpp")) {
      s += "  <cpp type=\"cpp\">";//\n";
      s += "<![CDATA[";//\n";
      String cb = c->getFooter("cpp")->getCodeBlock();
      fixupCData(cb);
      s += cb;
      s += "]]>";//\n";
      s += "</cpp>\n";
    }

    s += "</xhd>\n";
    if (o != NULL) {
      fwrite(s, STRLEN(s), 1, o);
      fclose(o);
    }
    xmlData = s;
  }
  return xmlData;
}

void XHDWriter::fixupCData(String &str) {
  char *d = str.getNonConstVal();
  if (d == NULL) return;
  while (*d) {
    if (!STRNCMP(d, "\x3C\x21\x5B\x43\x44\x41\x54\x41", 8))
      *(d+1) = '#';
    if (!STRNCMP(d, "\x5D\x5D\x3E", 3))
      *(d+1) = '#';
    d++;
  }
}

BigString XHDWriter::pttstr;
String XHDWriter::xmlData;
ProgressCallback * XHDWriter::m_ppercent = NULL;
ChunkCollection * XHDWriter::m_chunkcollection = NULL;
