#ifndef _SVC_SRCCLASSFACTORY_H
#define _SVC_SRCCLASSFACTORY_H

#include <bfc/dispatch.h>
#include <bfc/string/string.h>

#include "svc_srcClassFactory.h"
#include <api/service/servicei.h>

/*<?<autoheader/>*/
#include "svc_srcClassFactory.h"
#include "svc_srcClassFactoryX.h"

class Workspace;
class ClassFactoryChunk;
class ProgressCallback;
/*?>*/

/*[interface.header.h]  
#include "studio/services/services.h"
#include "bfc/platform/platform.h"
*/

class RootWnd;
class ClassFactoryChunk;

class svc_srcClassFactoryI : public svc_srcClassFactoryX {
protected:
  svc_srcClassFactoryI() {}
  virtual ~svc_srcClassFactoryI() {}
public:
    /*[interface]public:static FOURCC getServiceType() { return WaSvc::SRCCLASSFACTORY; }*/

    // the following formats are available
  DISPATCH(100) virtual int getNumAvailableFormats()=0;
  DISPATCH(120) virtual const char *enumAvailableFormat(int n)=0;

    // try loading a format
  DISPATCH(30) virtual int tryFormat(const char *format)=0; //

    // multiple factories will return NULL on this until tryFormat has been called
    // it'll then return this string. other factories expose their name thru this call
  DISPATCH(10) virtual const char *getName()=0;

    // set the chunk to operate on
  DISPATCH(70) virtual void setChunk(ClassFactoryChunk *chunk)=0;

    // do the processing
  DISPATCH(130) virtual int process()=0;

    // callback classes
  DISPATCH(80) virtual void setProgressCallback(ProgressCallback *cb)=0;
  DISPATCH(90) virtual void setWorkspace(Workspace *workspace)=0;

    // the members below this line must be called only AFTER process()

    // these are the actual files generated
  DISPATCH(20) virtual const char *getDescription()=0;
  DISPATCH(40) virtual int getNumModules()=0;
  DISPATCH(50) virtual const char *enumModuleName(int i)=0;
  DISPATCH(60) virtual const char *enumModuleData(int i)=0;

    // if this returns 1, the class should make this class friend
  DISPATCH(110) virtual int needFriend()=0;

};

template <class T>
class SrcClassFactoryCreator : public waServiceFactoryT<svc_srcClassFactory, T> {};

template <class T>
class SrcClassFactoryCreatorSingle : public waServiceFactoryTSingle<svc_srcClassFactory, T> {
public:
  svc_srcClassFactory *getHandler() {
    return getSingleService();
  }
};

/*[interface.footer.h]
#include "bfc/svc_enum.h"
#include "bfc/string.h"

class SrcClassFactoryEnum : public SvcEnumT<svc_srcClassFactory> {
public:
  SrcClassFactoryEnum(const char *_format = "") : format(_format) { }
protected:
  virtual int testService(svc_srcClassFactory *svc) {
    // If nobody asks for a specific format, enum all of them, please.
    return (format.isempty() || svc->tryFormat(format));
  }
private:
  String format;
};
*/

#endif // __SVC_SRCCLASSFACTORYI_H
