#include "precomp.h"
#include "pproject.h"
#include <bfc/string/stringdict.h>
#include <bfc/parse/pathparse.h>
#include <api/xml/xmlwrite.h>

enum XPP_TAGS {
  XPP_XPP,
  XPP_CLASS,
};

BEGIN_STRINGDICTIONARY(_XPPTAGS);
SDI("class", XPP_CLASS);
SDI("xpp", XPP_XPP);
END_STRINGDICTIONARY(_XPPTAGS, xppTags);

PreprocProject::PreprocProject() {
}

PreprocProject::~PreprocProject() {
  m_files.deleteAll();
}

int PreprocProject::loadFile(const char *projectfile) {
  inxpp = 0;
  if (ACCESS(projectfile, 0)) {
    printf("%s not found\n", projectfile);
    return 0;
  }

  if (Std::isDirChar(*projectfile)) m_projectdir += DIRCHARSTR;
  PathParser pp(projectfile);
  for (int i=0;i<pp.getNumStrings()-1;i++) m_projectdir += String(pp.enumString(i)) + DIRCHARSTR;

  XmlReader::registerCallback("*", static_cast<XmlReaderCallback*>(this));
  //printf("opening %s\n", projectfile);
  fflush(stdout);
  XmlReader::loadFile(projectfile, NULL, static_cast<XmlReaderCallback*>(this));
  XmlReader::unregisterCallback(static_cast<XmlReaderCallback*>(this));
  return 1;
}

void PreprocProject::writeFile(const char *projectfile) {
  XMLWrite w(projectfile, "xpp");
  foreach(m_files)
    w.pushCategoryAttrib("class", 1);
    w.writeCategoryAttrib("name", m_files.getfor()->getClassname());
    w.writeCategoryAttrib("file", m_files.getfor()->getFilename());
    w.closeCategoryAttrib(1);
  endfor;
  w.popCategory();
}

int PreprocProject::getNumFiles() {
  return m_files.getNumItems();
}

const char *PreprocProject::enumFile(int n) {
  return m_files.enumItem(n)->getFilename();
}

void PreprocProject::xmlReaderOnStartElementCallback(const char *xmlpath, const char *xmltag, XmlReaderParams *params) {
  fflush(stdout);
  switch (xppTags.getId(xmltag)) {
    case XPP_XPP: {
        inxpp++;
        workspace.setPrecomFile(params->getItemValue("precompfile"));
/*        String td = workspace.getTemplateDir();
        if (!td.isempty()) td += ";";
        td += params->getItemValue("templatedir");
        workspace.setTemplateDir(td);*/
      }
      break;
    case XPP_CLASS:
      m_files.addItem(new FileEntry(params->getItemValue("file"), params->getItemValue("name")));
      break;
  }
}

void PreprocProject::xmlReaderOnEndElementCallback(const char *xmlpath, const char *xmltag) {
  switch (xppTags.getId(xmltag)) {
    case XPP_XPP:
      inxpp--;
      break;
  }
}

