#ifndef _PPROJECT_H
#define _PPROJECT_H

#include "inputfactories/XPPXmlReaderI.h"
#include "pworkspace.h"
#include <api/xml/xmlreader.h>

class FileEntry {
public:
  FileEntry(const char *filename, const char *classname) : m_filename(filename),  m_classname(classname) {};
  virtual ~FileEntry() {}
  const char *getFilename() { return m_filename; }
  const char *getClassname() { return m_classname; }
private:
  String m_filename;
  String m_classname;
};

class COMEXP FileEntryComparator {
public:
  // comparator for sorting
  static int compareItem(FileEntry *p1, FileEntry * p2) {
    return STRCMP(p1->getClassname(), p2->getClassname());
  }
  // comparator for searching
  static int compareAttrib(const char *attrib, FileEntry *item) {
    return STRCMP(attrib, item->getClassname());
  }
};

class PreprocProject : public XPP_XmlReaderCallbackI {
public:
  PreprocProject();
  virtual ~PreprocProject();

  PreprocWorkspace *getWorkspace() { return &workspace; }
  virtual int loadFile(const char *projectfile);
  virtual void writeFile(const char *projectfile);
  virtual int getNumFiles();
  virtual const char *enumFile(int n);

  virtual void xmlReaderOnStartElementCallback(const char *xmlpath, const char *xmltag, XmlReaderParams *params);
  virtual void xmlReaderOnEndElementCallback(const char *xmlpath, const char *xmltag);

  const char *getProjectDir() { return m_projectdir; }

private:
  PreprocWorkspace workspace;
  PtrList<FileEntry> m_files;
  PtrListQuickSorted<FileEntry, FileEntryComparator> m_sortedfiles;
  int inxpp;
  String m_projectdir;
};

#endif
